/* SPDX-License-Identifier: GPL-2.0 */
#ifndef _LINUX_VMALLOC_H
#define _LINUX_VMALLOC_H

#include <linux/spinlock.h>
#include <linux/init.h>
#include <linux/list.h>
#include <linux/llist.h>
#include <asm/page.h>		/* pgprot_t */
#include <linux/rbtree.h>
#include <linux/overflow.h>
#include <linux/workqueue.h>
#include <linux/mm_types.h>

#include <asm/vmalloc.h>

struct vm_area_struct;		/* vma defining user mapping in mm_types.h */
struct notifier_block;		/* in notifier.h */
struct iov_iter;		/* in uio.h */

/* bits in flags of vmalloc's vm_struct below */
enum _vmalloc_flag_bits {
	_VM_IOREMAP,		/* ioremap() and friends */
	_VM_ALLOC,		/* vmalloc() */
	_VM_MAP,		/* vmap()ed pages */
	_VM_USERMAP,		/* suitable for remap_vmalloc_range */
	_VM_DMA_COHERENT,	/* dma_alloc_coherent */
	_VM_UNINITIALIZED,	/* vm_struct is not fully initialized */
	_VM_NO_GUARD,		/* ***DANGEROUS*** don't add guard page */
	_VM_KASAN,		/* has allocated kasan shadow memory */
	_VM_RANDOM,		/* randomize allocation within range */
	_VM_ALIGN_END,		/* don't page align, but align allocation to the end */
#ifndef CONFIG_PAX_KERNEXEC
	_VM_FLUSH_RESET_PERMS,	/* reset direct map and flush TLB on unmap, can't be freed in atomic context */
#endif
	_VM_MAP_PUT_PAGES,	/* put pages and free array in vfree */
	_VM_ALLOW_HUGE_VMAP,	/* Allow for huge pages on archs with HAVE_ARCH_HUGE_VMALLOC */

#if (defined(CONFIG_KASAN_GENERIC) || defined(CONFIG_KASAN_SW_TAGS)) && \
	!defined(CONFIG_KASAN_VMALLOC)
	_VM_DEFER_KMEMLEAK,	/* defer kmemleak object creation */
#endif

	_VM_TOPDOWN,		/* allocate from the high end of the range */
	_VM_USERCOPY,		/* allocation intended for copies to userland */

#if defined(CONFIG_X86) && defined(CONFIG_PAX_KERNEXEC)
	_VM_KERNEXEC,		/* allocate from executable kernel memory range */
#endif
	_VMALLOC_FLAGS_LAST_BIT
};

#define __VMALLOC_FLAG_BIT(nr)	(1UL << (nr))
#define __VMALLOC_FLAG_UNUSED	(0UL)

#define VM_IOREMAP		__VMALLOC_FLAG_BIT(_VM_IOREMAP)		/* ioremap() and friends */
#define VM_ALLOC		__VMALLOC_FLAG_BIT(_VM_ALLOC)		/* vmalloc() */
#define VM_MAP			__VMALLOC_FLAG_BIT(_VM_MAP)		/* vmap()ed pages */
#define VM_USERMAP		__VMALLOC_FLAG_BIT(_VM_USERMAP)		/* suitable for remap_vmalloc_range */
#define VM_DMA_COHERENT		__VMALLOC_FLAG_BIT(_VM_DMA_COHERENT)	/* dma_alloc_coherent */
#define VM_UNINITIALIZED	__VMALLOC_FLAG_BIT(_VM_UNINITIALIZED)	/* vm_struct is not fully initialized */
#define VM_NO_GUARD		__VMALLOC_FLAG_BIT(_VM_NO_GUARD)	/* ***DANGEROUS*** don't add guard page */
#define VM_KASAN		__VMALLOC_FLAG_BIT(_VM_KASAN)		/* has allocated kasan shadow memory */
#define VM_RANDOM		__VMALLOC_FLAG_BIT(_VM_RANDOM)		/* randomize allocation within range */
#define VM_ALIGN_END		__VMALLOC_FLAG_BIT(_VM_ALIGN_END)	/* don't page align, but align allocation to the end */
#ifdef CONFIG_PAX_KERNEXEC
#define VM_FLUSH_RESET_PERMS	__VMALLOC_FLAG_UNUSED
#else
#define VM_FLUSH_RESET_PERMS	__VMALLOC_FLAG_BIT(_VM_FLUSH_RESET_PERMS)	/* reset direct map and flush TLB on unmap, can't be freed in atomic context */
#endif
#define VM_MAP_PUT_PAGES	__VMALLOC_FLAG_BIT(_VM_MAP_PUT_PAGES)	/* put pages and free array in vfree */
#define VM_ALLOW_HUGE_VMAP	__VMALLOC_FLAG_BIT(_VM_ALLOW_HUGE_VMAP)	/* Allow for huge pages on archs with HAVE_ARCH_HUGE_VMALLOC */

#if (defined(CONFIG_KASAN_GENERIC) || defined(CONFIG_KASAN_SW_TAGS)) && \
	!defined(CONFIG_KASAN_VMALLOC)
#define VM_DEFER_KMEMLEAK	__VMALLOC_FLAG_BIT(_VM_DEFER_KMEMLEAK)	/* defer kmemleak object creation */
#else
#define VM_DEFER_KMEMLEAK	__VMALLOC_FLAG_UNUSED
#endif

#define VM_TOPDOWN		__VMALLOC_FLAG_BIT(_VM_TOPDOWN)		/* allocate from the high end of the range */
#define VM_USERCOPY		__VMALLOC_FLAG_BIT(_VM_USERCOPY)	/* allocation intended for copies to userland */

/* This would normally only be needed for X86_32, but ensure it is available for all x86 for out-of-tree VirtualBox use */
#if defined(CONFIG_X86) && defined(CONFIG_PAX_KERNEXEC)
#define VM_KERNEXEC		__VMALLOC_FLAG_BIT(_VM_KERNEXEC)	/* allocate from executable kernel memory range */
#endif

/* bits [20..32] reserved for arch specific ioremap internals */

/*
 * Maximum alignment for ioremap() regions.
 * Can be overridden by arch-specific value.
 */
#ifndef IOREMAP_MAX_ORDER
#define IOREMAP_MAX_ORDER	(7 + PAGE_SHIFT)	/* 128 pages */
#endif

struct vm_struct {
	struct vm_struct	*next;
	void			*addr;
	unsigned long		size;
	unsigned long		flags;
	struct page		**pages;
#ifdef CONFIG_HAVE_ARCH_HUGE_VMALLOC
	unsigned int		page_order;
#endif
	unsigned int		nr_pages;
	phys_addr_t		phys_addr;
	const void		*caller;
} __randomize_layout;

struct vmap_area {
	unsigned long va_start;
	unsigned long va_end;

	struct rb_node rb_node;         /* address sorted rbtree */
	struct list_head list;          /* address sorted list */

	/*
	 * The following two variables can be packed, because
	 * a vmap_area object can be either:
	 *    1) in "free" tree (root is free_vmap_area_root)
	 *    2) or "busy" tree (root is vmap_area_root)
	 */
	union {
		unsigned long subtree_max_size; /* in "free" tree */
		struct vm_struct *vm;           /* in "busy" tree */
	};
	unsigned long flags; /* mark type of vm_map_ram area */
} __randomize_layout;

/* archs that select HAVE_ARCH_HUGE_VMAP should override one or more of these */
#ifndef arch_vmap_p4d_supported
static inline bool arch_vmap_p4d_supported(pgprot_t prot)
{
	return false;
}
#endif

#ifndef arch_vmap_pud_supported
static inline bool arch_vmap_pud_supported(pgprot_t prot)
{
	return false;
}
#endif

#ifndef arch_vmap_pmd_supported
static inline bool arch_vmap_pmd_supported(pgprot_t prot)
{
	return false;
}
#endif

#ifndef arch_vmap_pte_range_map_size
static inline unsigned long arch_vmap_pte_range_map_size(unsigned long addr, unsigned long end,
							 u64 pfn, unsigned int max_page_shift)
{
	return PAGE_SIZE;
}
#endif

#ifndef arch_vmap_pte_supported_shift
static inline int arch_vmap_pte_supported_shift(unsigned long size)
{
	return PAGE_SHIFT;
}
#endif

#ifndef arch_vmap_pgprot_tagged
static inline pgprot_t arch_vmap_pgprot_tagged(pgprot_t prot)
{
	return prot;
}
#endif

/*
 *	Highlevel APIs for driver use
 */
extern void vm_unmap_ram(const void *mem, unsigned int count);
extern void *vm_map_ram(struct page **pages, unsigned int count, int node);
extern void vm_unmap_aliases(void);

#ifdef CONFIG_MMU
extern unsigned long vmalloc_nr_pages(void);
#else
static inline unsigned long vmalloc_nr_pages(void) { return 0; }
#endif

#ifdef CONFIG_MMU
extern void *_vmalloc(unsigned long size) __alloc_size(1) __malloc;
static inline __alloc_size(1) __malloc void *vmalloc(unsigned long size)
{
	return __builtin_constant_p(size) && size >= INT_MAX ? NULL:
		_vmalloc(size);
}

extern void *_vmalloc_usercopy(unsigned long size) __alloc_size(1) __malloc;
static inline __alloc_size(1) __malloc void *vmalloc_usercopy(unsigned long size)
{
	return __builtin_constant_p(size) && size >= INT_MAX ? NULL:
		_vmalloc_usercopy(size);
}

extern void *_vzalloc(unsigned long size) __alloc_size(1) __malloc;
static inline __alloc_size(1) __malloc void *vzalloc(unsigned long size)
{
	return __builtin_constant_p(size) && size >= INT_MAX ? NULL:
		_vzalloc(size);
}
#else
extern void *vmalloc(unsigned long size) __alloc_size(1) __malloc;
extern void *vmalloc_usercopy(unsigned long size) __alloc_size(1) __malloc;
extern void *vzalloc(unsigned long size) __alloc_size(1) __malloc;
#endif
extern void *vmalloc_user(unsigned long size) __alloc_size(1) __malloc;
extern void *vmalloc_node(unsigned long size, int node) __alloc_size(1) __malloc;
extern void *vzalloc_node(unsigned long size, int node) __alloc_size(1) __malloc;
extern void *vmalloc_32(unsigned long size) __alloc_size(1) __malloc;
extern void *vzalloc_32(unsigned long size) __alloc_size(1) __malloc;
extern void *vmalloc_32_user(unsigned long size) __alloc_size(1) __malloc;
extern void *__vmalloc(unsigned long size, gfp_t gfp_mask) __alloc_size(1) __malloc;
extern void *__vmalloc_node_range(unsigned long size, unsigned long align,
			unsigned long start, unsigned long end, gfp_t gfp_mask,
			pgprot_t prot, unsigned long vm_flags, int node,
			const void *caller) __alloc_size(1) __malloc;
void *__vmalloc_node(unsigned long size, unsigned long align, gfp_t gfp_mask,
		int node, const void *caller) __alloc_size(1) __malloc;
void *__vmalloc_node_flags(unsigned long size, unsigned long align,
			   gfp_t gfp_mask, unsigned long flags, int node, const void *caller) __alloc_size(1) __malloc;
void *vmalloc_huge(unsigned long size, gfp_t gfp_mask) __alloc_size(1) __malloc;

extern void *__vmalloc_array(size_t n, size_t size, gfp_t flags) __alloc_size(1, 2) __malloc;
extern void *vmalloc_array(size_t n, size_t size) __alloc_size(1, 2) __malloc;
extern void *__vcalloc(size_t n, size_t size, gfp_t flags) __alloc_size(1, 2) __malloc;
extern void *vcalloc(size_t n, size_t size) __alloc_size(1, 2) __malloc;

extern void vfree(const void *addr);
extern void vfree_atomic(const void *addr);

extern void *vmap(struct page **pages, unsigned int count,
			unsigned long flags, pgprot_t prot);
void *vmap_pfn(unsigned long *pfns, unsigned int count, pgprot_t prot);
extern void vunmap(const void *addr);
extern void vunmap_atomic(const void *addr);

#ifdef CONFIG_GRKERNSEC_KSTACKOVERFLOW
extern void unmap_process_stacks(struct task_struct *task);
#endif

extern int remap_vmalloc_range_partial(struct vm_area_struct *vma,
				       unsigned long uaddr, void *kaddr,
				       unsigned long pgoff, unsigned long size);

extern int remap_vmalloc_range(struct vm_area_struct *vma, void *addr,
							unsigned long pgoff);

/*
 * Architectures can set this mask to a combination of PGTBL_P?D_MODIFIED values
 * and let generic vmalloc and ioremap code know when arch_sync_kernel_mappings()
 * needs to be called.
 */
#ifndef ARCH_PAGE_TABLE_SYNC_MASK
#define ARCH_PAGE_TABLE_SYNC_MASK 0
#endif

/*
 * There is no default implementation for arch_sync_kernel_mappings(). It is
 * relied upon the compiler to optimize calls out if ARCH_PAGE_TABLE_SYNC_MASK
 * is 0.
 */
void arch_sync_kernel_mappings(unsigned long start, unsigned long end);

/*
 *	Lowlevel-APIs (not for driver use!)
 */

static inline size_t get_vm_area_size(const struct vm_struct *area)
{
	if (!(area->flags & VM_NO_GUARD))
		/* return actual size without guard page */
		return area->size - PAGE_SIZE;
	else
		return area->size;

}

extern struct vm_struct *get_vm_area(unsigned long size, unsigned long flags);
extern struct vm_struct *get_vm_area_caller(unsigned long size,
					unsigned long flags, const void *caller);
extern struct vm_struct *__get_vm_area_caller(unsigned long size,
					unsigned long flags,
					unsigned long start, unsigned long end,
					const void *caller);
void free_vm_area(struct vm_struct *area);
extern struct vm_struct *remove_vm_area(const void *addr);
extern struct vm_struct *find_vm_area(const void *addr);
extern struct vm_struct *find_vm_area_safe(const void *addr);
struct vmap_area *find_vmap_area(unsigned long addr, bool safe);

enum kstack_kind { CURRENT_KSTACK, FOREIGN_KSTACK, IRQ_KSTACK, KSTACK_KINDS };
#ifdef CONFIG_PAX_PRIVATE_KSTACKS
void map_kstack_on(enum kstack_kind kind, unsigned long addr, unsigned long const pfns[], unsigned int cpu, bool writable);
const struct task_struct *__map_kstack(const struct task_struct *tsk, bool *writable);
const struct task_struct *map_kstack(const struct task_struct *tsk, bool *writable) __opens(kstack);

void unmap_kstack_on(const struct task_struct *tsk, const struct task_struct *cur, unsigned int cpu, bool *writable);
void unmap_kstack(const struct task_struct *tsk, const struct task_struct *prev, bool writable) __closes(kstack);

__visible void prepare_kstack_switch(const struct task_struct *prev, const struct task_struct *next);
__visible void finish_kstack_switch(const struct task_struct *prev, const struct task_struct *next);
#else
static inline void map_kstack_on(enum kstack_kind kind, unsigned long addr, unsigned long const pfns[], unsigned int cpu, bool writable) {}
static inline const struct task_struct *map_kstack(const struct task_struct *tsk, bool *writable) { return NULL; }

static inline void unmap_kstack_on(const struct task_struct *tsk, const struct task_struct *cur, unsigned int cpu, bool *writable) {}
static inline void unmap_kstack(const struct task_struct *tsk, const struct task_struct *prev, bool writable) {}
#endif

static inline bool is_vm_area_hugepages(const void *addr)
{
	/*
	 * This may not 100% tell if the area is mapped with > PAGE_SIZE
	 * page table entries, if for some reason the architecture indicates
	 * larger sizes are available but decides not to use them, nothing
	 * prevents that. This only indicates the size of the physical page
	 * allocated in the vmalloc layer.
	 */
#ifdef CONFIG_HAVE_ARCH_HUGE_VMALLOC
	return find_vm_area(addr)->page_order > 0;
#else
	return false;
#endif
}

#ifdef CONFIG_MMU
void vunmap_range(unsigned long addr, unsigned long end);
static inline void set_vm_flush_reset_perms(void *addr)
{
#ifndef CONFIG_PAX_KERNEXEC
	struct vm_struct *vm = find_vm_area(addr);

	if (vm)
		vm->flags |= VM_FLUSH_RESET_PERMS;
#endif
}

#else
static inline void set_vm_flush_reset_perms(void *addr)
{
}
#endif

/* for /proc/kcore */
extern long vread_iter(struct iov_iter *iter, const char *addr, size_t count);

/*
 *	Internals.  Don't use..
 */
extern struct list_head vmap_area_list;
extern __init void vm_area_add_early(struct vm_struct *vm);
extern __init void vm_area_register_early(struct vm_struct *vm, size_t align);

#ifdef CONFIG_SMP
# ifdef CONFIG_MMU
struct vm_struct **pcpu_get_vm_areas(const unsigned long *offsets,
				     const size_t *sizes, size_t nr_vms,
				     size_t align);

void pcpu_free_vm_areas(struct vm_struct **vms, int nr_vms);
# else
static inline struct vm_struct **
pcpu_get_vm_areas(const unsigned long *offsets,
		const size_t *sizes, size_t nr_vms,
		size_t align)
{
	return NULL;
}

static inline void
pcpu_free_vm_areas(struct vm_struct **vms, int nr_vms)
{
}
# endif
#endif

#ifdef CONFIG_MMU
#define VMALLOC_TOTAL (VMALLOC_END - VMALLOC_START)
#else
#define VMALLOC_TOTAL 0UL
#endif

int register_vmap_purge_notifier(struct notifier_block *nb);
int unregister_vmap_purge_notifier(struct notifier_block *nb);

#if defined(CONFIG_MMU) && defined(CONFIG_PRINTK)
bool vmalloc_dump_obj(void *object);
#else
static inline bool vmalloc_dump_obj(void *object) { return false; }
#endif

#endif /* _LINUX_VMALLOC_H */
