/*
 * Copyright 2019-2024 by PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 *
 * Note: the choice of the license means that the compilation process is
 *       NOT 'eligible' as defined by gcc's library exception to the GPL v3,
 *       but for the kernel it doesn't matter since it doesn't link against
 *       any of the gcc libraries
 *
 * gcc plugin to convert k*alloc calls into type specific slab allocations
 *
 * TODO:
 * - too many to list
 *
 * BUGS:
 * - none known
 */

#include "autoslab.h"

__visible int plugin_is_GPL_compatible;

static struct plugin_info autoslab_plugin_info = {
	.version	= "202407231140",
	.help		= "verbose\tlog autoslab decisions\n"
			  "autostack\tenable autostack conversions\n"
};

bool verbose;
bool autostack;

void autoslab_find_decls(void *event_data, void *data __unused)
{
	tree decl = (tree)event_data;
	enum kmem_call_names kmem_call;

	if (decl == error_mark_node)
		return;

	switch (TREE_CODE(decl)) {
	default:
		return;

	case FUNCTION_DECL:
		if (DECL_ARTIFICIAL(decl))
			return;

		if (DECL_ABSTRACT_ORIGIN(decl) && DECL_ABSTRACT_ORIGIN(decl) != decl)
			return;

		gcc_assert(DECL_NAME_POINTER(decl)[0]);

		for (kmem_call = KMEM_CACHE_ALLOC; kmem_call < KMEM_CACHE_NUM; ++kmem_call) {
			if (!id_equal(DECL_NAME(decl), kmem_cache_decls[kmem_call].name))
				continue;

			if (kmem_cache_decls[kmem_call].decl != NULL_TREE) {
				gcc_assert(decl == kmem_cache_decls[kmem_call].decl);
				continue;
			}

			kmem_cache_decls[kmem_call].decl = decl;
			return;
		}
	}
}

static void register_attributes(void *event_data __unused, void *data __unused)
{
	register_attribute(&alloc_flags_attr);
	register_attribute(&alloc_node_attr);
	register_attribute(&alloc_attr);
}

__visible int plugin_init(struct plugin_name_args *plugin_info, struct plugin_gcc_version *version)
{
	int i;
	const int argc = plugin_info->argc;
	const struct plugin_argument * const argv = plugin_info->argv;
	const char * const plugin_name = plugin_info->base_name;
	struct opt_pass *reference_pass;

	PASS_INFO(autotypename, "increase_alignment", 1, PASS_POS_INSERT_AFTER);
	PASS_INFO(autoslab, "slsr", 1, PASS_POS_INSERT_AFTER);

	if (!plugin_default_version_check(version, &gcc_version)) {
		error_gcc_version(version);
		return 1;
	}

	for (i = 0; i < argc; ++i) {
		if (!strcmp(argv[i].key, "verbose")) {
			verbose = true;
			continue;
		}

		if (!strcmp(argv[i].key, "autostack")) {
			autostack = true;
			continue;
		}

		error(G_("unknown option '-fplugin-arg-%s-%s'"), plugin_name, argv[i].key);
	}

	register_callback(plugin_name, PLUGIN_INFO, NULL, &autoslab_plugin_info);
	register_callback(plugin_name, PLUGIN_FINISH_DECL, autoslab_find_decls, NULL);
	register_callback(plugin_name, PLUGIN_PRE_GENERICIZE, autoslab_find_decls, NULL);
#if (BUILDING_GCC_VERSION >= 9003 && BUILDING_GCC_VERSION < 10000) || BUILDING_GCC_VERSION >= 10001
	register_callback(plugin_name, PLUGIN_PRE_GENERICIZE, autoslab_resolve_typenames, NULL);
#endif
	register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &autotypename_pass_info);

#if BUILDING_GCC_VERSION >= 13000
	reference_pass = plugin_find_pass("respectre");
	if (!reference_pass)
#endif
		reference_pass = plugin_find_pass("size_overflow_transform");
	if (reference_pass) {
		autoslab_pass_info.reference_pass_name = reference_pass->name;
		autoslab_pass_info.pos_op = PASS_POS_INSERT_BEFORE;
	}

	register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &autoslab_pass_info);
	register_callback(plugin_name, PLUGIN_REGISTER_GGC_ROOTS, NULL, (void *)&gt_ggc_r_autoslab);
	register_callback(plugin_name, PLUGIN_ATTRIBUTES, register_attributes, NULL);

	gcc_assert(!regcomp(&sizeof_regex, "^[[:space:]\\(]*sizeof[[:space:]]*\\([[:space:]]*(struct|union)[[:space:]]*([_[:alnum:]]+)[[:space:]]*\\)[[:space:]\\)]*$", REG_EXTENDED));
	return 0;
}
