/*
 * Copyright 2019-2024 by PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 */

#include "autoslab.h"

GTY(()) struct kmem_cache_decl kmem_cache_decls[KMEM_CACHE_NUM] = {
#define KMEM_CACHE_DECL(NODE,NOPROF,INDEX,TYPE,node,noprof,index,type)	\
[KMEM_CACHE_ALLOC##NODE##NOPROF##INDEX##TYPE] = {			\
	.name = "kmem_cache_alloc" node noprof index type,		\
	.decl = NULL_TREE,						\
},
#include "autoslab.def"
};

EXPORTED_CONST struct ggc_root_tab gt_ggc_r_autoslab[] = {
	INIT_GGC_ROOT_TAB_TREES(kmem_cache_decls, decl),
	LAST_GGC_ROOT_TAB
};

const struct kalloc_function kalloc_functions[KMALLOC_NUM] = {
#define KMALLOC_DECL(NAME,NODE,TYPENAME,NOPROF,name,node,typename,noprof) \
[NAME##NODE##TYPENAME##NOPROF] = {					\
	.kalloc_name		= name node typename noprof,		\
	.kmem_name		= KMEM_CACHE_ALLOC##NOPROF,		\
	.kmem_name_type		= KMEM_CACHE_ALLOC##NOPROF##_TYPE,	\
	.kmem_name_index	= KMEM_CACHE_ALLOC##NOPROF##_INDEX,	\
	.kmem_name_index_type	= KMEM_CACHE_ALLOC##NOPROF##_INDEX_TYPE,\
},
#include "autoslab.def"
};

struct size_analysis {
	HOST_WIDE_INT c1;
	HOST_WIDE_INT c0;
	HOST_WIDE_INT minshift;
	HOST_WIDE_INT maxshift;
};

static void __autoslab_analyze_size(tree size, struct size_analysis *sa);

static size_t is_kalloc(const char *fnname)
{
	size_t i;

	for (i = 0; i < ARRAY_SIZE(kalloc_functions); i++)
		if (!strcmp(fnname, kalloc_functions[i].kalloc_name))
			break;
	return i;
}

bool is_cache_alloc(const char *fnname)
{
	size_t i;

	for (i = 0; i < ARRAY_SIZE(kmem_cache_decls); i++)
		if (!strcmp(fnname, kmem_cache_decls[i].name))
			return true;
	return false;
}

tree autoslab_build_var(location_t loc, tree id, tree type, tree ctor)
{
	tree var;

	var = build_decl(loc, VAR_DECL, id, type);
	DECL_INITIAL(var) = ctor;
	DECL_EXTERNAL(var) = 0;
	DECL_IGNORED_P(var) = 0;
	DECL_ARTIFICIAL(var) = 1;
	SET_DECL_ALIGN(var, UNITS_PER_WORD * BITS_PER_UNIT);
	DECL_USER_ALIGN(var) = 1;

	TREE_USED(var) = 1;
	return var;
}

static tree autoslab_build_global_var(location_t loc, tree id, tree type, tree ctor, const char *section)
{
	tree var;

	var = autoslab_build_var(loc, id, type, ctor);

	DECL_CONTEXT(var) = DECL_CONTEXT(current_function_decl);
	DECL_PRESERVE_P(var) = 1;

	TREE_ADDRESSABLE(var) = 1;
	TREE_STATIC(var) = 1;
	TREE_PUBLIC(var) = 1;

	if (section)
		set_decl_section_name(var, section);
	else
		TREE_READONLY(var) = 1;

	varpool_add_new_variable(var);
	return var;
}

static tree autoslab_build_cache_ptr(location_t loc, tree id, tree type, tree ctor)
{
	const char *section_name;
	tree section;

	section = lookup_attribute("section", DECL_ATTRIBUTES(current_function_decl));
	if (section && TREE_VALUE(section)) {
		section = TREE_VALUE(TREE_VALUE(section));

		if (!strncmp(TREE_STRING_POINTER(section), ".init.text", 10))
			section_name = ".data.init.autoslab";
		else if (!strncmp(TREE_STRING_POINTER(section), ".meminit.text", 13))
			section_name = ".data.init.autoslab";
		else
			section_name = ".data.autoslab";
		if (verbose)
			inform(loc, "autoslab: yes section %E %E", section, id);
	} else {
		section_name = ".data.autoslab";
		if (verbose)
			inform(loc, "autoslab: yes section \".text\" %E", id);
	}

	return autoslab_build_global_var(loc, id, type, ctor, section_name);
}

static tree get_alloc_type(tree *tp, int *walk_subtrees, void *data)
{
	struct walk_stmt_info *wi = (struct walk_stmt_info *)data;
	tree lhs = (tree)wi->info;

	if (TYPE_P(*tp)) {
		*walk_subtrees = 0;
		return NULL_TREE;
	}

	return operand_equal_p(*tp, lhs, 0) ? lhs : NULL_TREE;
}

static tree __autoslab_get_alloc_type(gimple_stmt_ptr use_stmt, tree lhs)
{
	tree lhs_type;

	switch (gimple_code(use_stmt)) {
	default:
		print_gimple_stmt(stderr, use_stmt, 0, TDF_LINENO);
		gcc_unreachable();

	case GIMPLE_ASM:
		// TODO
		return NULL_TREE;

	case GIMPLE_ASSIGN: {
		struct walk_stmt_info wi = { };

		wi.info = lhs;
		gcc_assert(walk_gimple_op(use_stmt, get_alloc_type, &wi));
		if (wi.is_lhs) {
			lhs = gimple_assign_lhs(use_stmt);
			lhs = get_inner_ref(lhs);
			switch (TREE_CODE(lhs)) {
			default:
				debug_gimple_stmt(use_stmt);
				debug_tree(lhs);
				gcc_unreachable();

			case MEM_REF:
				// FALLTHROUGH
			case TARGET_MEM_REF:
				lhs_type = TYPE_MAIN_VARIANT(TREE_TYPE(TREE_OPERAND(lhs, 1)));
				break;
			}
		} else {
			enum tree_code code;

			lhs = gimple_assign_rhs1(use_stmt);
			lhs = get_inner_ref(lhs);
			code = TREE_CODE(lhs);
			if (code == ADDR_EXPR) {
				lhs = TREE_OPERAND(lhs, 0);
				code = TREE_CODE(lhs);
			}
			switch (code) {
			default:
				lhs = gimple_assign_lhs(use_stmt);
				lhs_type = TREE_TYPE(lhs);
				break;

			case MEM_REF:
				// FALLTHROUGH
			case TARGET_MEM_REF:
				gcc_assert(lhs == get_inner_ref(lhs));
				lhs_type = TYPE_MAIN_VARIANT(TREE_TYPE(TREE_OPERAND(lhs, 1)));
				break;
			}
		}

		switch (TREE_CODE(lhs)) {
		default:
			fprintf(stderr, "USE %s %s ", get_tree_code_name(gimple_assign_rhs_code(use_stmt)), wi.is_lhs ? "lhs" : "rhs");
			debug_gimple_stmt(use_stmt);
			debug_tree(lhs);
			debug_tree(lhs_type);
			gcc_unreachable();

		case ARRAY_REF:
		case COMPONENT_REF:
		case PARM_DECL:
		case VAR_DECL:
			break;

		case MEM_REF:
			if (!integer_zerop(TREE_OPERAND(lhs, 1)))
				return NULL_TREE;
			break;

		case TARGET_MEM_REF:
			if (TMR_OFFSET(lhs) && !integer_zerop(TMR_OFFSET(lhs)))
				return NULL_TREE;
			break;

		case NOP_EXPR:
		case SSA_NAME:
			if (!wi.is_lhs)
				return NULL_TREE;
			lhs = gimple_assign_lhs(use_stmt);
			if (TREE_CODE(lhs) != SSA_NAME)
				break;
			return autoslab_get_alloc_type(lhs);

		case EQ_EXPR:
		case GE_EXPR:
		case GT_EXPR:
		case LE_EXPR:
		case LT_EXPR:
		case NE_EXPR:
		case PLUS_EXPR:
		case POINTER_PLUS_EXPR:
			return NULL_TREE;
		}
		break;
	}

	case GIMPLE_CALL: {
		unsigned int i, e;
		tree argtype;

		argtype = gimple_call_fntype(use_stmt);
		if (argtype == NULL_TREE)
			return NULL_TREE;
		argtype = TYPE_ARG_TYPES(gimple_call_fntype(use_stmt));
		lhs_type = NULL_TREE;
		for (i = 0, e = gimple_call_num_args(use_stmt); i < e; ++i, argtype = TREE_CHAIN(argtype)) {
			tree arg, type_size;

			if (argtype == NULL_TREE)
				return NULL_TREE;

			arg = gimple_call_arg(use_stmt, i);
			if (!operand_equal_p(lhs, arg, 0))
				continue;

			arg = TREE_VALUE(argtype);
			if (!POINTER_TYPE_P(arg))
				continue;
			arg = TREE_TYPE(arg);
			if (!RECORD_OR_UNION_TYPE_P(arg))
				continue;

			type_size = TYPE_SIZE(arg);
			if (type_size == NULL_TREE)
				continue;

			gcc_assert(TREE_CODE(type_size) == INTEGER_CST);

			if (lhs_type == NULL_TREE || tree_int_cst_compare(type_size, TYPE_SIZE(lhs_type)) > 0)
				lhs_type = TREE_VALUE(argtype);
		}
		if (lhs_type == NULL_TREE)
			return NULL_TREE;
		break;
	}

	case GIMPLE_COND:
		return NULL_TREE;

	case GIMPLE_DEBUG: {
		tree var, value;

		if (!gimple_debug_bind_p(use_stmt))
			return NULL_TREE;

		var = gimple_debug_bind_get_var(use_stmt);
		value = gimple_debug_bind_get_value(use_stmt);

		switch (TREE_CODE(value)) {
		default:
			debug_gimple_stmt(use_stmt);
			debug_tree(lhs);
			debug_tree(var);
			debug_tree(value);
			gcc_unreachable();

		// TODO
		case ADDR_EXPR:
		case ARRAY_REF:
		case BIT_FIELD_REF:
		case BIT_IOR_EXPR:
		case COMPONENT_REF:
		case EQ_EXPR:
		case NE_EXPR:
		case MEM_REF:
		case MINUS_EXPR:
		case NOP_EXPR:
		case PLUS_EXPR:
#if BUILDING_GCC_VERSION >= 8000
		case POINTER_DIFF_EXPR:
#endif
			return NULL_TREE;

		case POINTER_PLUS_EXPR:
			if (!integer_zerop(TREE_OPERAND(value, 1)))
				return NULL_TREE;
			break;

		case SSA_NAME:
			break;
		}

		lhs_type = TREE_TYPE(var);
		break;
	}

	case GIMPLE_PHI: {
		gphi *phi_stmt = as_a_gphi(use_stmt);
		imm_use_iterator imm_iter;
		tree type_size, alloc_type = NULL_TREE;

		if (gimple_visited_p(phi_stmt))
			return NULL_TREE;
		gimple_set_visited(phi_stmt, true);

		lhs = gimple_phi_result(phi_stmt);
		FOR_EACH_IMM_USE_STMT(use_stmt, imm_iter, lhs) {
			lhs_type = __autoslab_get_alloc_type(use_stmt, lhs);
			if (lhs_type == NULL_TREE)
				continue;

			type_size = TYPE_SIZE(lhs_type);
			if (type_size == NULL_TREE)
				continue;

			gcc_assert(TREE_CODE(type_size) == INTEGER_CST);
			if (alloc_type == NULL_TREE || tree_int_cst_compare(type_size, TYPE_SIZE(alloc_type)) > 0)
				alloc_type = lhs_type;
		}
		gimple_set_visited(phi_stmt, false);
		return alloc_type;
	}

	case GIMPLE_RETURN:
		lhs = gimple_return_retval(as_a_greturn(use_stmt));
		lhs_type = TREE_TYPE(lhs);
		break;
	}

	if (!POINTER_TYPE_P(lhs_type))
		return NULL_TREE;

	lhs_type = strip_array_types(TREE_TYPE(lhs_type));
	if (RECORD_OR_UNION_TYPE_P(lhs_type))
		return lhs_type;

	if (!POINTER_TYPE_P(lhs_type))
		return NULL_TREE;

	if (RECORD_OR_UNION_TYPE_P(TREE_TYPE(lhs_type)))
		return lhs_type;

	if (!POINTER_TYPE_P(TREE_TYPE(lhs_type)))
		return NULL_TREE;

	if (RECORD_OR_UNION_TYPE_P(TREE_TYPE(TREE_TYPE(lhs_type))))
		return lhs_type;

	return NULL_TREE;
}

tree autoslab_get_alloc_type(tree lhs)
{
	tree lhs_type, ptr_alloc_type, ptr_ptr_alloc_type;
	imm_use_iterator imm_iter;
	gimple_stmt_ptr use_stmt;

	gcc_assert(lhs);
	lhs_type = TREE_TYPE(lhs);
	gcc_assert(POINTER_TYPE_P(lhs_type));
	lhs_type = TREE_TYPE(lhs_type);

	if (RECORD_OR_UNION_TYPE_P(lhs_type))
		return lhs_type;

	ptr_ptr_alloc_type = ptr_alloc_type = NULL_TREE;
	FOR_EACH_IMM_USE_STMT(use_stmt, imm_iter, lhs) {
		tree type_size;

		lhs_type = __autoslab_get_alloc_type(use_stmt, lhs);

		if (lhs_type == NULL_TREE)
			continue;

		if (POINTER_TYPE_P(lhs_type))
			type_size = TYPE_SIZE(TREE_TYPE(lhs_type));
		else
			type_size = TYPE_SIZE(lhs_type);

		if (type_size == NULL_TREE)
			continue;

		gcc_assert(TREE_CODE(type_size) == INTEGER_CST);

		if (POINTER_TYPE_P(lhs_type)) {
			if (ptr_ptr_alloc_type == NULL_TREE) {
				ptr_ptr_alloc_type = lhs_type;
				continue;
			}
		} else {
			if (ptr_alloc_type == NULL_TREE) {
				ptr_alloc_type = lhs_type;
				continue;
			}
		}

#if 0
		if (POINTER_TYPE_P(lhs_type)) {
			gcc_assert(POINTER_TYPE_P(ptr_ptr_alloc_type));
			if (tree_int_cst_compare(type_size, TYPE_SIZE(TREE_TYPE(ptr_ptr_alloc_type))) > 0)
				ptr_ptr_alloc_type = lhs_type;
		} else {
			gcc_assert(!POINTER_TYPE_P(ptr_alloc_type));
			if (tree_int_cst_compare(type_size, TYPE_SIZE(ptr_alloc_type)) > 0)
				ptr_alloc_type = lhs_type;
		}
#endif

		if (lhs_type != ptr_alloc_type)
			RETURN_FROM_IMM_USE_STMT(imm_iter, NULL_TREE);
	}

	if (ptr_alloc_type == NULL_TREE && ptr_ptr_alloc_type == NULL_TREE)
		return NULL_TREE;

	if ((ptr_alloc_type != NULL_TREE) != (ptr_ptr_alloc_type != NULL_TREE))
		return ptr_alloc_type ? : ptr_ptr_alloc_type;

	if (tree_int_cst_compare(TYPE_SIZE(TREE_TYPE(ptr_ptr_alloc_type)), TYPE_SIZE(ptr_alloc_type)) > 0)
		return ptr_ptr_alloc_type;
	return ptr_alloc_type;
}

static tree autoslab_gen_name(const char *kind,
			      const char *type,
			      gimple_stmt_iterator *gsi,
			      unsigned HOST_WIDE_INT size_val,
			      unsigned HOST_WIDE_INT lhs_align_val,
			      unsigned HOST_WIDE_INT flags_val,
			      sbitmap gfp_zones,
			      HOST_WIDE_INT minshift,
			      HOST_WIDE_INT maxshift,
			      HOST_WIDE_INT c1,
			      HOST_WIDE_INT c0)
{
	static unsigned int counter;
	gimple_stmt_ptr stmt;
	location_t loc1, loc2;
	char temp[tree_to_uhwi(KSYM_NAME_LEN)];
	char *name;
	int len;
	tree autoslab_name;

	stmt = gsi_stmt(*gsi);
	loc1 = gimple_inline_point_location(stmt);
	loc2 = gimple_expansion_point_location(stmt);

	gcc_assert(strlen(TREE_STRING_POINTER(KBUILD_MODNAME)));
	gcc_assert(strlen(TREE_STRING_POINTER(AUTOSLAB_BASENAME)));
	len = snprintf(temp, sizeof temp,
		"autoslab_%s%s"
		"M_%s_P_%s_%u_%u_%u_%u_S_"
		HOST_WIDE_INT_PRINT_UNSIGNED "_A_" HOST_WIDE_INT_PRINT_UNSIGNED
		"_%s",
		kind, type,
		TREE_STRING_POINTER(KBUILD_MODNAME),
		TREE_STRING_POINTER(AUTOSLAB_BASENAME),
		LOCATION_LINE(loc1),
		LOCATION_COLUMN(loc1),
		LOCATION_LINE(loc2),
		LOCATION_COLUMN(loc2),
		size_val,
		lhs_align_val,
		sbitmap_count_bits(gfp_zones) > 1 ? "i" :
		ZONE_DMA && sbitmap_bit_p(gfp_zones, tree_to_uhwi(ZONE_DMA)) ? "d" :
		ZONE_DMA32 && sbitmap_bit_p(gfp_zones, tree_to_uhwi(ZONE_DMA32)) ? "D" :
		!sbitmap_bit_p(gfp_zones, tree_to_uhwi(ZONE_NORMAL)) ? NULL :
		flags_val & tree_to_uhwi(__GFP_SEALED) ? "s" :
		flags_val & tree_to_uhwi(__GFP_RECLAIMABLE) ? "r" : "n");

	if (len == -1 || (unsigned long)len >= sizeof temp)
		return NULL_TREE;

	if (strcmp(kind, "sizes_"))
		len = asprintf(&name, "%s_%u", temp, counter++);
	else
		len = asprintf(&name,
			"%s_MIN_" HOST_WIDE_INT_PRINT_DEC
			"_MAX_" HOST_WIDE_INT_PRINT_DEC
			"_C1_" HOST_WIDE_INT_PRINT_DEC
			"_C0_" HOST_WIDE_INT_PRINT_DEC "_%u",
			temp,
			minshift,
			maxshift,
			c1,
			c0,
			counter++);

	if (len == -1)
		return NULL_TREE;

	autoslab_name = get_identifier(name);
	free(name);

	if ((unsigned HOST_WIDE_INT)len < tree_to_uhwi(KSYM_NAME_LEN))
		return autoslab_name;

	if (!seen_error())
		error_at(loc2, "autoslab: increase KSYM_NAME_LEN to %qd at least because of %qs!", len, name);
	return NULL_TREE;
}

static tree autoslab_create_info(gimple_stmt_iterator *gsi,
				 tree size,
				 tree flags,
				 sbitmap gfp_zones,
				 const char *type,
				 const char *lhs_type_name_ptr,
				 unsigned HOST_WIDE_INT lhs_align_val)
{
	gcall *kalloc_call_stmt;
	tree field, value, ctor, autoslab_info;
	unsigned HOST_WIDE_INT size_val, flags_val;
	tree autoslab_info_name, autoslab_name;
	varpool_node_ptr var;
	location_t loc;
	vec<constructor_elt, va_gc> *initializer = NULL;

	kalloc_call_stmt = as_a_gcall(gsi_stmt(*gsi));
	loc = gimple_expansion_point_location(kalloc_call_stmt);
	size_val = tree_to_uhwi(size);
	flags_val = TREE_CODE(flags) == INTEGER_CST ? tree_to_uhwi(flags) : 0;

	// build:
	// const struct autoslab_info i __initdata = &(struct autoslab_info){"name", "typename", size, align, flags, useroffset, usersize};
	// struct autoslab_info *pi = &i;

	// construct the initializer for autoslab_info
	gcc_assert(autoslab_info_type);
	field = TYPE_FIELDS(autoslab_info_type);
	vec_alloc(initializer, list_length(field));

	// .name
	autoslab_name = autoslab_gen_name("", type, gsi, size_val, lhs_align_val, flags_val, gfp_zones, 0, 0, 0, 0);
	gcc_assert(autoslab_name);
	value = build_string_literal(IDENTIFIER_LENGTH(autoslab_name) + 1, IDENTIFIER_POINTER(autoslab_name));
	CONSTRUCTOR_APPEND_ELT(initializer, field, value);

	// .typename
	field = TREE_CHAIN(field);
	value = build_string_literal(strlen(lhs_type_name_ptr) + 1, lhs_type_name_ptr);
	CONSTRUCTOR_APPEND_ELT(initializer, field, value);

	// .size
	field = TREE_CHAIN(field);
	value = fold_binary_to_constant(MAX_EXPR, TREE_TYPE(field), size, build_int_cst(TREE_TYPE(field), 8));
	gcc_assert(value);
	CONSTRUCTOR_APPEND_ELT(initializer, field, value);

	// .align
	field = TREE_CHAIN(field);
	value = fold_convert_loc(loc, TREE_TYPE(field), build_int_cst(TREE_TYPE(field), lhs_align_val));
	gcc_assert(value);
	CONSTRUCTOR_APPEND_ELT(initializer, field, value);

	// .flags
	field = TREE_CHAIN(field);
	if (flags_val & tree_to_uhwi(__GFP_DMA))
		value = fold_convert_loc(loc, TREE_TYPE(field), SLAB_CACHE_DMA);
	else if (flags_val & tree_to_uhwi(__GFP_DMA32))
		value = fold_convert_loc(loc, TREE_TYPE(field), SLAB_CACHE_DMA32);
	else if (flags_val & tree_to_uhwi(__GFP_RECLAIMABLE))
		value = fold_convert_loc(loc, TREE_TYPE(field), SLAB_RECLAIM_ACCOUNT);
	else if (flags_val & tree_to_uhwi(__GFP_USERCOPY))
		value = fold_convert_loc(loc, TREE_TYPE(field), SLAB_USERCOPY);
	else if (flags_val & tree_to_uhwi(__GFP_SEALED))
		value = fold_convert_loc(loc, TREE_TYPE(field), SLAB_SEALED);
	else
		value = fold_convert_loc(loc, TREE_TYPE(field), integer_zero_node);
	CONSTRUCTOR_APPEND_ELT(initializer, field, value);

	// .useroffset
	field = TREE_CHAIN(field);
	value = build_int_cst(TREE_TYPE(field), 0);
	CONSTRUCTOR_APPEND_ELT(initializer, field, value);

	// .usersize
	field = TREE_CHAIN(field);
	if (flags_val & tree_to_uhwi(__GFP_SEALED))
		value = fold_convert_loc(loc, TREE_TYPE(field), integer_zero_node);
	else
		value = fold_convert_loc(loc, TREE_TYPE(field), size);
	CONSTRUCTOR_APPEND_ELT(initializer, field, value);

	ctor = build_constructor(autoslab_info_type, initializer);

	TREE_STATIC(ctor) = 1;
	TREE_PUBLIC(ctor) = 0;
	TREE_CONSTANT(ctor) = 1;
	TREE_USED(ctor) = 1;

	autoslab_info_name = autoslab_gen_name("info_", type, gsi, size_val, lhs_align_val, flags_val, gfp_zones, 0, 0, 0, 0);
	gcc_assert(autoslab_info_name);
	var = varpool_get_for_asmname(autoslab_info_name);
	gcc_assert(!var);
	autoslab_info = autoslab_build_global_var(loc, autoslab_info_name, autoslab_info_type, ctor, ".init.data.autoslab");

	return autoslab_info;
}

typedef vec<constructor_elt, va_gc> *ctor_values;

static ctor_values autoslab_create_infos_initializer(gimple_stmt_iterator *gsi,
						     tree size,
						     sbitmap gfp_zones,
				 		     const char *lhs_type_name_ptr,
						     unsigned HOST_WIDE_INT lhs_align_val)
{
	gcall *kalloc_call_stmt;
	tree field, value;
	location_t loc;
	ctor_values ctor_values = NULL;
	vec_alloc(ctor_values, list_length(TYPE_FIELDS(autoslab_infos_type)));

	kalloc_call_stmt = as_a_gcall(gsi_stmt(*gsi));
	loc = gimple_expansion_point_location(kalloc_call_stmt);

	gcc_assert(autoslab_infos_type);
	for (field = TYPE_FIELDS(autoslab_infos_type); field != NULL_TREE; field = TREE_CHAIN(field)) {
		tree flags;
		const char *type;

		if (id_equal(DECL_NAME(field), "normal")) {
			// .normal
			flags = integer_zero_node;
			type = "normal_";
#if 0
		} else if (id_equal(DECL_NAME(field), "reclaimable")) {
			// .reclaimable
			flags = __GFP_RECLAIMABLE;
			type = "reclaimable_";
#endif
		} else if (id_equal(DECL_NAME(field), "dma")) {
			// .dma
			flags = __GFP_DMA;
			type = "dma_";
		} else if (id_equal(DECL_NAME(field), "dma32")) {
			// .dma32
			flags = __GFP_DMA32;
			type = "dma32_";
		} else {
			debug_tree(autoslab_infos_type);
			debug_tree(field);
			gcc_unreachable();
		}

		value = autoslab_create_info(gsi, size, flags, gfp_zones, type, lhs_type_name_ptr, lhs_align_val);
		value = build1_loc(loc, ADDR_EXPR, TYPE_POINTER_TO(autoslab_info_type), value);
		CONSTRUCTOR_APPEND_ELT(ctor_values, field, value);
	}

	return ctor_values;
}

static tree autoslab_create_infos_type(gimple_stmt_iterator *gsi,
				       tree size,
				       sbitmap gfp_zones,
				       const char *lhs_type_name_ptr,
				       unsigned HOST_WIDE_INT lhs_align_val)
{
	gcall *kalloc_call_stmt;
	tree ctor, autoslab_infos;
	unsigned HOST_WIDE_INT size_val;
	tree autoslab_infos_name;
	varpool_node_ptr var;
	location_t loc;

	kalloc_call_stmt = as_a_gcall(gsi_stmt(*gsi));
	loc = gimple_expansion_point_location(kalloc_call_stmt);
	size_val = tree_to_uhwi(size);

	autoslab_infos_name = autoslab_gen_name("infos_", "", gsi, size_val, lhs_align_val, 0, gfp_zones, 0, 0, 0, 0);
	gcc_assert(autoslab_infos_name);
	var = varpool_get_for_asmname(autoslab_infos_name);
	gcc_assert(!var);

	// build:
	// struct autoslab_infos autoslabs = { .normal = autoslab_normal, .dma = autoslab_dma, .dma32 = autoslab_dma32 };

	ctor = build_constructor(autoslab_infos_type, autoslab_create_infos_initializer(gsi, size, gfp_zones, lhs_type_name_ptr, lhs_align_val));

	autoslab_infos = autoslab_build_cache_ptr(loc, autoslab_infos_name, autoslab_infos_type, ctor);

	return autoslab_infos;
}

static tree autoslab_analyze_size(gimple_stmt_iterator *gsi,
				  tree size,
				  struct size_analysis *sa)
{
	gcall *kalloc_call_stmt;
	unsigned HOST_WIDE_INT kmalloc_max, minsize, maxsize, maxsize_ceil;

	gcc_assert(TREE_CODE(size) == INTEGER_CST || TREE_CODE(size) == SSA_NAME);

	kalloc_call_stmt = as_a_gcall(gsi_stmt(*gsi));

	__autoslab_analyze_size(size, sa);

	if (sa->c1 == 0) {
		sa->minshift = ceil_log2(sa->c0);
		sa->maxshift = ceil_log2(sa->c0);

		if (verbose)
			inform(gimple_inline_point_location(kalloc_call_stmt),
			       "autoslab: CONSTSIZE"
			       " minshift " HOST_WIDE_INT_PRINT_DEC
			       " maxshift " HOST_WIDE_INT_PRINT_DEC
			       " c1 " HOST_WIDE_INT_PRINT_DEC
			       " c0 " HOST_WIDE_INT_PRINT_DEC,
			       sa->minshift, sa->maxshift, sa->c1, sa->c0);

		return build_int_cstu(sizetype, sa->c0);
	}

	gcc_assert(TREE_CODE(size) == SSA_NAME);
	minsize = tree_to_uhwi(KMALLOC_MIN_SIZE);
	maxsize = kmalloc_max = tree_to_uhwi(KMALLOC_MAX_SIZE);

	wide_int minrange, maxrange;
	location_t loc = gimple_expansion_point_location(kalloc_call_stmt);

	switch (get_range_info(kalloc_call_stmt, size, &minrange, &maxrange)) {
	default:
		break;

	case VR_ANTI_RANGE:
		if (minrange.to_uhwi() == 1)
			break;
		if (minrange.to_uhwi() >= kmalloc_max + 1)
			break;
		if (maxrange.to_uhwi() <= kmalloc_max - 1)
			break;

		maxsize = minrange.to_uhwi() - 1;
		if (minsize > maxsize)
			maxsize = minsize;
		if (verbose)
			inform(loc, "autoslab: SIZE ANTIRANGE 0-" HOST_WIDE_INT_PRINT_UNSIGNED "," HOST_WIDE_INT_PRINT_UNSIGNED "-inf c1 " HOST_WIDE_INT_PRINT_DEC " c0 " HOST_WIDE_INT_PRINT_DEC,
				minrange.to_uhwi() - 1, maxrange.to_uhwi() + 1, sa->c1, sa->c0);
		break;

	case VR_RANGE:
		if (maxrange.to_uhwi() >= kmalloc_max)
			break;

		if (minrange.to_uhwi() > minsize)
			minsize = minrange.to_uhwi();
		if (maxrange.to_uhwi() > minsize)
			maxsize = maxrange.to_uhwi();
		else
			maxsize = minsize;

		if (verbose)
			inform(loc, "autoslab: SIZE RANGE " HOST_WIDE_INT_PRINT_UNSIGNED "-" HOST_WIDE_INT_PRINT_UNSIGNED  " c1 " HOST_WIDE_INT_PRINT_DEC " c0 " HOST_WIDE_INT_PRINT_DEC,
				minrange.to_uhwi(), maxrange.to_uhwi(), sa->c1, sa->c0);
		break;
	}

	if (sa->c0 > maxsize / 2 || sa->c0 + sa->c1 > maxsize / 2) {
		sa->c0 = 0;
		sa->c1 = 1;
	}

	sa->minshift = minsize < sa->c0 + sa->c1 ? ceil_log2(sa->c1 + sa->c0) : ceil_log2(minsize);
	gcc_assert(maxsize >= sa->c0 + sa->c1);
	sa->maxshift = ceil_log2(maxsize);

	if (verbose)
		inform(gimple_inline_point_location(kalloc_call_stmt),
		       "autoslab: VARSIZE minsize " HOST_WIDE_INT_PRINT_UNSIGNED
		       " maxsize " HOST_WIDE_INT_PRINT_UNSIGNED
		       " minshift " HOST_WIDE_INT_PRINT_DEC
		       " maxshift " HOST_WIDE_INT_PRINT_DEC
		       " c1 " HOST_WIDE_INT_PRINT_DEC
		       " c0 " HOST_WIDE_INT_PRINT_DEC,
		       minsize, maxsize, sa->minshift, sa->maxshift, sa->c1, sa->c0);

	maxsize_ceil = ((1UL << sa->maxshift) - sa->c0) / sa->c1 * sa->c1 + sa->c0;
	if (maxsize_ceil < maxsize)
		maxsize = maxsize_ceil;
	size = build_int_cstu(unsigned_type_node, maxsize);
	gcc_assert(tree_int_cst_compare(size, KMALLOC_MAX_SIZE) <= 0);
	return size;
}

static tree autoslab_create_infos_size(gimple_stmt_iterator *gsi,
				       tree flags,
				       sbitmap gfp_zones,
				       const char *lhs_type_name_ptr,
				       struct size_analysis *sa,
				       unsigned HOST_WIDE_INT maxsize,
				       unsigned HOST_WIDE_INT lhs_align_val)
{
	tree ctor, autoslab_infos;
	tree autoslab_infos_name;
	varpool_node_ptr var;
	location_t loc;
	unsigned HOST_WIDE_INT flags_val;
	unsigned int i;
	bool varflags = sbitmap_count_bits(gfp_zones) > 1;

	loc = gimple_expansion_point_location(gsi_stmt(*gsi));
	flags_val = TREE_CODE(flags) == INTEGER_CST ? tree_to_uhwi(flags) : 0;
	autoslab_infos_name = autoslab_gen_name("sizes_", "", gsi, 0, lhs_align_val, flags_val, gfp_zones, sa->minshift, sa->maxshift, sa->c1, sa->c0);
	gcc_assert(autoslab_infos_name);
	var = varpool_get_for_asmname(autoslab_infos_name);
	gcc_assert(!var);

	ctor_values initializer = NULL;
	vec_alloc(initializer, sa->maxshift - sa->minshift + 1);

	// build:
	// struct autoslab_infos *autoslabs = { [i] = &autoslab_size_<2^i/c1*c1>, ... };

	// loop over all sizes and create the autoslab_info for each
	for (i = sa->minshift; i <= sa->maxshift; i++) {
		unsigned HOST_WIDE_INT size;

		size = ((1UL << i) - sa->c0) / sa->c1 * sa->c1 + sa->c0;
		if (size > maxsize)
			size = maxsize;
		gcc_assert(size <= tree_to_uhwi(KMALLOC_MAX_SIZE));

		if (varflags) {
			ctor_values values;
			tree value;
			unsigned int j;

			values = autoslab_create_infos_initializer(gsi, build_int_cstu(long_unsigned_type_node, size), gfp_zones, lhs_type_name_ptr, lhs_align_val);
			FOR_EACH_CONSTRUCTOR_VALUE(values, j, value)
				CONSTRUCTOR_APPEND_ELT(initializer, build_int_cstu(unsigned_type_node, values->length() * (i - sa->minshift) + j), value);
		} else {
			tree value;

			value = autoslab_create_info(gsi, build_int_cstu(long_unsigned_type_node, size), flags, gfp_zones, "size_", lhs_type_name_ptr, lhs_align_val);
			value = build1_loc(loc, ADDR_EXPR, TYPE_POINTER_TO(autoslab_info_type), value);
			CONSTRUCTOR_APPEND_ELT(initializer, build_int_cstu(unsigned_type_node, i - sa->minshift), value);
		}
	}

	ctor = build_constructor(build_array_type_nelts(TYPE_POINTER_TO(autoslab_info_type), initializer->length()), initializer);
	autoslab_infos = autoslab_build_cache_ptr(loc, autoslab_infos_name, TREE_TYPE(ctor), ctor);

	return autoslab_infos;
}

static void __autoslab_convert_call(gimple_stmt_iterator *gsi,
				    enum kmem_call_names kmem_call,
				    tree size,
				    tree flags,
				    sbitmap gfp_zones,
				    tree node,
				    tree minshift,
				    tree maxshift,
				    const char *lhs_type_name_ptr,
				    unsigned HOST_WIDE_INT lhs_size_val,
				    unsigned HOST_WIDE_INT lhs_align_val,
				    tree kmem_cache)
{
	gcall *kalloc_call_stmt, *kmem_call_stmt;
	tree autoslab, kmem_cache_alloc;
	unsigned HOST_WIDE_INT size_val;
	location_t loc1, loc2;
	const char *autoslab_kind;
	cgraph_node_ptr callee;
	basic_block bb;
	int frequency;

	kalloc_call_stmt = as_a_gcall(gsi_stmt(*gsi));
	loc1 = gimple_inline_point_location(kalloc_call_stmt);
	loc2 = gimple_expansion_point_location(kalloc_call_stmt);
	size_val = TREE_CONSTANT(size) ? tree_to_uhwi(size) : 0;

	// handle the allocation of a single object
	if (DECL_P(kmem_cache))
		autoslab = create_tmp_var(TREE_TYPE(kmem_cache), DECL_NAME_POINTER(kmem_cache));
	else
		autoslab = create_tmp_var(TREE_TYPE(kmem_cache), DECL_NAME_POINTER(TREE_OPERAND(TREE_OPERAND(kmem_cache, 0), 0)));
	autoslab = make_ssa_name(autoslab, NULL);

	gimple_stmt_ptr assign = gimple_build_assign(autoslab, kmem_cache);
	gimple_set_location(assign, loc2);
	gimple_set_block(assign, gimple_block(kalloc_call_stmt));
	gsi_insert_before(gsi, assign, GSI_SAME_STMT);
	update_stmt(assign);

	kmem_cache_alloc = kmem_cache_decls[kmem_call].decl;
	switch (kmem_call) {
	default:
		gcc_unreachable();

	case KMEM_CACHE_ALLOC:
	case KMEM_CACHE_ALLOC_NOPROF:
	case KMEM_CACHE_ALLOC_TYPE:
	case KMEM_CACHE_ALLOC_NOPROF_TYPE:
		kmem_call_stmt = as_a_gcall(gimple_build_call(kmem_cache_alloc, 2, autoslab, flags));
		break;

	case KMEM_CACHE_ALLOC_INDEX:
	case KMEM_CACHE_ALLOC_NOPROF_INDEX:
	case KMEM_CACHE_ALLOC_INDEX_TYPE:
	case KMEM_CACHE_ALLOC_NOPROF_INDEX_TYPE:
		kmem_call_stmt = as_a_gcall(gimple_build_call(kmem_cache_alloc, 5, autoslab, size, flags, minshift, maxshift));
		break;

	case KMEM_CACHE_ALLOC_NODE:
	case KMEM_CACHE_ALLOC_NODE_NOPROF:
	case KMEM_CACHE_ALLOC_NODE_TYPE:
	case KMEM_CACHE_ALLOC_NODE_NOPROF_TYPE:
		kmem_call_stmt = as_a_gcall(gimple_build_call(kmem_cache_alloc, 3, autoslab, flags, node));
		break;

	case KMEM_CACHE_ALLOC_NODE_INDEX:
	case KMEM_CACHE_ALLOC_NODE_NOPROF_INDEX:
	case KMEM_CACHE_ALLOC_NODE_INDEX_TYPE:
	case KMEM_CACHE_ALLOC_NODE_NOPROF_INDEX_TYPE:
		kmem_call_stmt = as_a_gcall(gimple_build_call(kmem_cache_alloc, 6, autoslab, size, flags, node, minshift, maxshift));
		break;
	}

	gimple_call_copy_flags(kmem_call_stmt, kalloc_call_stmt);
	gimple_set_location(kmem_call_stmt, loc2);
	gimple_set_block(kmem_call_stmt, gimple_block(kalloc_call_stmt));
	gimple_call_set_lhs(kmem_call_stmt, gimple_call_lhs(kalloc_call_stmt));
//	update_stmt(kmem_call_stmt);
	gsi_replace(gsi, kmem_call_stmt, false);

	// update the cgraph
	bb = gsi_bb(*gsi);
	callee = cgraph_get_create_node(gimple_call_fndecl(kmem_call_stmt));
	gcc_assert(callee);
	frequency = compute_call_stmt_bb_frequency(current_function_decl, bb);
	cgraph_create_edge(cgraph_get_node(current_function_decl), callee, kmem_call_stmt, bb->count, frequency);

	if (!verbose)
		return;

	if (sbitmap_count_bits(gfp_zones) == 1)
		autoslab_kind = TREE_CODE(size) == INTEGER_CST ? "constsize" : "varsize";
	else
		autoslab_kind = TREE_CODE(size) == INTEGER_CST ? "varflags" : "varsizeflags";

	inform(loc2, "autoslab: yes %s "
		HOST_WIDE_INT_PRINT_UNSIGNED "/" HOST_WIDE_INT_PRINT_HEX_PURE " "
		HOST_WIDE_INT_PRINT_UNSIGNED "/" HOST_WIDE_INT_PRINT_HEX_PURE " "
		HOST_WIDE_INT_PRINT_UNSIGNED "/" HOST_WIDE_INT_PRINT_HEX_PURE " "
		HOST_WIDE_INT_PRINT_UNSIGNED "/" HOST_WIDE_INT_PRINT_HEX_PURE " "
		HOST_WIDE_INT_PRINT_HEX_PURE " "
		"%lx/%lx "
		HOST_WIDE_INT_PRINT_UNSIGNED "/" HOST_WIDE_INT_PRINT_HEX_PURE " "
		"%qs %u:%u %u:%u",
		autoslab_kind,
		lhs_size_val ? size_val/lhs_size_val : 0, lhs_size_val ? size_val/lhs_size_val : 0,
		size_val, size_val,
		lhs_size_val, lhs_size_val,
		lhs_align_val, lhs_align_val,
		TREE_CODE(flags) == INTEGER_CST ? tree_to_uhwi(flags) : ~0UL,
		0UL, 0UL, size_val, size_val, lhs_type_name_ptr,
		LOCATION_LINE(loc1), LOCATION_COLUMN(loc1),
		LOCATION_LINE(loc2), LOCATION_COLUMN(loc2));
}

static void autoslab_analyze_minus(tree rhs1, tree rhs2, struct size_analysis *sa_out)
{
	struct size_analysis sa_in_1, sa_in_2;

	__autoslab_analyze_size(rhs2, &sa_in_2);
	if (sa_in_2.c1)
		return;

	__autoslab_analyze_size(rhs1, &sa_in_1);
	if (sa_in_1.c0 < sa_in_2.c0)
		return;

	sa_out->c0 = sa_in_1.c0 - sa_in_2.c0;
	sa_out->c1 = sa_in_1.c1 - sa_in_2.c1;
}

static void autoslab_analyze_plus(tree rhs1, tree rhs2, struct size_analysis *sa_out)
{
	struct size_analysis sa_in_1, sa_in_2;

	__autoslab_analyze_size(rhs1, &sa_in_1);
	__autoslab_analyze_size(rhs2, &sa_in_2);

	sa_out->c0 = sa_in_1.c0 + sa_in_2.c0;
	if (!sa_in_1.c1 == !sa_in_2.c1)
		sa_out->c1 = 1;
	else
		sa_out->c1 = sa_in_1.c1 + sa_in_2.c1;
}

static void autoslab_analyze_mult(tree rhs1, tree rhs2, struct size_analysis *sa_out)
{
	struct size_analysis sa_in_1, sa_in_2;

	__autoslab_analyze_size(rhs1, &sa_in_1);
	__autoslab_analyze_size(rhs2, &sa_in_2);

	sa_out->c0 = sa_in_1.c0 * sa_in_2.c0;
	if (!sa_in_1.c1 == !sa_in_2.c1)
		sa_out->c1 = 1;
	else
		sa_out->c1 = sa_in_1.c0 * sa_in_2.c1 + sa_in_2.c0 * sa_in_1.c1;
}

static void __autoslab_analyze_size(tree size, struct size_analysis *sa_out)
{
	switch (TREE_CODE(size)) {
	default:
		debug_tree(size);
		gcc_unreachable();

	case ADDR_EXPR:
		sa_out->c0 = 0;
		sa_out->c1 = 0;
		break;

	case INTEGER_CST: {

		if (tree_fits_shwi_p(size)) {
			HOST_WIDE_INT c0_in = tree_to_shwi(size);

			sa_out->c0 = c0_in <= INT_MAX ? c0_in : 0;
		} else
			sa_out->c0 = 0;
		sa_out->c1 = 0;
		break;
	}

	case SSA_NAME: {
		gimple_stmt_ptr def_stmt = SSA_NAME_DEF_STMT(size);

		sa_out->c0 = 0;
		sa_out->c1 = 1;

		switch (gimple_code(def_stmt)) {
		default:
			inform(gimple_inline_point_location(def_stmt), "c1 " HOST_WIDE_INT_PRINT_DEC " c0 " HOST_WIDE_INT_PRINT_DEC, sa_out->c1, sa_out->c0);
			debug_gimple_stmt(def_stmt);
			debug_tree(size);
			gcc_unreachable();

		case GIMPLE_ASM:
			break;

		case GIMPLE_ASSIGN:
			switch (gimple_assign_rhs_code(def_stmt)) {
			default:
				inform(gimple_inline_point_location(def_stmt), "code %s", get_tree_code_name(gimple_assign_rhs_code(def_stmt)));
				inform(gimple_expansion_point_location(def_stmt), "code %s", get_tree_code_name(gimple_assign_rhs_code(def_stmt)));
				debug_gimple_stmt(def_stmt);
				debug_tree(size);
				gcc_unreachable();

			case ABS_EXPR:
			case ADDR_EXPR:
			case ARRAY_REF:
			case BIT_FIELD_REF:
			case BIT_NOT_EXPR:
			case BIT_XOR_EXPR:
			case COND_EXPR:
			case EQ_EXPR:
			case GE_EXPR:
			case GT_EXPR:
			case INDIRECT_REF:
			case LE_EXPR:
			case LROTATE_EXPR:
			case LT_EXPR:
			case MEM_REF:
			case MIN_EXPR:
			case NE_EXPR:
			case NEGATE_EXPR:
			case PARM_DECL:
#if BUILDING_GCC_VERSION >= 8000
			case POINTER_DIFF_EXPR:
#endif
			case RROTATE_EXPR:
			case TARGET_MEM_REF:
			case VAR_DECL:
				break;

			case COMPONENT_REF: {
				tree rhs, var, field, index, value;
				unsigned int i;

				rhs = gimple_assign_rhs1(def_stmt);
				var = TREE_OPERAND(rhs, 0);
				if (!VAR_P(var) || !DECL_INITIAL(var))
					break;

				field = TREE_OPERAND(rhs, 1);
				gcc_assert(TREE_CODE(field) == FIELD_DECL);
				if (!TREE_READONLY(var) && !TREE_READONLY(field))
					break;

				FOR_EACH_CONSTRUCTOR_ELT(CONSTRUCTOR_ELTS(DECL_INITIAL(var)), i, index, value) {
					if (field == index)
						return __autoslab_analyze_size(value, sa_out);
				}
				break;
			}

			case BIT_AND_EXPR: {
				tree rhs1, rhs2;
				union {
					HOST_WIDE_INT smask;
					unsigned HOST_WIDE_INT umask;
				} mask;
				struct size_analysis sa_in;

				rhs1 = gimple_assign_rhs1(def_stmt);
				rhs2 = gimple_assign_rhs2(def_stmt);

				if (TREE_CODE(rhs2) != INTEGER_CST)
					break;
				if (tree_fits_shwi_p(rhs2))
					mask.smask = tree_to_shwi(rhs2);
				else if (tree_fits_uhwi_p(rhs2))
					mask.umask = tree_to_uhwi(rhs2);
				else
					gcc_unreachable();
				if (!pow2p_hwi(mask.umask + 1))
					break;

				__autoslab_analyze_size(rhs1, &sa_in);

				sa_out->c0 = sa_in.c0 & mask.umask;
				break;
			}

			case BIT_IOR_EXPR: {
				tree rhs1, rhs2;
				union {
					HOST_WIDE_INT smask;
					unsigned HOST_WIDE_INT umask;
				} mask;
				struct size_analysis sa_in;

				rhs1 = gimple_assign_rhs1(def_stmt);
				rhs2 = gimple_assign_rhs2(def_stmt);

				if (TREE_CODE(rhs2) != INTEGER_CST)
					break;
				if (tree_fits_shwi_p(rhs2))
					mask.smask = tree_to_shwi(rhs2);
				else if (tree_fits_uhwi_p(rhs2))
					mask.umask = tree_to_uhwi(rhs2);
				else
					gcc_unreachable();
				if (!pow2p_hwi(mask.umask + 1))
					break;

				__autoslab_analyze_size(rhs1, &sa_in);

				sa_out->c0 = sa_in.c0 | mask.umask;

				break;
			}

			case LSHIFT_EXPR: {
				tree rhs1, rhs2;
				unsigned HOST_WIDE_INT lshift;
				struct size_analysis sa_in;

				rhs1 = gimple_assign_rhs1(def_stmt);
				rhs2 = gimple_assign_rhs2(def_stmt);

				if (TREE_CODE(rhs2) != INTEGER_CST)
					break;

				__autoslab_analyze_size(rhs1, &sa_in);

				lshift = tree_to_uhwi(rhs2);
				sa_out->c0 = sa_in.c0 << lshift;
				sa_out->c1 = sa_in.c1 << lshift;

				break;
			}

			case MAX_EXPR: {
				tree rhs1, rhs2;
				struct size_analysis sa_in_1, sa_in_2;

				rhs1 = gimple_assign_rhs1(def_stmt);
				rhs2 = gimple_assign_rhs2(def_stmt);

				__autoslab_analyze_size(rhs1, &sa_in_1);
				__autoslab_analyze_size(rhs2, &sa_in_2);

				if (sa_in_1.c1 && sa_in_2.c1)
					break;

				sa_out->c0 = MAX(sa_in_1.c0, sa_in_2.c0);
				sa_out->c1 = MAX(sa_in_1.c1, sa_in_2.c1);
				break;
			}

			case MINUS_EXPR:
				autoslab_analyze_minus(gimple_assign_rhs1(def_stmt), gimple_assign_rhs2(def_stmt), sa_out);
				break;

			case MULT_EXPR:
				autoslab_analyze_mult(gimple_assign_rhs1(def_stmt), gimple_assign_rhs2(def_stmt), sa_out);
				break;

			case PLUS_EXPR:
			case POINTER_PLUS_EXPR:
				autoslab_analyze_plus(gimple_assign_rhs1(def_stmt), gimple_assign_rhs2(def_stmt), sa_out);
				break;

			case EXACT_DIV_EXPR:
			case TRUNC_DIV_EXPR:
			case TRUNC_MOD_EXPR:
			case RSHIFT_EXPR:
				break;

			case REALPART_EXPR:
				__autoslab_analyze_size(TREE_OPERAND(gimple_assign_rhs1(def_stmt), 0), sa_out);
				break;

			case COMPLEX_EXPR:
			case CONVERT_EXPR:
			case INTEGER_CST:
			case NOP_EXPR:
			case SSA_NAME:
				__autoslab_analyze_size(gimple_assign_rhs1(def_stmt), sa_out);
				break;

			case VIEW_CONVERT_EXPR:
				__autoslab_analyze_size(TREE_OPERAND(gimple_assign_rhs1(def_stmt), 0), sa_out);
				break;
			}
			break;

		case GIMPLE_CALL: {
			if (!gimple_call_internal_p(def_stmt))
				break;

			switch (gimple_call_internal_fn(def_stmt)) {
			default:
				debug_gimple_stmt(def_stmt);
				debug_tree(size);
				gcc_unreachable();

			case IFN_UBSAN_CHECK_ADD:
				autoslab_analyze_plus(gimple_call_arg(def_stmt, 0), gimple_call_arg(def_stmt, 1), sa_out);
				break;

			case IFN_UBSAN_CHECK_SUB:
//				autoslab_analyze_minus(gimple_call_arg(def_stmt, 0), gimple_call_arg(def_stmt, 1), sa_out);
				break;

			case IFN_UBSAN_CHECK_MUL:
				autoslab_analyze_mult(gimple_call_arg(def_stmt, 0), gimple_call_arg(def_stmt, 1), sa_out);
				break;

			case IFN_ADD_OVERFLOW:
				autoslab_analyze_plus(gimple_call_arg(def_stmt, 0), gimple_call_arg(def_stmt, 1), sa_out);
				break;

			case IFN_SUB_OVERFLOW:
//				autoslab_analyze_minus(gimple_call_arg(def_stmt, 0), gimple_call_arg(def_stmt, 1), sa_out);
				break;

			case IFN_MUL_OVERFLOW:
				autoslab_analyze_mult(gimple_call_arg(def_stmt, 0), gimple_call_arg(def_stmt, 1), sa_out);
				break;
			}

			break;
		}

		case GIMPLE_NOP:
			break;

		case GIMPLE_PHI: {
			gphi *phi_stmt;
			unsigned int i, num_args;
			tree origin = NULL_TREE;

			phi_stmt = as_a_gphi(def_stmt);

			if (gimple_visited_p(phi_stmt))
				break;
			gimple_set_visited(phi_stmt, true);

			for (i = 0, num_args =  gimple_phi_num_args(phi_stmt); i < num_args; i++) {
				tree phi_arg = gimple_phi_arg_def(phi_stmt, i);

				switch (TREE_CODE(phi_arg)) {
				default:
					debug_gimple_stmt(def_stmt);
					debug_tree(phi_arg);
					debug_tree(size);
					gcc_unreachable();

				case ADDR_EXPR:
					continue;

				case COMPLEX_CST:
					phi_arg = TREE_REALPART(phi_arg);
					// FALLTHROUGH

				case INTEGER_CST:
					if (integer_zerop(phi_arg))
						continue;
					if (tree_int_cst_lt(KMALLOC_MAX_SIZE, phi_arg))
						continue;
					break;

				case SSA_NAME:
					if (origin == NULL_TREE) {
						origin = phi_arg;
						continue;
					}

					if (origin == phi_arg)
						continue;
					break;
				}

				break;
			}

			if (i == num_args && origin != NULL_TREE)
				__autoslab_analyze_size(origin, sa_out);
			gimple_set_visited(phi_stmt, false);

			break;
		}
		}
	}
	}
}

static sbitmap autoslab_all_zones(void)
{
	sbitmap gfp_zones;

	gfp_zones = sbitmap_alloc(tree_to_uhwi(__MAX_NR_ZONES));
	sbitmap_clear(gfp_zones);

	if (ZONE_DMA != NULL_TREE)
		sbitmap_set_bit(gfp_zones, tree_to_uhwi(ZONE_DMA));
	if (ZONE_DMA32 != NULL_TREE)
		sbitmap_set_bit(gfp_zones, tree_to_uhwi(ZONE_DMA32));
	sbitmap_set_bit(gfp_zones, tree_to_uhwi(ZONE_NORMAL));

	return gfp_zones;
}

static sbitmap autoslab_extract_flags(tree flags)
{
	sbitmap gfp_zones;
	unsigned HOST_WIDE_INT flags_val;

	flags_val = tree_to_uhwi(flags);

	gfp_zones = sbitmap_alloc(tree_to_uhwi(__MAX_NR_ZONES));
	sbitmap_clear(gfp_zones);

	if (flags_val & tree_to_uhwi(__GFP_DMA) && ZONE_DMA != NULL_TREE)
		sbitmap_set_bit(gfp_zones, tree_to_uhwi(ZONE_DMA));
	else if (flags_val & tree_to_uhwi(__GFP_DMA32) && ZONE_DMA32 != NULL_TREE)
		sbitmap_set_bit(gfp_zones, tree_to_uhwi(ZONE_DMA32));
	else
		sbitmap_set_bit(gfp_zones, tree_to_uhwi(ZONE_NORMAL));

	return gfp_zones;
}

static sbitmap autoslab_analyze_flags(tree flags, bool *atomic)
{
	switch (TREE_CODE(flags)) {
	default:
		debug_tree(flags);
		gcc_unreachable();

	case INTEGER_CST: {
		unsigned HOST_WIDE_INT flags_val;

		flags_val = tree_to_uhwi(flags);

		// bail out on __GFP_HIGHMEM/__GFP_MOVABLE
		if (flags_val & (tree_to_uhwi(__GFP_HIGHMEM) | tree_to_uhwi(__GFP_MOVABLE)))
			return NULL;
		// bail out on more than a single zone
		if (!!(flags_val & tree_to_uhwi(__GFP_DMA)) + !!(flags_val & tree_to_uhwi(__GFP_DMA32)) + !!(flags_val & tree_to_uhwi(__GFP_RECLAIMABLE)) > 1)
			return NULL;

		/*
		 * KSTACKOVERFLOW has no problem with these converted GFP_ATOMIC allocations,
		 * and we've disabled the informational message in convert_autostack()
		 * This commenting is only safe because AUTOSTACK depends on !VMAP_STACK
		 * and !PAX_PRIVATE_KSTACKS
		 */
		//if (!(flags_val & (tree_to_uhwi(__GFP_DIRECT_RECLAIM))))
		//	*atomic = true;

		return autoslab_extract_flags(flags);
	}

	case SSA_NAME: {
		gimple_stmt_ptr def_stmt;

		def_stmt = SSA_NAME_DEF_STMT(flags);

		switch (gimple_code(def_stmt)) {
		default:
			debug_gimple_stmt(def_stmt);
			debug_tree(flags);
			gcc_unreachable();

		case GIMPLE_ASSIGN: {
			sbitmap gfp_zones;
			tree rhs1;

			rhs1 = gimple_assign_rhs1(def_stmt);

			switch (gimple_assign_rhs_code(def_stmt)) {
			default:
				inform(gimple_expansion_point_location(def_stmt), "code %s", get_tree_code_name(gimple_assign_rhs_code(def_stmt)));
				debug_gimple_stmt(def_stmt);
				debug_tree(flags);
				gcc_unreachable();

			case BIT_AND_EXPR: {
				tree rhs2;
				sbitmap rhs_gfp_zones;

				// shortcut const masks
				gcc_assert(TREE_CODE(rhs1) != INTEGER_CST);
				rhs2 = gimple_assign_rhs2(def_stmt);
				if (TREE_CODE(rhs2) == INTEGER_CST)
					return autoslab_extract_flags(rhs2);

				gfp_zones = autoslab_analyze_flags(rhs2, atomic);
				if (!gfp_zones)
					return NULL;

				rhs_gfp_zones = autoslab_analyze_flags(rhs1, atomic);
				if (!rhs_gfp_zones) {
					sbitmap_free(gfp_zones);
					return NULL;
				}
				sbitmap_and(gfp_zones, gfp_zones, rhs_gfp_zones);
				sbitmap_free(rhs_gfp_zones);

				return gfp_zones;
			}

			case COMPONENT_REF:
			case MEM_REF:
			case RSHIFT_EXPR:
				// in IPA/LTO we can refine this
				return autoslab_all_zones();

			case COND_EXPR:
			case BIT_IOR_EXPR: {
				sbitmap rhs_gfp_zones;

				gfp_zones = autoslab_analyze_flags(gimple_assign_rhs2(def_stmt), atomic);
				if (!gfp_zones)
					return NULL;

				rhs_gfp_zones = autoslab_analyze_flags(gimple_assign_rhs_code(def_stmt) == BIT_IOR_EXPR ? rhs1 : gimple_assign_rhs3(def_stmt), atomic);
				if (!rhs_gfp_zones) {
					sbitmap_free(gfp_zones);
					return NULL;
				}
				sbitmap_ior(gfp_zones, gfp_zones, rhs_gfp_zones);
				sbitmap_free(rhs_gfp_zones);

				return gfp_zones;
			}

			case EQ_EXPR:
			case GE_EXPR:
			case GT_EXPR:
			case LE_EXPR:
			case LT_EXPR:
			case NE_EXPR:
				return autoslab_extract_flags(integer_one_node);

			case MULT_EXPR: {
				tree rhs2;
				unsigned HOST_WIDE_INT rhs2_val;

				rhs2 = gimple_assign_rhs2(def_stmt);
				if (TREE_CODE(rhs2) != INTEGER_CST)
					return autoslab_all_zones();
				rhs2_val = tree_to_uhwi(rhs2);
				if (!pow2p_hwi(rhs2_val))
					return autoslab_all_zones();
				if (rhs2_val >= tree_to_uhwi(__MAX_NR_ZONES))
					return autoslab_all_zones();

				// TODO: if we ever run into this
				return autoslab_all_zones();
			}

#if 0
			case PLUS_EXPR: {
				sbitmap rhs_gfp_zones;

				gfp_zones = autoslab_analyze_flags(gimple_assign_rhs2(def_stmt), atomic);
				if (!gfp_zones)
					return NULL;

				rhs_gfp_zones = autoslab_analyze_flags(rhs1, atomic);
				if (!rhs_gfp_zones) {
					sbitmap_free(gfp_zones);
					return NULL;
				}
				sbitmap_ior(gfp_zones, gfp_zones, rhs_gfp_zones);
				sbitmap_free(rhs_gfp_zones);

				return gfp_zones;
			}
#endif

			case NOP_EXPR:
			case VAR_DECL:
				return autoslab_analyze_flags(rhs1, atomic);
			}
		}

		case GIMPLE_CALL:
			// in IPA/LTO we can refine this
			return autoslab_all_zones();

		case GIMPLE_NOP: {
			sbitmap gfp_zones;

			// in IPA/LTO we can refine this
			gfp_zones = autoslab_all_zones();

#if BUILDING_GCC_VERSION >= 7000
			tree _value;
			widest_int _mask;
			unsigned HOST_WIDE_INT value, mask;
			location_t loc = DECL_SOURCE_LOCATION(current_function_decl);

			gcc_assert(SSA_NAME_VAR(flags));
			flags = SSA_NAME_VAR(flags);
			gcc_assert(TREE_CODE(flags) == PARM_DECL);

			if (!ipcp_get_parm_bits(flags, &_value, &_mask))
				return gfp_zones;

			value = tree_to_uhwi(_value);
			mask = _mask.to_uhwi();

			if (ZONE_DMA != NULL_TREE && !(mask & tree_to_uhwi(__GFP_DMA)) && !(value & tree_to_uhwi(__GFP_DMA)))
				sbitmap_clear_bit(gfp_zones, tree_to_uhwi(ZONE_DMA));
			if (ZONE_DMA32 != NULL_TREE && !(mask & tree_to_uhwi(__GFP_DMA32)) && !(value & tree_to_uhwi(__GFP_DMA32)))
				sbitmap_clear_bit(gfp_zones, tree_to_uhwi(ZONE_DMA32));

			if (verbose)
				inform(loc, "autoslab: FLAGS %qD VALUE " HOST_WIDE_INT_PRINT_HEX_PURE " MASK " HOST_WIDE_INT_PRINT_HEX_PURE, flags, value, mask);
#endif

			return gfp_zones;
		}

		case GIMPLE_PHI: {
			sbitmap phi_gfp_zones;
			gphi *phi_stmt;
			unsigned int i, num_args;

			phi_gfp_zones = sbitmap_alloc(tree_to_uhwi(__MAX_NR_ZONES));
			sbitmap_clear(phi_gfp_zones);

			if (gimple_visited_p(def_stmt))
				return phi_gfp_zones;
			gimple_set_visited(def_stmt, true);

			phi_stmt = as_a_gphi(def_stmt);

			for (i = 0, num_args = gimple_phi_num_args(phi_stmt); i < num_args; i++) {
				sbitmap arg_gfp_zones;

				arg_gfp_zones = autoslab_analyze_flags(gimple_phi_arg_def(phi_stmt, i), atomic);
				if (!arg_gfp_zones) {
					gimple_set_visited(def_stmt, false);
					sbitmap_free(phi_gfp_zones);
					return NULL;
				}

				sbitmap_ior(phi_gfp_zones, phi_gfp_zones, arg_gfp_zones);
				sbitmap_free(arg_gfp_zones);
			}

			gimple_set_visited(def_stmt, false);
			return phi_gfp_zones;
		}
		}
	}

	case VAR_DECL:
		return autoslab_all_zones();
	}
}

/*
 * convert a k*alloc*(size, flags, ...) call into kmem_cache_*alloc*(cache, flags, ...)
 *
 * in particular:
 * 1. kmalloc            -> kmem_cache_alloc
 * 2. kmalloc_node       -> kmem_cache_alloc_node
 * 3. kzalloc            -> kmem_cache_alloc
 * 4. kzalloc_node       -> kmem_cache_alloc_node
 * 5. kcalloc            -> kmem_cache_alloc
 * 6. kmalloc_array      -> kmem_cache_alloc
 * 7. kcalloc_node       -> kmem_cache_alloc_node
 * 8. kmalloc_array_node -> kmem_cache_alloc_node
 */
static void autoslab_convert_call(gimple_stmt_iterator *gsi,
				  unsigned int kalloc_func,
				  tree count_arg,
				  tree size_arg,
				  tree flags_arg,
				  tree node_arg)
{
	gcall *kalloc_call_stmt;
	enum kmem_call_names kmem_call;
	tree lhs, lhs_type, size, flags, node, lhs_size, maxsize;
	const_tree lhs_type_name;
	unsigned HOST_WIDE_INT size_val, lhs_size_val, lhs_align_val;
	location_t loc;
	char *lhs_type_name_ptr = NULL;
	sbitmap gfp_zones;
	bool atomic;
	unsigned int align;
	struct size_analysis sa = {};

	gcc_assert(count_arg == NULL_TREE);

	kalloc_call_stmt = as_a_gcall(gsi_stmt(*gsi));
	loc = gimple_expansion_point_location(kalloc_call_stmt);

	lhs_type_name = lhs_type = NULL_TREE;
	if (kalloc_func == KMALLOC_TYPENAME) {
		tree cache_typename;

		cache_typename = gimple_call_arg(kalloc_call_stmt, 2);
		gcc_assert(TREE_CODE(cache_typename) == ADDR_EXPR);
		cache_typename = TREE_OPERAND(cache_typename, 0);
		gcc_assert(TREE_CODE(cache_typename) == STRING_CST);
		cache_typename = TREE_TYPE(cache_typename);
		if (POINTER_TYPE_P(cache_typename)) {
			cache_typename = TREE_TYPE(cache_typename);
			gcc_assert(RECORD_OR_UNION_TYPE_P(cache_typename));

			lhs_type = cache_typename;
			lhs_type_name = type_name(lhs_type);
			gcc_assert(lhs_type_name);
		} else
			gcc_assert(TREE_CODE(cache_typename) == ARRAY_TYPE);
	}

	lhs = gimple_call_lhs(kalloc_call_stmt);
	if (lhs_type == NULL_TREE) {
		tree alloc_type;

		lhs_type = alloc_type = autoslab_get_alloc_type(lhs) ? : TREE_TYPE(TREE_TYPE(lhs));

		if (POINTER_TYPE_P(alloc_type))
			alloc_type = TREE_TYPE(alloc_type);
		if (POINTER_TYPE_P(alloc_type))
			alloc_type = TREE_TYPE(alloc_type);
		if (RECORD_OR_UNION_TYPE_P(alloc_type)) {
			lhs_type_name = type_name(TYPE_MAIN_VARIANT(alloc_type));
			if (lhs_type_name == NULL_TREE)
				lhs_type_name = type_name(alloc_type);
			if (verbose && lhs_type_name == NULL_TREE) {
				inform(loc, "autoslab: unnamed lhs type %qT", alloc_type);
			}
		}
	}

	size = gimple_call_arg(kalloc_call_stmt, tree_to_uhwi(size_arg) - 1);
	flags = gimple_call_arg(kalloc_call_stmt, tree_to_uhwi(flags_arg) - 1);
	node = node_arg ? gimple_call_arg(kalloc_call_stmt, tree_to_uhwi(node_arg) - 1) : NULL_TREE;

	if (lhs_type_name == NULL_TREE)
		lhs_type_name_ptr = xstrdup("");
	else if (POINTER_TYPE_P(lhs_type))
		gcc_assert(asprintf(&lhs_type_name_ptr, "%s%s", POINTER_TYPE_P(TREE_TYPE(lhs_type)) ? "**" : "*", IDENTIFIER_POINTER(lhs_type_name)) != -1);
	else {
		gcc_assert(asprintf(&lhs_type_name_ptr, "%s", IDENTIFIER_POINTER(lhs_type_name)) != -1);
	}

	lhs_size = TYPE_SIZE_UNIT(lhs_type);
	if (lhs_size != NULL_TREE) {
		gcc_assert(TREE_CODE(lhs_size) == INTEGER_CST);
		lhs_size_val = tree_to_uhwi(lhs_size);
	} else
		lhs_size_val = 0;

	atomic = false;
	gfp_zones = autoslab_analyze_flags(flags, &atomic);

	if (!gfp_zones) {
		error_at(loc, "autoslab: no badflags size: %qE lhssize:%qE flags:%qE lhstype:%qs",
			 size, lhs_size ?: null_pointer_node, flags, lhs_type_name_ptr);
		goto out;
	}

	gcc_assert(sbitmap_count_bits(gfp_zones) > 0);

	maxsize = autoslab_analyze_size(gsi, size, &sa);

	// we managed to reduce varsize to constsize
	if (sa.minshift == sa.maxshift)
		size = maxsize;
	size_val = tree_to_uhwi(maxsize);
	if (lhs_size_val == 0)
		lhs_size_val = size_val;

	// handle the special sizes that return a const ptr
	if (integer_zerop(size)) {
		// convert 0-sized allocations to ZERO_SIZE_PTR
		gcc_assert(!SSA_NAME_OCCURS_IN_ABNORMAL_PHI(lhs));
		replace_call_with_value(gsi, ZERO_SIZE_PTR);

		if (verbose)
			inform(loc, "autoslab: yes nullsize %qE %qs", flags, lhs_type_name_ptr);
		goto out;
	}

	if (TREE_CODE(size) == INTEGER_CST && tree_int_cst_lt(KMALLOC_MAX_SIZE, size)) {
		// convert oversized allocations to NULL
		gcc_assert(!SSA_NAME_OCCURS_IN_ABNORMAL_PHI(lhs));
		replace_call_with_value(gsi, null_pointer_node);

		if (verbose)
			inform(loc, "autoslab: yes maxsize "
				HOST_WIDE_INT_PRINT_UNSIGNED "/" HOST_WIDE_INT_PRINT_HEX_PURE
				" %qE %qs",
				size_val, size_val, flags, lhs_type_name_ptr);
		goto out;
	}

	// compute the required alignment
	lhs_align_val = MAX(TYPE_ALIGN_UNIT(lhs_type), tree_to_uhwi(ARCH_KMALLOC_MINALIGN));
	align = get_pointer_alignment(lhs) / BITS_PER_UNIT;
	lhs_align_val = MAX(lhs_align_val, align);

	// try to convert to an automatic variable first
	if (can_convert_to_automatic(gsi, maxsize, gfp_zones, atomic, lhs_align_val)) {
		autoslab_convert_to_automatic(gsi, maxsize, flags, gfp_zones, lhs_size_val, lhs_align_val);

		if (verbose)
			inform(loc, "autoslab: yes stack "
				    HOST_WIDE_INT_PRINT_UNSIGNED "/" HOST_WIDE_INT_PRINT_HEX_PURE " "
				    HOST_WIDE_INT_PRINT_UNSIGNED "/" HOST_WIDE_INT_PRINT_HEX_PURE " "
				    "%qE %s " HOST_WIDE_INT_PRINT_UNSIGNED " %u %qT",
				lhs_size_val, lhs_size_val,
				size_val, size_val,
				flags, node ? "node" : "nonode",
				lhs_align_val,
				TYPE_ALIGN_UNIT(TREE_TYPE(TREE_TYPE(lhs))),
				TYPE_MAIN_VARIANT(TREE_TYPE(TREE_TYPE(lhs))));
	}

	if (TREE_CODE(size) == INTEGER_CST) {
		align = MIN(1U << ceil_log2(size_val), tree_to_uhwi(SMP_CACHE_BYTES));
		align = MIN(1U << ctz_hwi(size_val), align);
		lhs_align_val = MAX(lhs_align_val, align);
	} else if (sa.c0 > lhs_align_val) {
		align = MIN(1U << ceil_log2(sa.c0), tree_to_uhwi(SMP_CACHE_BYTES));
		align = MIN(1U << ctz_hwi(sa.c0), align);
		lhs_align_val = MAX(lhs_align_val, align);
	}

	if (TREE_CODE(size) != INTEGER_CST) {
		tree kmem_caches;

		if (sbitmap_count_bits(gfp_zones) == 1)
			// build varsize:
			// ptr = kmem_cache_alloc_index(&autoslabs, size, flags, sa.minshift, sa.maxshift);
			kmem_call = kalloc_functions[kalloc_func].kmem_name_index;
		else
			// build varsizeflags:
			// ptr = kmem_cache_alloc_index_type(&autoslabs, size, flags, sa.minshift, sa.maxshift);
			kmem_call = kalloc_functions[kalloc_func].kmem_name_index_type;

		if (kmem_cache_decls[kmem_call].decl == NULL_TREE) {
			error_at(loc, "autoslab did not find %qs to replace %qD", kmem_cache_decls[kmem_call].name, gimple_call_fndecl(kalloc_call_stmt));
			goto out;
		}

		gcc_assert(sa.minshift < sa.maxshift);
		kmem_caches = autoslab_create_infos_size(gsi, flags, gfp_zones, lhs_type_name_ptr, &sa, size_val, lhs_align_val);
		kmem_caches = build1_loc(loc, ADDR_EXPR, TYPE_POINTER_TO(autoslab_infos_type), kmem_caches);
		kmem_caches = fold_convert_loc(loc, TYPE_POINTER_TO(TYPE_POINTER_TO(autoslab_info_type)), kmem_caches);
		__autoslab_convert_call(gsi, kmem_call, size, flags, gfp_zones, node, build_int_cstu(unsigned_type_node, sa.minshift), build_int_cstu(unsigned_type_node, sa.maxshift), lhs_type_name_ptr, lhs_size_val, lhs_align_val, kmem_caches);
	} else if (sbitmap_count_bits(gfp_zones) == 1) {
		tree autoslab_cache_name, kmem_cache, autoslab_info, ctor;
		varpool_node_ptr var;
		unsigned HOST_WIDE_INT flags_val;

		flags_val = TREE_CODE(flags) == INTEGER_CST ? tree_to_uhwi(flags) : 0;

		// build constsize:
		// ptr = kmem_cache_alloc(autoslab, flags);
		kmem_call = kalloc_functions[kalloc_func].kmem_name;
		if (kmem_cache_decls[kmem_call].decl == NULL_TREE) {
			error_at(loc, "autoslab did not find %qs to replace %qD", kmem_cache_decls[kmem_call].name, gimple_call_fndecl(kalloc_call_stmt));
			goto out;
		}

		autoslab_cache_name = autoslab_gen_name("cache_", "", gsi, size_val, lhs_align_val, flags_val, gfp_zones, 0, 0, 0, 0);
		gcc_assert(autoslab_cache_name);
		var = varpool_get_for_asmname(autoslab_cache_name);
		gcc_assert(!var);

		autoslab_info = autoslab_create_info(gsi, size, flags, gfp_zones, "const_", lhs_type_name_ptr, lhs_align_val);
		gcc_assert(TYPE_POINTER_TO(autoslab_info_type));
		gcc_assert(POINTER_TYPE_P(TYPE_POINTER_TO(autoslab_info_type)));
		ctor = build1_loc(loc, ADDR_EXPR, TYPE_POINTER_TO(autoslab_info_type), autoslab_info);
		kmem_cache = autoslab_build_cache_ptr(loc, autoslab_cache_name, TREE_VALUE(TYPE_ARG_TYPES(TREE_TYPE(kmem_cache_decls[kmem_call].decl))), ctor);
		__autoslab_convert_call(gsi, kmem_call, size, flags, gfp_zones, node, NULL_TREE, NULL_TREE, lhs_type_name_ptr, lhs_size_val, lhs_align_val, kmem_cache);
	} else {
		tree kmem_caches;

		gcc_assert(TREE_CODE(flags) == SSA_NAME);

		// build varflags:
		// ptr = kmem_cache_alloc_type(&autoslabs, flags);
		kmem_call = kalloc_functions[kalloc_func].kmem_name_type;
		if (kmem_cache_decls[kmem_call].decl == NULL_TREE) {
			error_at(loc, "autoslab did not find %qs to replace %qD", kmem_cache_decls[kmem_call].name, gimple_call_fndecl(kalloc_call_stmt));
			goto out;
		}

		kmem_caches = autoslab_create_infos_type(gsi, size, gfp_zones, lhs_type_name_ptr, lhs_align_val);
		kmem_caches = build1_loc(loc, ADDR_EXPR, TYPE_POINTER_TO(autoslab_infos_type), kmem_caches);
		kmem_caches = fold_convert_loc(loc, TYPE_POINTER_TO(TYPE_POINTER_TO(autoslab_info_type)), kmem_caches);
		__autoslab_convert_call(gsi, kmem_call, size, flags, gfp_zones, node, NULL_TREE, NULL_TREE, lhs_type_name_ptr, lhs_size_val, lhs_align_val, kmem_caches);
	}

out:
	free(lhs_type_name_ptr);
	sbitmap_free(gfp_zones);
}

static unsigned int autoslab_execute(void)
{
	basic_block bb;

	loop_optimizer_init(LOOPS_NORMAL | LOOPS_HAVE_RECORDED_EXITS);
	gcc_assert(current_loops);

	calculate_dominance_info(CDI_DOMINATORS);
	calculate_dominance_info(CDI_POST_DOMINATORS);

	gcc_assert(dom_info_available_p(CDI_DOMINATORS));
	gcc_assert(dom_info_available_p(CDI_POST_DOMINATORS));

	scev_initialize();

#if BUILDING_GCC_VERSION >= 12000
	enable_ranger(cfun);
#endif

	find_unreachable_blocks();

#if BUILDING_GCC_VERSION >= 8000 && BUILDING_GCC_VERSION < 9000
	if (!ipa_call_summaries)
		ipa_call_summaries = new ipa_call_summary_t(symtab, false);
#endif

	FOR_EACH_BB_FN(bb, cfun) {
		gphi_iterator gsi;

		for (gsi = gsi_start_phis(bb); !gsi_end_p(gsi); gsi_next(&gsi))
			gimple_set_visited(gsi_gphi(gsi), false);
	}

	// 1. loop through BBs and GIMPLE statements
	FOR_EACH_BB_FN(bb, cfun) {
		gimple_stmt_iterator gsi;

		for (gsi = gsi_start_bb(bb); !gsi_end_p(gsi);) {
			tree fndecl, count_arg, size_arg, flags_arg, node_arg;
			gimple_stmt_ptr stmt;
			gcall *call_stmt;
			unsigned int kalloc_func;

			// 2. is it a call ...
			stmt = gsi_stmt(gsi);
			if (!is_gimple_call(stmt))
				goto next;
			call_stmt = as_a_gcall(stmt);
			fndecl = gimple_call_fndecl(call_stmt);
			if (fndecl == NULL_TREE)
				goto next;
			if (TREE_CODE(fndecl) != FUNCTION_DECL)
				goto next;

			// 3. ... to k*alloc
			kalloc_func = is_kalloc(DECL_NAME_POINTER(fndecl));
			if (kalloc_func >= ARRAY_SIZE(kalloc_functions))
				goto next;

			// 4. remove stray kmalloc calls left behind by certain transformations
			if (gimple_call_lhs(call_stmt) == NULL_TREE) {
				unlink_stmt_vdef(call_stmt);
				gsi_remove(&gsi, true);
				continue;
			}

			// 5. the alloc/malloc/alloc_size/alloc_flags attrs must be present on the decl/type
			gcc_assert(lookup_attribute("alloc", DECL_ATTRIBUTES(fndecl)));
			gcc_assert(lookup_attribute("malloc", DECL_ATTRIBUTES(fndecl)));
			autoslab_get_alloc_attrs(fndecl, &count_arg, &size_arg, &flags_arg, NULL, &node_arg);
			gcc_assert(flags_arg);

			// 6. try to convert the allocation into a type specific slab allocation
			autoslab_convert_call(&gsi, kalloc_func, count_arg, size_arg, flags_arg, node_arg);
next:			gsi_next(&gsi);
		}
	}

#if BUILDING_GCC_VERSION >= 12000
	disable_ranger(cfun);
#endif

	scev_finalize();

	free_dominance_info(CDI_POST_DOMINATORS);
	free_dominance_info(CDI_DOMINATORS);
	loop_optimizer_finalize();

	return 0;
}

#define PASS_NAME autoslab
#define NO_GATE
#define PROPERTIES_REQUIRED PROP_cfg
#define TODO_FLAGS_START TODO_rebuild_cgraph_edges
#define TODO_FLAGS_FINISH TODO_verify_il | TODO_remove_unused_locals | TODO_update_ssa | TODO_cleanup_cfg | TODO_rebuild_cgraph_edges | TODO_update_address_taken
#include "gcc-generate-gimple-pass.h"
