/*
 * Copyright 2012-2024 by PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 *
 * Homepage: http://pax.grsecurity.net/
 *
 * Usage:
 * $ # for C++ based 5+
 * $ g++ -I`g++ -print-file-name=plugin`/include -I`g++ -print-file-name=plugin`/include/c-family -fPIC -shared -O2 -o rap_plugin.so rap_plugin.c
 * $ gcc -fplugin=./rap_plugin.so -fplugin-arg-rap_plugin-check=call test.c -O2
 *
 * This plugin implements the Reuse Attack Protector (RAP) Control Flow Integrity (CFI) scheme
 * along with a poor man's emulation of non-executable kernel memory on amd64 (KERNEXEC) and
 * a defense mechanism against Spectre v2 (CVE-2017-5715, branch target injection). Since all
 * of these defenses instrument around the same code, there's an ordering requirement and thus
 * the need to implement them together. That said, they can all be individually enabled during
 * compilation.
 *
 * The instrumented code under various configurations looks like this:
 *
 * Indirect calls (forward edge), including tail calls (implemented as indirect jumps)
 *
 *                           all inlined                thunked retpoline                  no retpoline
 *    ---------------------------------------------------------------------------------------------------------
 *    load fptr              mov (fptr), %rax           mov (fptr), %rax                   mov (fptr), %rax
 *    ---------------------------------------------------------------------------------------------------------
 *    kernexec               bts $63, %rax              bts $63, %rax                      bts $63, %rax
 *    ---------------------------------------------------------------------------------------------------------
 *    rap fptr check         cmpq $hash, -8(%rax)       cmpq $hash, -8(%rax)               cmpq $hash, -8(%rax)
 *                           jnz .bad                   jnz .bad                           jnz .bad
 *    ---------------------------------------------------------------------------------------------------------
 *    retpoline              jmp 1f
 *                        2: call 3f
 *                        4: pause
 *                           lfence
 *                           jmp 4b
 *                        3: mov %rax, (%rsp)
 *                           ret
 *    ---------------------------------------------------------------------------------------------------------
 *                                                      jmp 1f                             jmp 1f
 *    rap return check       movabs -$hash, %rax        movabs -$hash, %rax                movabs -$hash, %rax
 *                           .fill padding              .fill padding                      .fill padding
 *    ---------------------------------------------------------------------------------------------------------
 *    retpoline           1: call 2b                 1: call __x86_indirect_thunk_rax
 *    ---------------------------------------------------------------------------------------------------------
 *    original call                                                                     1: call *%rax
 *
 */

#include "rap.h"

__visible int plugin_is_GPL_compatible;

static struct plugin_info rap_plugin_info = {
	.version	= "202403312200",
	.help		= "typecheck=ret,call,nospec,cast\tenable the corresponding type hash checking based features\n"
			  "callabort=ud2\t\toverride __builtin_trap with specified asm for indirect call checking\n"
			  "retabort=ud2\t\toverride __builtin_trap with specified asm for type hash based return address checking\n"
			  "xorabort=ud2\t\toverride __builtin_trap with specified asm for xor cookie based return address checking\n"
			  "hash=abs,abs-finish,abs-ops,abs-attr,const,volatile\n"
			  "report=func,fptr,abs\n"
			  "include\t\tassume that asm macros are included into all translation units to provide various RAP macros instead of the builtin ones\n"
			  "kernexec_method=[bts|or]\tKERNEXEC instrumentation method\n"
			  "sls\tSLS instrumentation\n"
			  "rethunk\tSLS compatible -mfunction-return=thunk-extern instrumentation\n"
};

rap_hash_flags_t imprecise_rap_hash_flags = {
	.qual_const	= 1,
	.qual_volatile	= 1,
	.func_name	= 0,
	.func_scope	= 0,
};

rap_hash_flags_t precise_rap_hash_flags = {
	.qual_const	= 1,
	.qual_volatile	= 1,
	.func_name	= 1,
	.func_scope	= 1,
};

tree rap_cookie_decl;
#ifdef TARGET_386
rtx leaveq_rtx, pop_rbp_rtx;
#endif
unsigned int rap_reg_id;
const char *rap_reg_name;
const char *asm_call_insn_pattern;
regex_t asm_call_insn_regex;
const char *rap_call_abort;
const char *rap_ret_abort;
const char *rap_xor_abort;
bool rap_include;

bool enable_xor = false;
bool enable_type_ret = false;
bool enable_type_call = false;
bool enable_type_nospec = false;
bool enable_type_cast = false;
bool enable_abs_finish = false;
bool enable_abs_ops = false;
bool enable_abs_attr = false;
static int rethooks;

bool enable_sls = false;
static bool enable_rethunk = false;

static bool report_func_hash, report_abs_hash;

gimple_stmt_ptr ibarrier(tree var)
{
	gimple_stmt_ptr stmt;
	gasm *asm_stmt;
	vec<tree, va_gc> *inputs = NULL;
	vec<tree, va_gc> *outputs = NULL;

	tree input, output;

	input = build_tree_list(NULL_TREE, build_const_char_string(2, "0"));
	input = chainon(NULL_TREE, build_tree_list(input, var));
	VEC_safe_push(tree, gc, inputs, input);

	output = build_tree_list(NULL_TREE, build_const_char_string(4, "=rm"));
	if (SSA_NAME_VAR(var))
		var = make_ssa_name(SSA_NAME_VAR(var), NULL);
	else
		var = make_temp_ssa_name(TREE_TYPE(var), NULL, "rap_ibarrier");
	output = chainon(NULL_TREE, build_tree_list(output, var));
	VEC_safe_push(tree, gc, outputs, output);

	stmt = gimple_build_asm_vec("lfence", inputs, outputs, NULL, NULL);
	asm_stmt = as_a_gasm(stmt);
	gimple_asm_set_volatile(asm_stmt, true);
	SSA_NAME_DEF_STMT(var) = stmt;

	return stmt;
}

static void check_global_variable_casts(void *event_data __unused, void *data __unused)
{
	varpool_node_ptr vnode;

	FOR_EACH_VARIABLE(vnode)
		check_variable_casts(NODE_DECL(vnode));
}

static void check_global_register_variables(void *event_data __unused, void *data __unused)
{
	varpool_node_ptr vnode;

	FOR_EACH_VARIABLE(vnode)
		check_register_variable(NODE_DECL(vnode));
}

static const struct gcc_debug_hooks *old_debug_hooks;
static struct gcc_debug_hooks rap_debug_hooks;

static bool rap_fptr_var_is_dereferenced(cgraph_node_ptr node, struct ipa_ref *ref, varpool_node_ptr varnode)
{
	tree vardecl, vartype;

	vardecl = NODE_DECL(varnode);
	gcc_assert(VAR_P(vardecl));
	vartype = TREE_TYPE(vardecl);

	switch (TREE_CODE(vartype)) {
	default:
		debug_cgraph_node(node);
		dump_varpool_node(stderr, varnode);
		debug_gimple_stmt(ref->stmt);
		debug_tree(vardecl);
		debug_tree(vartype);
		gcc_unreachable();

	case ARRAY_TYPE:
		vartype = TREE_TYPE(vartype);
		// FALLTHROUGH
	case POINTER_TYPE:
		return lookup_attribute("noderef", TYPE_ATTRIBUTES(vartype)) == NULL_TREE;

	case INTEGER_TYPE:
		// TODO: triggers on linux-6.8.1/samples/ftrace/ftrace-direct-modify.c:my_ip=schedule
		return false;

	case RECORD_TYPE:
	case UNION_TYPE: {
		tree index, value;
		unsigned int i;

		if (DECL_INITIAL(vardecl) == NULL_TREE)
			return true;

		if (DECL_INITIAL(vardecl) == error_mark_node)
			return true;

		FOR_EACH_CONSTRUCTOR_ELT(CONSTRUCTOR_ELTS(DECL_INITIAL(vardecl)), i, index, value) {
			gcc_assert(TREE_CODE(index) == FIELD_DECL);

			switch (TREE_CODE(value)) {
			default:
				debug_cgraph_node(node);
				dump_varpool_node(stderr, varnode);
				debug_gimple_stmt(ref->stmt);
				debug_tree(vardecl);
				debug_tree(vartype);
				debug_tree(value);
				gcc_unreachable();

			case ADDR_EXPR:
				if (!operand_equal_p(TREE_OPERAND(value, 0), NODE_DECL(node), 0))
					continue;
				return lookup_attribute("noderef", TYPE_ATTRIBUTES(TREE_TYPE(index))) == NULL_TREE;

			case CONSTRUCTOR:
				return true;

			case INTEGER_CST:
			case STRING_CST:
				continue;
			}
		}
		return false;
	}
	}
}

static bool rap_fptr_arg_is_dereferenced(cgraph_node_ptr node, struct ipa_ref *ref)
{
	gimple_stmt_ptr ref_stmt;

	ref_stmt = ref->stmt;

	gcc_assert(TREE_CODE(NODE_DECL(ipa_ref_referring_cgraph_node(ref))) == FUNCTION_DECL);

	switch (gimple_code(ref->stmt)) {
	default:
		debug_gimple_stmt(ref_stmt);
		gcc_unreachable();

	case GIMPLE_ASSIGN:
		return true;

	case GIMPLE_ASM:
		return true;

	case GIMPLE_CALL: {
		unsigned int i, e;
		tree argtype;
		unsigned long found = 0;

		argtype = TYPE_ARG_TYPES(gimple_call_fntype(ref_stmt));
		for (i = 0, e = gimple_call_num_args(ref_stmt); i < e; ++i, argtype = TREE_CHAIN(argtype)) {
			tree arg;

			if (argtype == NULL_TREE)
				return true;

			arg = gimple_call_arg(ref_stmt, i);
			if (TREE_CODE(arg) != ADDR_EXPR)
				continue;
			if (!operand_equal_p(TREE_OPERAND(arg, 0), NODE_DECL(node), 0))
				continue;
			found++;
			if (!lookup_attribute("noderef", TYPE_ATTRIBUTES(TREE_VALUE(argtype))))
				return true;
		}

		gcc_assert(found);
		return false;
	}

	case GIMPLE_COND:
		// an fptr comparison is not a dereference
		return false;

	case GIMPLE_PHI:
		return true;

	case GIMPLE_RETURN:
		return true;
	}
}

static bool rap_fptr_is_dereferenced(cgraph_node_ptr node, struct ipa_ref *ref)
{
	cgraph_node_ptr caller;
	varpool_node_ptr varnode;

	varnode = ipa_ref_referring_varpool_node(ref);
	if (varnode)
		return rap_fptr_var_is_dereferenced(node, ref, varnode);

	caller = ipa_ref_referring_cgraph_node(ref);
	gcc_assert(caller);
	return ref->stmt ? rap_fptr_arg_is_dereferenced(node, ref) : true;
}

static bool rap_cgraph_indirectly_called(cgraph_node_ptr node, void *data __unused)
{
	if (node->address_taken) {
		int idx;
		struct ipa_ref *ref = NULL, *addressable = NULL;

		// EXPORT_SYMBOL and static_call_key create an address reference that does not indicate an indirect call
		// ignore it if it's the only kind of address reference to the function
		for (idx = 0; ipa_ref_list_referring_iterate(&node->ref_list, idx, ref); idx++) {
			switch (ref->use) {
			default:
				debug_cgraph_node(node);
				debug_cgraph_node(ipa_ref_referring_cgraph_node(ref));
				gcc_unreachable();

			case IPA_REF_ADDR:
				if (rap_fptr_is_dereferenced(node, ref))
					return true;

				addressable = ref;
				continue;

			case IPA_REF_ALIAS:
				continue;
			}
		}

		return idx && addressable == NULL;
	}

	if (DECL_STATIC_CONSTRUCTOR(NODE_DECL(node)) || DECL_STATIC_DESTRUCTOR(NODE_DECL(node)))
		return true;

	return false;
}

static sbitmap indirectly_called;
static struct cgraph_node_hook_list *function_insertion_hook_holder;
static struct cgraph_node_hook_list *node_removal_hook_holder;
static struct cgraph_2node_hook_list *node_duplication_hook_holder;
static bool after_execute;

static bool rap_indirectly_called(cgraph_node_ptr node)
{
	gcc_assert(indirectly_called);
	if (NODE_UID(node) >= SBITMAP_SIZE(indirectly_called)) {
		tree fndecl;

		fndecl = NODE_DECL(node);
		if (DECL_STATIC_CONSTRUCTOR(fndecl) || DECL_STATIC_DESTRUCTOR(fndecl))
			return true;

		gcc_assert(fndecl_built_in_class_p(fndecl, BUILT_IN_NORMAL));
		gcc_assert(DECL_FUNCTION_CODE(fndecl) == BUILT_IN_UNREACHABLE);

		return false;
	}
	return sbitmap_bit_p(indirectly_called, NODE_UID(node));
}

static void rap_add_new_function(cgraph_node_ptr node, void *data __unused)
{
	gcc_assert(!in_lto_p || !after_execute);
	gcc_assert(!NODE_GLOBAL(node)->inlined_to);
	gcc_unreachable();
}

static void rap_remove_node_data(cgraph_node_ptr node, void *data __unused)
{
	gcc_assert(indirectly_called);
	gcc_assert(NODE_UID(node) < SBITMAP_SIZE(indirectly_called));
	sbitmap_set_bit(indirectly_called, NODE_UID(node));
}

static void rap_duplicate_node_data(cgraph_node_ptr src, cgraph_node_ptr dst, void *data __unused)
{
	gcc_assert(!in_lto_p || !after_execute);
	gcc_assert(NODE_UID(src) != NODE_UID(dst));

	if (NODE_UID(dst) >= SBITMAP_SIZE(indirectly_called))
		indirectly_called = sbitmap_resize(indirectly_called, NODE_UID(dst) + 1, 1);

	if (!sbitmap_bit_p(indirectly_called, NODE_UID(src)))
		sbitmap_clear_bit(indirectly_called, NODE_UID(dst));
}

static void rap_indirectly_called_register_hooks(void)
{
	static bool init_p = false;

	if (init_p)
		return;

	init_p = true;

	// XXX gcc 5-7 are buggy here, see https://gcc.gnu.org/bugzilla/show_bug.cgi?id=82256
	// TODO: note that while create_version_clone_with_body calls call_cgraph_insertion_hooks
	// with the wrong node, before that it calls create_version_clone which calls call_cgraph_duplication_hooks
	// with the correct target so it could be inferred in the insertion hook as there should be
	// no other callbacks in-between
	function_insertion_hook_holder = cgraph_add_function_insertion_hook(&rap_add_new_function, NULL);

	node_removal_hook_holder = cgraph_add_node_removal_hook(&rap_remove_node_data, NULL);
	node_duplication_hook_holder = cgraph_add_node_duplication_hook(&rap_duplicate_node_data, NULL);
}

static void rap_indirectly_called_generate_summary(void)
{
	cgraph_node_ptr node;

	gcc_assert(!indirectly_called);

	indirectly_called = sbitmap_alloc(cgraph_max_uid);
	sbitmap_ones(indirectly_called);

	rap_indirectly_called_register_hooks();

	FOR_EACH_FUNCTION(node) {
		gcc_assert(!NODE_GLOBAL(node)->inlined_to);
		gcc_assert(NODE_UID(node) < SBITMAP_SIZE(indirectly_called));

		if (NODE_THUNK(node) || node->alias)
			continue;
		if (!cgraph_for_node_and_aliases(node, rap_cgraph_indirectly_called, NULL, true))
			sbitmap_clear_bit(indirectly_called, NODE_UID(node));
	}
}

static unsigned int rap_indirectly_called_execute(void)
{
	after_execute = true;
	return 0;
}

// Omit the IPA/LTO callbacks until https://gcc.gnu.org/bugzilla/show_bug.cgi?id=61311 gets fixed (license concerns)
void __weak rap_indirectly_called_write_summary(void) {}
void __weak rap_indirectly_called_write_optimization_summary(void) {}
void __weak rap_indirectly_called_read_summary(void);
void __weak rap_indirectly_called_read_optimization_summary(void);

#define PASS_NAME rap_indirectly_called
#define NO_STMT_FIXUP
#define NO_FUNCTION_TRANSFORM
#define NO_VARIABLE_TRANSFORM
#define NO_GATE
#include "gcc-generate-ipa-pass.h"

static bool rap_cgraph_indirectly_callable(cgraph_node_ptr node, void *data __unused)
{
	if (node->externally_visible)
		return true;

	return false;
}

static bool rap_tree_indirectly_callable(tree decl, void *data __unused)
{
	return TREE_PUBLIC(decl);
}

static bool rap_indirectly_callable(cgraph_node_ptr node)
{
	if (cgraph_for_node_and_aliases(node, rap_cgraph_indirectly_callable, NULL, true))
		return true;

	if (tree_for_decl_and_aliases(NODE_DECL(node), rap_tree_indirectly_callable, NULL))
		return true;

	return false;
}

static void rap_hash_align(const_tree decl)
{
	unsigned HOST_WIDE_INT rap_hash_offset;
	unsigned HOST_WIDE_INT skip;

	if (DECL_USER_ALIGN(decl))
		skip = DECL_ALIGN_UNIT(decl);
	else
		skip = 1ULL << align_functions_log;

//	if (!optimize_function_for_speed_p(cfun))
//		return;

	rap_hash_offset = UNITS_PER_WORD / 4 * sizeof(rap_hash_t);
#ifdef TARGET_386
	if (skip > rap_hash_offset)
		skip -= rap_hash_offset;
	else if (skip < UNITS_PER_WORD / 4)
		skip = UNITS_PER_WORD / 4;

	{
		unsigned char padding[skip];
		unsigned HOST_WIDE_INT i = 0;

		// this byte sequence helps disassemblers not trip up on the following rap hash
		memset(padding, 0xcc, sizeof padding - 1);
		padding[sizeof padding - 1] = 0xb8;
		if (UNITS_PER_WORD == 8 && sizeof padding > 1)
			padding[sizeof padding - 2] = 0x48;

		gcc_assert(targetm.asm_out.byte_op);
		while (i++ < skip) {
			if (i % 8 == 1)
				fputs(targetm.asm_out.byte_op, asm_out_file);
			fprintf(asm_out_file, "0x%02x%s", padding[i - 1], i == skip || i % 8 == 0 ? "\n" : ", ");
		}
	}
#else
	if (skip > rap_hash_offset)
		ASM_OUTPUT_SKIP(asm_out_file, skip - rap_hash_offset);
#endif
}

static unsigned int count_aliases(cgraph_node_ptr node, tree *aliasdecl)
{
	unsigned int aliases;
	struct ipa_ref *ref;

	aliases = 0;
	*aliasdecl = NULL_TREE;
	ref = NULL;

	FOR_EACH_ALIAS(node, ref) {
		cgraph_node_ptr refnode;

		if (ref->use != IPA_REF_ALIAS)
			continue;

		refnode = ipa_ref_referring_cgraph_node(ref);
		if (!refnode)
			continue;

		gcc_assert(refnode->externally_visible);
		*aliasdecl = NODE_DECL(refnode);
		aliases++;
	}

	return aliases;
}

static void rap_begin_function(tree decl)
{
	cgraph_node_ptr node;
	rap_hash_t imprecise_rap_hash;
	const char *asmsize, *asmname;
	char *raphashname;

	// chain to previous callback
	if (old_debug_hooks && old_debug_hooks->begin_function)
		old_debug_hooks->begin_function(decl);

	asmsize = UNITS_PER_WORD == 8 ? ".quad" : ".long";
	gcc_assert(DECL_ASSEMBLER_NAME(decl));
	asmname = IDENTIFIER_POINTER(DECL_ASSEMBLER_NAME(decl));

	gcc_assert(asprintf(&raphashname, "rap_hash.%s", asmname) != -1);
	if (DECL_WEAK(decl))
		ASM_OUTPUT_WEAK_ALIAS(asm_out_file, raphashname, NULL);
	else if (TREE_PUBLIC(decl))
		targetm.asm_out.globalize_label(asm_out_file, raphashname);
	fprintf(asm_out_file, "\t%s:\n", raphashname);
	fprintf(asm_out_file, "\t.type %s, @function\n", raphashname);

	// align the rap hash if necessary
	rap_hash_align(decl);

	// don't compute hash for functions called only directly
	node = cgraph_get_node(decl);
	gcc_assert(node);

	if (rap_indirectly_called(node))
		imprecise_rap_hash = rap_lookup_imprecise_rap_hash(decl);
	else if (rap_indirectly_callable(node)) {
		tree aliasdecl = NULL;
		const char *aliasasmname = asmname;

		imprecise_rap_hash = rap_lookup_imprecise_rap_hash(decl);

		// emit the weak hash if
		// - the function is public and has no public aliases
		// - the function is static and has a single public alias whose name will be used for the weak hash
		switch (count_aliases(node, &aliasdecl)) {
		case 0:
			gcc_assert(!aliasdecl);
			if (TREE_PUBLIC(decl))
				aliasdecl = decl;
			break;

		case 1:
			if (!TREE_PUBLIC(decl)) {
				aliasasmname = IDENTIFIER_POINTER(DECL_ASSEMBLER_NAME(aliasdecl));
				break;
			}
			// FALLTHROUGH

		default:
			aliasdecl = NULL;
		}

		if (aliasdecl) {
			char *name = NULL;

			gcc_assert(asprintf(&name, "__rap_hash_icall_%s", aliasasmname) != -1);

			if (report_func_hash)
				inform(DECL_SOURCE_LOCATION(decl), "ifunc rap_hash: %x %s", imprecise_rap_hash.hash, aliasasmname);

			ASM_OUTPUT_WEAK_ALIAS(asm_out_file, name, NULL);
			ASM_OUTPUT_TYPE_DIRECTIVE(asm_out_file, name, "object");
			fprintf(asm_out_file, "\t%s %s\n", asmsize, name);
			free(name);
			goto out;
		}
	} else
		imprecise_rap_hash.hash = 0;

	if (report_func_hash)
		inform(DECL_SOURCE_LOCATION(decl), "func rap_hash: %x %s", imprecise_rap_hash.hash, asmname);

	if (UNITS_PER_WORD == 8)
		fprintf(asm_out_file, "\t%s %#llx\t%s __rap_hash_call_%s\n", asmsize, (long long)imprecise_rap_hash.hash, ASM_COMMENT_START, asmname);
	else
		fprintf(asm_out_file, "\t%s %#x\t%s __rap_hash_call_%s\n", asmsize, imprecise_rap_hash.hash, ASM_COMMENT_START, asmname);

out:
	fprintf(asm_out_file, "\t.size %s, .-%s\n", raphashname, raphashname);
	free(raphashname);

#if 0
	// this is a hack to satisfy objtool by providing non-vanishing aliasing symbol to weak functions
	if (!DECL_WEAK(decl))
		return;

	gcc_assert(asprintf(&raphashname, "%s_weak", asmname) != -1);
	ASM_OUTPUT_DEF(asm_out_file, raphashname, asmname);
	ASM_OUTPUT_WEAK_ALIAS(asm_out_file, raphashname, NULL);
	free(raphashname);
#endif
}

static void rap_start_unit_xor(void *gcc_data __unused, void *user_data __unused)
{
	//register unsigned long rap_cookie asm("rap_reg_name");
	rap_cookie_decl = build_decl(UNKNOWN_LOCATION, VAR_DECL, get_identifier("rap_cookie"), long_unsigned_type_node); //ptr_type_node);
	DECL_ARTIFICIAL(rap_cookie_decl) = 1;
	DECL_INITIAL(rap_cookie_decl) = NULL;
	DECL_COMMON(rap_cookie_decl) = 1;
	TREE_PUBLIC(rap_cookie_decl) = 1;
	TREE_STATIC(rap_cookie_decl) = 1;
	TREE_USED(rap_cookie_decl) = 1;
	DECL_PRESERVE_P(rap_cookie_decl) = 1;
//	C_DECL_REGISTER(rap_cookie_decl) = 1;
	DECL_REGISTER(rap_cookie_decl) = 1;
	DECL_HARD_REGISTER(rap_cookie_decl) = 1;
	set_user_assembler_name(rap_cookie_decl, rap_reg_name);
	make_decl_rtl(rap_cookie_decl);
}

static void rap_start_unit_ret(void *gcc_data __unused, void *user_data __unused)
{
#ifdef TARGET_386
#if BUILDING_GCC_VERSION >= 10000
	leaveq_rtx = PATTERN(gen_leave(word_mode));
#else
	leaveq_rtx = TARGET_64BIT ? gen_leave_rex64() : gen_leave();;
#endif
	pop_rbp_rtx = gen_pop(hard_frame_pointer_rtx);
#endif
}

static bool rap_hash_section_open;

static void rap_hash_open_section(const char *caller)
{
	gcc_assert(!rap_hash_section_open);
	rap_hash_section_open = true;
	fprintf(asm_out_file, "\t.pushsection .rap_hash,\"a\" %s %s\n", ASM_COMMENT_START, caller);
}
#define rap_hash_open_section() rap_hash_open_section(__func__)

static void rap_hash_close_section(const char *caller)
{
	gcc_assert(rap_hash_section_open);
	rap_hash_section_open = false;
	fprintf(asm_out_file, "\t.popsection %s %s\n", ASM_COMMENT_START, caller);
}
#define rap_hash_close_section() rap_hash_close_section(__func__)

static void rap_emit_hash_symbol(const char *type, const char *asmname, rap_hash_t hash)
{
	char *name = NULL;

	gcc_assert(rap_hash_section_open);
	gcc_assert(asprintf(&name, "__rap_hash_%s_%s", type, asmname) != -1);

	fprintf(asm_out_file, GLOBAL_ASM_OP " %s\n", name);
	if (UNITS_PER_WORD == 8)
		fprintf(asm_out_file, "\t.offset %#018llx\n", (long long)hash.hash);
	else if (UNITS_PER_WORD == 4)
		fprintf(asm_out_file, "\t.offset %#010x\n", hash.hash);
	else
		gcc_unreachable();

	ASM_OUTPUT_TYPE_DIRECTIVE(asm_out_file, name, "object");
	ASM_OUTPUT_LABEL(asm_out_file, name);
	free(name);
}

static void rap_emit_hash_symbols(const char *asmname, rap_hash_t hash)
{
	rap_emit_hash_symbol("call", asmname, hash);
	hash.hash = -hash.hash;
	rap_emit_hash_symbol("ret", asmname, hash);
}

static void rap_emit_nonweak_hash(void *gcc_data __unused, void *user_data __unused)
{
	cgraph_node_ptr node;
	rap_hash_t hash;

	rap_hash_open_section();
	FOR_EACH_FUNCTION(node) {
		tree fndecl;
		const char *asmname;

		if (NODE_THUNK(node))
			continue;

		if (!rap_indirectly_called(node))
			continue;
		if (node->alias && !rap_cgraph_indirectly_called(node, NULL))
			continue;

		gcc_assert(node->alias || cgraph_function_or_thunk_node(node, NULL) == node);

		fndecl = NODE_DECL(node);
		gcc_assert(fndecl);
		if (DECL_IS_UNDECLARED_BUILTIN(fndecl) && DECL_BUILT_IN_CLASS(fndecl) == BUILT_IN_NORMAL)
			continue;

		if (!TREE_PUBLIC(fndecl))
			continue;

		if (DECL_ARTIFICIAL(fndecl))
			continue;

		if (DECL_ABSTRACT_ORIGIN(fndecl) && DECL_ABSTRACT_ORIGIN(fndecl) != fndecl)
			continue;

		gcc_assert(DECL_ASSEMBLER_NAME(fndecl));
		asmname = IDENTIFIER_POINTER(DECL_ASSEMBLER_NAME(fndecl));
		if (strchr(asmname, '.'))
			continue;

		if (asmname[0] == '*')
			asmname++;

		gcc_assert(asmname[0]);

		hash = rap_lookup_imprecise_rap_hash(fndecl);
		if (report_func_hash)
			inform(DECL_SOURCE_LOCATION(fndecl), "finish ifunc rap_hash: %x %s", hash.hash, IDENTIFIER_POINTER(DECL_ASSEMBLER_NAME(fndecl)));
		rap_emit_hash_symbol("icall", asmname, hash);
	}
	rap_hash_close_section();
}

/*
   emit an absolute symbol for each function that may be referenced through the plt
     - all externs
     - non-static functions
       - use visibility instead?

   .globl __rap_hash_call_func
   .offset 0xhash_for_func
   .type __rap_hash_call_func, @object
   __rap_hash_call_func:
   .previous
*/
static void rap_finish_unit(void *gcc_data __unused, void *user_data __unused)
{
	cgraph_node_ptr node;
	rap_hash_t hash;

	hash.hash = 0;
	rap_hash_open_section();
	FOR_EACH_FUNCTION(node) {
		tree fndecl;
		const char *asmname;

		if (NODE_THUNK(node) || node->alias)
			continue;
		if (cgraph_function_body_availability(node) >= AVAIL_INTERPOSABLE) {
			if (!rap_indirectly_called(node) && !rap_indirectly_callable(node))
				continue;
		}

		gcc_assert(cgraph_function_or_thunk_node(node, NULL) == node);

		fndecl = NODE_DECL(node);
		gcc_assert(fndecl);

		if (!TREE_PUBLIC(fndecl))
			continue;

		if (DECL_ARTIFICIAL(fndecl))
			continue;

//		fprintf(stderr, "[%s:%d] '%s'...\n", __func__,  __LINE__, DECL_NAME_POINTER(fndecl));
		if (DECL_ABSTRACT_ORIGIN(fndecl) && DECL_ABSTRACT_ORIGIN(fndecl) != fndecl)
			continue;

		gcc_assert(DECL_ASSEMBLER_NAME(fndecl));
		asmname = IDENTIFIER_POINTER(DECL_ASSEMBLER_NAME(fndecl));
		if (strchr(asmname, '.'))
			continue;

		if (asmname[0] == '*')
			asmname++;

		gcc_assert(asmname[0]);

		if (DECL_IS_UNDECLARED_BUILTIN(fndecl) &&
		    DECL_BUILT_IN_CLASS(fndecl) == BUILT_IN_NORMAL &&
		    !strcmp(asmname, DECL_NAME_POINTER(fndecl)))
			continue;
//		fprintf(stderr, "[%s:%d] '%s' '%s'...\n", __func__,  __LINE__, DECL_NAME_POINTER(fndecl), IDENTIFIER_POINTER(DECL_ASSEMBLER_NAME(fndecl)));

		hash = rap_lookup_imprecise_rap_hash(fndecl);
		if (report_abs_hash)
			inform(DECL_SOURCE_LOCATION(fndecl), "finish abs rap_hash: %x %s", hash.hash, IDENTIFIER_POINTER(DECL_ASSEMBLER_NAME(fndecl)));
		rap_emit_hash_symbols(asmname, hash);
	}
	rap_hash_close_section();
}

static bool rap_reserve_register(void)
{
	int reg;

#ifdef TARGET_386
	if (TARGET_64BIT) {
		rap_reg_name = "%r12";
		asm_call_insn_pattern = "[[:space:]]?callq?[[:space:]]+";
	} else {
		rap_reg_name = "%ebx";
		asm_call_insn_pattern = "[[:space:]]?l?call[[:space:]]+";
	}
#elif defined(TARGET_ARM) || defined(TARGET_THUMB)
	rap_reg_name = "%r10";
	asm_call_insn_pattern	= "[[:space:]]?blx?[[:space:]]+";
#elif defined(AARCH64_ABI_DEFAULT)
	rap_reg_name = "x19";
	asm_call_insn_pattern	= "[[:space:]]?bl?[[:space:]]+";
#elif defined(SPARC_DEFAULT_CMODEL)
	rap_reg_name = "%g5";
	asm_call_insn_pattern	= "[[:space:]]?call?[[:space:]]+";
#else
#error unsupported target
#endif

	if (enable_xor) {
		reg = decode_reg_name(rap_reg_name);
		if (reg < 0)
			return false;

		rap_reg_id = reg;
		if (fixed_regs[rap_reg_id] == 1)
			return false;
#ifdef TARGET_386
		if (TARGET_64BIT && fixed_regs[rap_reg_id] == 3)
			return false;
		if (!TARGET_64BIT && fixed_regs[rap_reg_id] == 2)
			return false;
#endif
		fix_register(rap_reg_name, 1, 1);
	}

	return true;
}

static void (*orig_option_override)(void);

static void rap_target_option_override(void)
{
	if (orig_option_override)
		orig_option_override();

	if (!rap_reserve_register())
		error(G_("register %s is already fixed"), rap_reg_name);

	if (regcomp(&asm_call_insn_regex, asm_call_insn_pattern, REG_EXTENDED | REG_ICASE | REG_NOSUB))
		error(G_("regcomp of %s failed"), asm_call_insn_pattern);
}

static void rap_reserve_register_hook(void)
{
	orig_option_override = targetm.target_option.override;
	targetm.target_option.override = rap_target_option_override;
}

// emit the rap hash as an absolute symbol for all functions seen in the frontend
// this is necessary as later unreferenced nodes will be removed yet we'd like to emit as many hashes as possible
static void rap_emit_hash_symbols_finish_decl(void *event_data, void *data __unused)
{
	tree fndecl = (tree)event_data;
	rap_hash_t hash;
	const char *asmname;

	if (fndecl == error_mark_node)
		return;

	if (TREE_CODE(fndecl) != FUNCTION_DECL)
		return;

	gcc_assert(TREE_CODE(TREE_TYPE(fndecl)) == FUNCTION_TYPE);
	if (!TREE_PUBLIC(fndecl))
		return;

	if (DECL_ARTIFICIAL(fndecl))
		return;

	if (DECL_ABSTRACT_ORIGIN(fndecl) && DECL_ABSTRACT_ORIGIN(fndecl) != fndecl)
		return;

	asmname = DECL_NAME_POINTER(fndecl);
	gcc_assert(asmname[0]);
	gcc_assert(!strchr(asmname, '.'));

	gcc_assert(lookup_attribute("rap_hash", TYPE_ATTRIBUTES(TREE_TYPE(fndecl))));
	hash = rap_lookup_imprecise_rap_hash(fndecl);

	rap_hash_open_section();
	rap_emit_hash_symbols(asmname, hash);
	rap_hash_close_section();
	if (report_abs_hash)
		inform(DECL_SOURCE_LOCATION(fndecl), "decl abs rap_hash: %x %s", hash.hash, asmname);
}

static void rap_emit_hash_symbols_type(const_tree type, const char *prefix)
{
	const_tree field;

	if (TYPE_FIELDS(type) == NULL_TREE)
		return;

	// TODO skip constified types for now
	if (TYPE_READONLY(type))
		return;

	// create the prefix if it hasn't been done yet
	if (!*prefix) {
		const_tree name = type_name(type);

		// skip an anonymous struct embedded inside another one
		// we'll see it when we walk the parent later
		if (!name)
			return;

		prefix = IDENTIFIER_POINTER(name);
		gcc_assert(*prefix);
	}

	for (field = TYPE_FIELDS(type); field; field = TREE_CHAIN(field)) {
		const_tree fieldtype, fieldname;
		char *hashname = NULL, *newprefix = NULL;
		rap_hash_t hash;

		fieldtype = TREE_TYPE(field);
		switch (TREE_CODE(fieldtype)) {
		default:
			continue;

		case RECORD_TYPE:
		case UNION_TYPE:
			fieldname = DECL_NAME(field);
			if (!fieldname)
				continue;
			gcc_assert(asprintf(&newprefix, "%s.%s", prefix, IDENTIFIER_POINTER(fieldname)) != -1);
			rap_emit_hash_symbols_type(fieldtype, newprefix);
			free(newprefix);
			continue;

		case POINTER_TYPE:
			fieldtype = TREE_TYPE(fieldtype);
			if (TREE_CODE(fieldtype) != FUNCTION_TYPE)
				continue;

			hash = rap_lookup_imprecise_rap_hash(fieldtype);
			fieldname = DECL_NAME(field);
			gcc_assert(fieldname);
			gcc_assert(asprintf(&hashname, "%s.%s", prefix, IDENTIFIER_POINTER(fieldname)) != -1);
			rap_emit_hash_symbols(hashname, hash);
			if (report_abs_hash)
				inform(DECL_SOURCE_LOCATION(field), "struct abs rap_hash: %x %s", hash.hash, hashname);
			free(hashname);
			continue;
		}
	}
}

static void rap_emit_hash_symbols_finish_type(void *event_data, void *data __unused)
{
	const_tree type = (const_tree)event_data;

	if (type == NULL_TREE || type == error_mark_node)
		return;

	if (!COMPLETE_TYPE_P(type))
		return;

	if (!lookup_attribute("rap_hash", TYPE_ATTRIBUTES(type)))
		return;

	switch (TREE_CODE(type)) {
	default:
		debug_tree(type);
		gcc_unreachable();

	case ENUMERAL_TYPE:
	case UNION_TYPE:
		break;

	case RECORD_TYPE:
		rap_hash_open_section();
		rap_emit_hash_symbols_type(type, "");
		rap_hash_close_section();
		break;
	}
}

#if BUILDING_GCC_VERSION >= 9000
static void rap_handle_typedef_main_variant(void *event_data, void *data __unused)
{
	tree deftype, maintype, typedecl;

	typedecl = (tree)event_data;
	if (typedecl == error_mark_node)
		return;

	if (TREE_CODE(typedecl) != TYPE_DECL)
		return;

	deftype = TREE_TYPE(typedecl);
	if (!TYPE_NAME(deftype))
		return;

	maintype = TYPE_MAIN_VARIANT(deftype);
	if (TYPE_NAME(maintype))
		return;

	if (lookup_attribute("first_variant", TYPE_ATTRIBUTES(deftype))) {
		TYPE_ATTRIBUTES(deftype) = remove_attribute("first_variant", TYPE_ATTRIBUTES(deftype));
		return;
	}

	TYPE_ATTRIBUTES(deftype) = copy_list(TYPE_ATTRIBUTES(deftype));
	TYPE_ATTRIBUTES(deftype) = tree_cons(get_identifier("first_variant"), NULL_TREE, TYPE_ATTRIBUTES(deftype));
}
#endif

static void rap_assembly_start(void)
{
	// chain to previous callback
	if (old_debug_hooks && old_debug_hooks->assembly_start)
		old_debug_hooks->assembly_start();

#ifdef TARGET_386
	if (enable_type_call || enable_type_ret) {
		fprintf(asm_out_file, "%s",
			"\t.macro rap_abort kind:req reg:req\n"
			"\t\t.ifc \\reg,%eax\n"
			"\t\t\t.byte 0x0f,0xb9,0x00|\\kind\n"
			"\t\t\t.exitm\n"
			"\t\t.endif\n"

			"\t\t.ifc \\reg,%ecx\n"
			"\t\t\t.byte 0x0f,0xb9,0x08|\\kind\n"
			"\t\t\t.exitm\n"
			"\t\t.endif\n"

			"\t\t.ifc \\reg,%edx\n"
			"\t\t\t.byte 0x0f,0xb9,0x10|\\kind\n"
			"\t\t\t.exitm\n"
			"\t\t.endif\n"

			"\t\t.ifc \\reg,%ebx\n"
			"\t\t\t.byte 0x0f,0xb9,0x18|\\kind\n"
			"\t\t\t.exitm\n"
			"\t\t.endif\n"

			"\t\t.ifc \\reg,%esp\n"
			"\t\t\t.byte 0x0f,0xb9,0x20|\\kind\n"
			"\t\t\t.exitm\n"
			"\t\t.endif\n"

			"\t\t.ifc \\reg,%ebp\n"
			"\t\t\t.byte 0x0f,0xb9,0x28|\\kind\n"
			"\t\t\t.exitm\n"
			"\t\t.endif\n"

			"\t\t.ifc \\reg,%esi\n"
			"\t\t\t.byte 0x0f,0xb9,0x30|\\kind\n"
			"\t\t\t.exitm\n"
			"\t\t.endif\n"

			"\t\t.ifc \\reg,%edi\n"
			"\t\t\t.byte 0x0f,0xb9,0x38|\\kind\n"
			"\t\t\t.exitm\n"
			"\t\t.endif\n"

			"\t\t.ifc \\reg,%r8d\n"
			"\t\t\t.byte 0x0f,0xb9,0xc0|\\kind\n"
			"\t\t\t.exitm\n"
			"\t\t.endif\n"

			"\t\t.ifc \\reg,%r9d\n"
			"\t\t\t.byte 0x0f,0xb9,0xc8|\\kind\n"
			"\t\t\t.exitm\n"
			"\t\t.endif\n"

			"\t\t.ifc \\reg,%r10d\n"
			"\t\t\t.byte 0x0f,0xb9,0xd0|\\kind\n"
			"\t\t\t.exitm\n"
			"\t\t.endif\n"

			"\t\t.ifc \\reg,%r11d\n"
			"\t\t\t.byte 0x0f,0xb9,0xd8|\\kind\n"
			"\t\t\t.exitm\n"
			"\t\t.endif\n"

			"\t\t.ifc \\reg,%r12d\n"
			"\t\t\t.byte 0x0f,0xb9,0xe0|\\kind\n"
			"\t\t\t.exitm\n"
			"\t\t.endif\n"

			"\t\t.ifc \\reg,%r13d\n"
			"\t\t\t.byte 0x0f,0xb9,0xe8|\\kind\n"
			"\t\t\t.exitm\n"
			"\t\t.endif\n"

			"\t\t.ifc \\reg,%r14d\n"
			"\t\t\t.byte 0x0f,0xb9,0xf0|\\kind\n"
			"\t\t\t.exitm\n"
			"\t\t.endif\n"

			"\t\t.ifc \\reg,%r15d\n"
			"\t\t\t.byte 0x0f,0xb9,0xf8|\\kind\n"
			"\t\t\t.exitm\n"
			"\t\t.endif\n"

			"\t\t.error \"unknown register \\reg\"\n"
			"\t.endm\n"
		);
	}

	if (enable_type_call || enable_type_ret) {
		if (!rap_include) {
			fprintf(asm_out_file,
				"\t.macro rap_retloc callee=\"\" hashval=0\n"
				"\tjmp .Lrap_retloc_end_\\@:\n"
				"\t\t%1$s\n"
				"\t\t.ifnb \\callee\n"
				"\t\t%2$s __rap_hash_ret_\\callee\n"
				"\t\t.elseif \\hashval\n"
				"\t\t%2$s \\hashval\n"
				"\t\t.else\n"
				"\t\t.error \"either 'callee' or 'hashval' must be specified\"\n"
				"\t\t.endif\n"
				"\t.Lrap_retloc_end_\\@:\n"
				"\t.endm\n",
				(UNITS_PER_WORD == 8 ? ".byte 0x48, 0xb8" : ".byte 0xb8"),
				(UNITS_PER_WORD == 8 ? ".quad" : ".long")
			);
			gcc_assert(RAP_RET_HASH_OFFSET == NULL_TREE);
			RAP_RET_HASH_OFFSET = build_int_cst(integer_type_node, UNITS_PER_WORD == 8 ? 5 + 2 : 2 + 1);
		}

		fprintf(asm_out_file,
			"\t.macro rap_indirect_call target:req hashsym:req\n"
			ASM_COMMENT_START "\t\t.p2align 5,,2\n"
			"\t\tcall \\target\n"
			"\t\trap_retloc callee=\\hashsym\n"
			"\t.endm\n"
		);

		fprintf(asm_out_file,
			"\t.macro rap_direct_call target:req hashsym=\"\"\n"
			"\t\t.ifb \\hashsym\n"
			"\t\t\trap_indirect_call \\target, \\target\n"
			"\t\t.else\n"
			"\t\t\trap_indirect_call \\target, \\hashsym\n"
			"\t\t.endif\n"
			"\t.endm\n"
		);

		fprintf(asm_out_file,
			"\t.macro rap_call_abort reg:req\n"
			"\t\trap_abort 0,\\reg\n"
			"\t.endm\n"
		);
	}

	if (enable_xor || enable_type_ret) {
		if (!rap_include)
			fprintf(asm_out_file,
				"\t.macro rap_ret func:req\n"
				"\t\tret\n"
				"\t.endm\n"
			);

		fprintf(asm_out_file,
			"\t.macro rap_ret_abort reg:req\n"
			"\t\trap_abort 1,\\reg\n"
			"\t.endm\n"
		);

		fprintf(asm_out_file,
			"\t.macro rap_xor_abort reg:req\n"
			"\t\trap_abort 2,\\reg\n"
			"\t.endm\n"
		);
	}
#elif defined(TARGET_ARM) || defined(TARGET_THUMB)
	if (enable_type_call || enable_type_ret) {
		fprintf(asm_out_file,
			"\t.macro rap_indirect_call target hashsym\n"
			"\t\tbl \\target\n"
			"\t\tb .Lrap_indirect_call_end_\\@\n"
			"\t\t%s __rap_hash_ret_\\hashsym\n"
			"\t.Lrap_indirect_call_end_\\@:\n"
			"\t.endm\n",
			(UNITS_PER_WORD == 8 ? ".quad" : ".long")
		);

		fprintf(asm_out_file,
			"\t.macro rap_direct_call target hashsym=\"\"\n"
			"\t\t.ifb \\hashsym\n"
			"\t\trap_indirect_call \\target \\target\n"
			"\t\t.else\n"
			"\t\trap_indirect_call \\target \\hashsym\n"
			"\t\t.endif\n"
			"\t.endm\n"
		);
	}

	if (enable_type_ret) {
		fprintf(asm_out_file,
			"\t.macro rap_ret target\n"
			"\t\tmov pc,lr\n"
			"\t.endm\n"
		);
	}
#elif defined(AARCH64_ABI_DEFAULT)
	if (enable_type_call || enable_type_ret) {
		fprintf(asm_out_file,
			"\t.macro rap_indirect_call target hashsym\n"
			"\t\tbl \\target\n"
			"\t\tb .Lrap_indirect_call_end_\\@\n"
			"\t\t%s __rap_hash_ret_\\hashsym\n"
			"\t.Lrap_indirect_call_end_\\@:\n"
			"\t.endm\n",
			(UNITS_PER_WORD == 8 ? ".quad" : ".long")
		);

		fprintf(asm_out_file,
			"\t.macro rap_direct_call target hashsym=\"\"\n"
			"\t\t.ifb \\hashsym\n"
			"\t\trap_indirect_call \\target \\target\n"
			"\t\t.else\n"
			"\t\trap_indirect_call \\target \\hashsym\n"
			"\t\t.endif\n"
			"\t.endm\n"
		);
	}

	if (enable_type_ret) {
		fprintf(asm_out_file,
			"\t.macro rap_ret target\n"
			"\t\tmov pc,lr\n"
			"\t.endm\n"
		);
	}
#elif defined(SPARC_DEFAULT_CMODEL)
	if (enable_type_call || enable_type_ret) {
		fprintf(asm_out_file,
			"\t.macro rap_indirect_call target hashsym\n"
			"\t\tcall \\target\n"
			"\t\tnop\n"
			"\t\tjmp .Lrap_indirect_call_end_\\@; nop\n"
			"\t\t%s __rap_hash_ret_\\hashsym\n"
			"\t.Lrap_indirect_call_end_\\@:\n"
			"\t.endm\n",
			(UNITS_PER_WORD == 8 ? ".quad" : ".long")
		);

		fprintf(asm_out_file,
			"\t.macro rap_direct_call target hashsym=\"\"\n"
			"\t\t.ifb \\hashsym\n"
			"\t\trap_indirect_call \\target \\target\n"
			"\t\t.else\n"
			"\t\trap_indirect_call \\target \\hashsym\n"
			"\t\t.endif\n"
			"\t.endm\n"
		);
	}

	if (enable_type_ret) {
		fprintf(asm_out_file,
			"\t.macro rap_ret target\n"
			"\t\tret\n"
			"\t.endm\n"
		);
	}
#else
#error unsupported target
#endif
}

static void rap_update_eni(void)
{
	// Tweak the inliner to be aware of our instrumentation overhead
	if (enable_sls) {
		// account for the XOR
		eni_size_weights.call_cost += 1;
		eni_size_weights.indirect_call_cost += 1;
	}

	if (enable_type_ret) {
		// account for the ret_hash code sequence
		eni_size_weights.call_cost += 3;
		eni_size_weights.indirect_call_cost += 3;

		// account for the ret_hash check
		eni_size_weights.return_cost += 2;
		eni_time_weights.return_cost += 2;

		// account for KRETPROBES/RETHOOK/FTRACE trampolines
		eni_size_weights.call_cost += rethooks * (2 + enable_sls);
		eni_size_weights.indirect_call_cost += rethooks * (2 + enable_sls);
	}

	if (enable_type_call || enable_type_nospec) {
		// account for the call_hash
		eni_size_weights.call_cost += 3;
		eni_size_weights.indirect_call_cost += 3;

		// account for the call_hash check
		eni_size_weights.indirect_call_cost += 2 + !!enable_type_nospec;
	}

	if (enable_xor) {
		// account for the stack slot load&stores and XORs
		eni_time_weights.call_cost += 3;
		eni_time_weights.indirect_call_cost += 3;
		eni_size_weights.return_cost += 5;
		eni_time_weights.return_cost += 5;
	}

	if (kernexec_instrument_fptr) {
		// account for the BTS / OR fptr mangling
		eni_size_weights.indirect_call_cost += 1;
	}

	if (kernexec_instrument_retaddr) {
		// account for the BTS / OR ret addr mangling
		eni_size_weights.return_cost += 1;
		eni_time_weights.return_cost += 1;
	}
}

static void (*old_override_options_after_change)(void);
static void (*old_override_asm_out_print_operand)(FILE *file, rtx x, int code);
#if BUILDING_GCC_VERSION >= 6000
# ifndef ix86_indirect_branch_register
static rtx_insn *(*old_override_gen_tablejump)(rtx, rtx);
# endif
#endif

static void rap_override_options_after_change(void)
{
	if (old_override_options_after_change)
		old_override_options_after_change();

	flag_stack_protect = 0;
	flag_ipa_icf_functions = 0;

#if BUILDING_GCC_VERSION >= 7000
	flag_code_hoisting = 0;
#endif

#if BUILDING_GCC_VERSION >= 8000
	if ((enable_xor || enable_type_call || enable_type_ret))
		flag_cf_protection = CF_NONE;
	flag_reorder_blocks_and_partition = 0;
#endif

	if (enable_type_call || enable_type_ret)
		flag_crossjumping = 0;

	if (enable_type_nospec) {
#ifdef ix86_indirect_branch_register
		gcc_assert(ix86_indirect_branch_register == 1);
#endif

#if BUILDING_GCC_VERSION < 6000
		flag_jump_tables = 0;
#endif
	}
}

static void rap_override_asm_out_print_operand(FILE *file, rtx x, int code)
{
#ifdef TARGET_386
	if (ASSEMBLER_DIALECT == ASM_ATT && code == 'V') {
		if (!REG_P(x)) {
			print_rtl_single(stderr, x);
			gcc_unreachable();
		}
		ASSEMBLER_DIALECT = ASM_INTEL;
		print_reg(x, code, file);
		ASSEMBLER_DIALECT = ASM_ATT;
	} else if (code == 'h' && GET_CODE(x) == SYMBOL_REF) {
		tree fndecl;
		rap_hash_t hash;

		gcc_assert(SYMBOL_REF_FUNCTION_P(x));
		fndecl = SYMBOL_REF_DECL(x);
		gcc_assert(fndecl);
		if (enable_type_call || enable_type_ret)
			hash = rap_lookup_imprecise_rap_hash(fndecl);
		else
			hash.hash = 0;
		fprintf(file, HOST_WIDE_INT_PRINT_HEX, (HOST_WIDE_INT)-hash.hash);
	} else
#endif
		old_override_asm_out_print_operand(file, x, code);
}

#if BUILDING_GCC_VERSION >= 6000
#ifndef ix86_indirect_branch_register
static rtx_insn *rap_override_gen_tablejump(rtx operand0, rtx operand1)
{
	rtx_insn *retpoline;
	rtx reg, body;
	rtvec argvec, constraintvec, labelvec;

	gcc_assert(REG_P(operand0));
	reg = operand0;

	argvec = rtvec_alloc(1);
	constraintvec = rtvec_alloc(1);
	labelvec = rtvec_alloc(0);

	start_sequence();

	body = gen_rtx_ASM_OPERANDS(GET_MODE(reg), ggc_strdup(""), empty_string, 0, argvec, constraintvec, labelvec, UNKNOWN_LOCATION);
	MEM_VOLATILE_P(body) = 1;
	ASM_OPERANDS_INPUT(body, 0) = reg;
	ASM_OPERANDS_INPUT_CONSTRAINT_EXP(body, 0) = gen_rtx_ASM_INPUT_loc(GET_MODE(reg), ggc_strdup("0"), UNKNOWN_LOCATION);
	ASM_OPERANDS_OUTPUT_CONSTRAINT(body) = ggc_strdup("=r");
	emit_insn(gen_rtx_set(reg, body));

	emit_jump_insn(gen_rtx_PARALLEL(VOIDmode, gen_rtvec(2, gen_rtx_set(pc_rtx, operand0), gen_rtx_USE(VOIDmode, gen_rtx_LABEL_REF(VOIDmode, operand1)))));
	retpoline = get_insns();
	end_sequence();

	return retpoline;
}
#endif
#endif

static tree (*old_builtin_function)(tree decl);
static tree (*old_builtin_function_ext_scope)(tree decl);

static tree rap_builtin_function(tree decl)
{
	rap_calculate_fndecl_hash(decl);
	return old_builtin_function(decl);
}

static tree rap_builtin_function_ext_scope(tree decl)
{
	rap_calculate_fndecl_hash(decl);
	return old_builtin_function_ext_scope(decl);
}

#ifdef TARGET_386
static void(*old_print_operand)(FILE *file, rtx x, int code);

static void rap_print_cs_override(FILE *file, rtx x, int code)
{
	tree fptr;

	// rtl match (mem:SI (plus:SI (reg:SI 0 ax [102]) (const_int -12 [0xfffffffffffffff4])) [5 MEM[(long int *)rap_fptr2.52_25 + 4294967284B]+0 S4 A32])

	if (x == NULL_RTX)
		return;
	if (!MEM_P(x))
		return;

	fptr = MEM_EXPR(x);
	if (fptr == NULL_TREE)
		return;
	if (TREE_CODE(fptr) != MEM_REF)
		return;
	fptr = TREE_OPERAND(fptr, 0);
	if (TYPE_ADDR_SPACE(TREE_TYPE(fptr)) != ADDR_SPACE_KERNEXEC)
		return;
	gcc_assert(!code);

	fputs("%cs:", file);
}

static void rap_print_operand(FILE *file, rtx x, int code)
{
	rap_print_cs_override(file, x, code);
	old_print_operand(file, x, code);
}
#endif

static void rap_start_unit_common(void *gcc_data __unused, void *user_data __unused)
{
	rap_update_eni();

	rap_hash_typemap = typemap_init();

	if (debug_hooks)
		rap_debug_hooks = *debug_hooks;

	if (enable_xor || enable_type_call || enable_type_ret)
		rap_debug_hooks.assembly_start = rap_assembly_start;
	if (enable_type_call || enable_type_ret)
		rap_debug_hooks.begin_function = rap_begin_function;

	old_debug_hooks = debug_hooks;
	debug_hooks = &rap_debug_hooks;

	old_override_options_after_change = targetm.override_options_after_change;
	targetm.override_options_after_change = rap_override_options_after_change;

#ifdef TARGET_386
	if (!TARGET_64BIT && (enable_type_call || enable_type_ret)) {
		old_print_operand = targetm.asm_out.print_operand;
		targetm.asm_out.print_operand = rap_print_operand;
	}
#endif

	if (enable_type_call || enable_type_ret || enable_type_nospec) {
		if (targetm.asm_out.print_operand) {
			old_override_asm_out_print_operand = targetm.asm_out.print_operand;
			targetm.asm_out.print_operand = rap_override_asm_out_print_operand;
		}
	}

	if (enable_type_nospec) {
#if BUILDING_GCC_VERSION >= 6000
# ifndef ix86_indirect_branch_register
		if (targetm.gen_tablejump) {
			old_override_gen_tablejump = targetm.gen_tablejump;
			targetm.gen_tablejump = rap_override_gen_tablejump;
		}
# endif
#endif
	}
}

static bool rap_unignore_gate(void)
{
	if (!DECL_IGNORED_P(current_function_decl))
		return false;

	DECL_IGNORED_P(current_function_decl) = 0;
	return false;
}

#define PASS_NAME rap_unignore
#define NO_EXECUTE
#include "gcc-generate-rtl-pass.h"

static void handle_rap_hash_func(tree fndecl, bool icall)
{
	rap_hash_t hash;
	const char *asmname;

	if (!enable_abs_attr)
		return;

	asmname = DECL_NAME_POINTER(fndecl);
	gcc_assert(asmname[0]);

	hash = rap_lookup_imprecise_rap_hash(fndecl);

	rap_hash_open_section();
	if (icall) {
		rap_emit_hash_symbol("icall", asmname, hash);
		if (report_abs_hash)
			inform(DECL_SOURCE_LOCATION(fndecl), "abs ifunc rap_hash: %x %s", hash.hash, asmname);
	} else {
		rap_emit_hash_symbols(asmname, hash);
		if (report_abs_hash)
			inform(DECL_SOURCE_LOCATION(fndecl), "abs rap_hash: %x %s", hash.hash, asmname);
	}
	rap_hash_close_section();
}

static tree handle_rap_hash_attribute(tree *node, tree name, tree args, int flags, bool *no_add_attrs)
{
	location_t loc;

	loc = DECL_P(*node) ? DECL_SOURCE_LOCATION(*node) : UNKNOWN_LOCATION;
	*no_add_attrs = true;

	if (args != NULL_TREE) {
#if BUILDING_GCC_VERSION >= 9000
		// allow the 'copy' attr introduced in gcc 9
		if (node[1] != NULL_TREE) {
			tree fndecl;

			gcc_assert(DECL_P(*node));
			fndecl = node[1];
			gcc_assert(DECL_P(fndecl));
			gcc_assert(lookup_attribute("rap_hash", DECL_ATTRIBUTES(fndecl)));
			*no_add_attrs = false;
			return NULL_TREE;
		}
#endif

		error_at(loc, "%qE attribute on %qD must not have arguments", name, *node);
		return NULL_TREE;
	}

	switch (TREE_CODE(*node)) {
	default:
		error_at(loc, "%qE attribute applies to structure types and function declarations only (%qE)", name, *node);
		return NULL_TREE;

	case FUNCTION_DECL: {
		tree fndecl;

		fndecl = *node;
		rap_calculate_fndecl_hash(fndecl);
		handle_rap_hash_func(fndecl, false);
		return NULL_TREE;
	}

	case RECORD_TYPE:
		*no_add_attrs = false;
		return NULL_TREE;
	}
}

static tree handle_rap_hash_icall_attribute(tree *node, tree name, tree args, int flags, bool *no_add_attrs)
{
	tree fndecl;
	location_t loc;

	fndecl = *node;
	loc = DECL_SOURCE_LOCATION(fndecl);
	*no_add_attrs = true;

	if (TREE_CODE(fndecl) == FUNCTION_DECL) {
		rap_calculate_fndecl_hash(fndecl);
		handle_rap_hash_func(fndecl, true);
	} else
		error_at(loc, "%qE attribute applies to function declarations only (%qE)", name, *node);

	return NULL_TREE;
}

static tree handle_rap_hash_override_attribute(tree *node, tree name, tree args, int flags, bool *no_add_attrs)
{
	location_t loc;
	tree attr, fndecl;

	loc = DECL_SOURCE_LOCATION(*node);
	*no_add_attrs = true;

	gcc_assert(args != NULL_TREE);

	if (TREE_CODE(*node) != FUNCTION_DECL) {
		error_at(loc, "%qE attribute applies to functions only, not %qD", name, *node);
		return NULL_TREE;
	}

	fndecl = TREE_VALUE(args);
	if (fndecl == NULL_TREE || TREE_CODE(fndecl) != FUNCTION_DECL) {
		error_at(loc, "%qE attribute on %qD must have a function argument instead of %qE", name, *node, fndecl);
		return NULL_TREE;
	}

	attr = lookup_attribute("rap_hash_override", DECL_ATTRIBUTES(fndecl));
	if (attr == NULL_TREE)
		attr = lookup_attribute("rap_hash", TYPE_ATTRIBUTES(TREE_TYPE(fndecl)));
	if (attr == NULL_TREE) {
		error_at(loc, "%qE attribute on %qD refers to %qD without a rap hash", name, *node, fndecl);
		return NULL_TREE;
	}

	fndecl = *node;
	DECL_ATTRIBUTES(fndecl) = copy_list(DECL_ATTRIBUTES(fndecl));
	DECL_ATTRIBUTES(fndecl) = tree_cons(name, TREE_VALUE(attr), DECL_ATTRIBUTES(fndecl));
	handle_rap_hash_func(fndecl, false);
	return NULL_TREE;
}

static tree handle_indirect_branch_attribute(tree *node, tree name, tree args __unused, int flags, bool *no_add_attrs)
{
	tree value;
	const char *ibranch_type;

	*no_add_attrs = true;

	gcc_assert(DECL_P(*node));

	switch (TREE_CODE(*node)) {
	default:
		error_at(DECL_SOURCE_LOCATION(*node), "%qE attribute applies to functions only (%qD)", name, *node);
		return NULL_TREE;

	case FUNCTION_DECL:
		break;
	}

	value = TREE_VALUE(args);
	if (TREE_CODE(value) != STRING_CST) {
		error_at(DECL_SOURCE_LOCATION(*node), "%qE attribute requires a string constant argument, not %qE", name, value);
		return NULL_TREE;
	}

	ibranch_type = TREE_STRING_POINTER(value);
#if 0
	// TODO we don't support gcc generated retpoline thunks for now
	if (strcmp(ibranch_type, "keep") && strcmp(ibranch_type, "thunk") && strcmp(ibranch_type, "thunk-inline") && strcmp(ibranch_type, "thunk-extern")) {
		error_at(DECL_SOURCE_LOCATION(*node), "argument to %qE attribute is %qE, must be one of keep|thunk|thunk-inline|thunk-extern", name, value);
#else
	if (strcmp(ibranch_type, "keep") && strcmp(ibranch_type, "thunk-extern")) {
		error_at(DECL_SOURCE_LOCATION(*node), "argument to %qE attribute is %qE, must be keep or thunk-extern", name, value);
#endif
		return NULL_TREE;
	}

	*no_add_attrs = false;
	return NULL_TREE;
}

static tree handle_convertible_attribute(tree *node, tree name, tree args __unused, int flags, bool *no_add_attrs)
{
	tree *type;

	*no_add_attrs = true;

	if (TREE_CODE(*node) != TYPE_DECL && TREE_CODE(*node) != FIELD_DECL) {
		error("%qE attribute applies to typedefs and fields only (%qD)", name, *node);
		return NULL_TREE;
	}

	type = &TREE_TYPE(*node);

	if (*type == error_mark_node)
		return NULL_TREE;


	if (!(flags & (int)ATTR_FLAG_TYPE_IN_PLACE)) {
		tree origtype = *type;

		*type = build_variant_type_copy(*type);
		if (TREE_CODE(*node) == TYPE_DECL && DECL_ORIGINAL_TYPE(*node) == NULL_TREE) {
			DECL_ORIGINAL_TYPE(*node) = origtype;
			TYPE_NAME(*type) = *node;
			TREE_USED(*type) = TREE_USED(*node);
			TREE_TYPE(*node) = *type;
		}
	}

	if (TREE_CODE(*type) == POINTER_TYPE && TREE_CODE(TREE_TYPE(*type)) == FUNCTION_TYPE)
		add_type_attr(*type, IDENTIFIER_POINTER(name), NULL_TREE);
	else
		TYPE_ATTRIBUTES(*type) = tree_cons(name, NULL_TREE, TYPE_ATTRIBUTES(*type));

	*no_add_attrs = false;
	return NULL_TREE;
}

#if BUILDING_GCC_VERSION >= 9000
static tree handle_first_variant_attribute(tree *node, tree name, tree args, int flags, bool *no_add_attrs)
{
	error_at(DECL_SOURCE_LOCATION(*node), "%qE attribute cannot be used on %qD", name, *node);
	return NULL_TREE;
}
#endif

static struct attribute_spec rap_hash_attr = ATTRIBUTE_SPEC_INIT("rap_hash", 0, 1, false, false, false, handle_rap_hash_attribute, true, NULL);
static struct attribute_spec rap_hash_icall_attr = ATTRIBUTE_SPEC_INIT("rap_hash_icall", 0, 0, true, false, false, handle_rap_hash_icall_attribute, true, NULL);
static struct attribute_spec rap_hash_override_attr = ATTRIBUTE_SPEC_INIT("rap_hash_override", 1, 1, true, false, false, handle_rap_hash_override_attribute, true, NULL);
static struct attribute_spec indirect_branch_attr = ATTRIBUTE_SPEC_INIT("indirect_branch", 1, 1, true, false, false, handle_indirect_branch_attribute, false, NULL);
static struct attribute_spec convertible_attr = ATTRIBUTE_SPEC_INIT("convertible", 0, 0, true, false, false, handle_convertible_attribute, true, NULL);
#if BUILDING_GCC_VERSION >= 9000
static struct attribute_spec first_variant_attr = ATTRIBUTE_SPEC_INIT("first_variant", 0, 0, true, false, false, handle_first_variant_attribute, false, NULL);
#endif

static void register_attributes(void *event_data __unused, void *data __unused)
{
	if (enable_type_cast || enable_abs_finish || enable_abs_ops || enable_abs_attr || enable_type_call || enable_type_ret) {
		register_attribute(&rap_hash_attr);
		register_attribute(&rap_hash_icall_attr);
		register_attribute(&rap_hash_override_attr);
		register_attribute(&convertible_attr);
	}

#ifdef TARGET_386
	if (enable_type_nospec && !lookup_attribute_spec(get_identifier("indirect_branch")))
		register_attribute(&indirect_branch_attr);
#endif

#if BUILDING_GCC_VERSION >= 9000
	if (enable_xor || enable_type_call || enable_type_ret)
		register_attribute(&first_variant_attr);
#endif
}

EXPORTED_CONST struct ggc_root_tab gt_ggc_r_rap[] = {
	INIT_GGC_ROOT_TAB_TREE(rap_cookie_decl),
#ifdef TARGET_386
	INIT_GGC_ROOT_TAB_RTX(leaveq_rtx),
	INIT_GGC_ROOT_TAB_RTX(pop_rbp_rtx),
#endif
	LAST_GGC_ROOT_TAB
};

__visible int plugin_init(struct plugin_name_args *plugin_info, struct plugin_gcc_version *version)
{
	int i;
	const char * const plugin_name = plugin_info->base_name;
	const int argc = plugin_info->argc;
	const struct plugin_argument * const argv = plugin_info->argv;
	bool enable_abs = false;

	PASS_INFO(rap_reload,				"rap_ret",		1, PASS_POS_INSERT_BEFORE);
	PASS_INFO(rap_ret_ipa,				"static-var",		1, PASS_POS_INSERT_AFTER);
	PASS_INFO(rap_indirectly_called,		"static-var",		1, PASS_POS_INSERT_AFTER);
	PASS_INFO(rap_check_tail_calls,			"tailc",		1, PASS_POS_INSERT_AFTER);
	PASS_INFO(rap_ret,				"optimized",		1, PASS_POS_INSERT_AFTER);
	PASS_INFO(merge_phi,				"rap_ret",		1, PASS_POS_INSERT_BEFORE);
	PASS_INFO(rap_fptr,				"rap_ret",		1, PASS_POS_INSERT_AFTER);
#ifdef TARGET_386
# ifndef ix86_indirect_branch
	PASS_INFO(rap_indirect_branch_register,		"vregs",		1, PASS_POS_INSERT_BEFORE);
# endif
#endif
	PASS_INFO(rap_check_local_variable_casts,	"ssa",			1, PASS_POS_INSERT_AFTER);
	PASS_INFO(rap_check_local_register_variables,	"ssa",			1, PASS_POS_INSERT_AFTER);
	PASS_INFO(rap_rewrite_retaddr_check,		"pro_and_epilogue",	1, PASS_POS_INSERT_AFTER);
	PASS_INFO(rap_xor_elimination,			"*all-postreload",	1, PASS_POS_INSERT_BEFORE);
//	PASS_INFO(rap_fptr_safety,			"vartrack",		1, PASS_POS_INSERT_BEFORE);
	PASS_INFO(rap_mark_retloc,			"mach",			1, PASS_POS_INSERT_AFTER);
	PASS_INFO(rap_rethunk,				"rap_sls",		1, PASS_POS_INSERT_BEFORE);
	PASS_INFO(rap_sls,				"rap_retpoline",	1, PASS_POS_INSERT_BEFORE);
	PASS_INFO(rap_retpoline,			"shorten",		1, PASS_POS_INSERT_BEFORE);
	PASS_INFO(rap_unignore,				"final",		1, PASS_POS_INSERT_BEFORE);

	PASS_INFO(kernexec_reload,			"early_optimizations",	1, PASS_POS_INSERT_BEFORE);
	// unfortunately PRE can screw up fptr types from unions...
	// see cpuhp_step_startup/cpuhp_step_teardown and kernel/cpu.c:cpuhp_invoke_callback
	PASS_INFO(kernexec_fptr,			"optimized",		1, PASS_POS_INSERT_BEFORE);
	PASS_INFO(kernexec_retaddr,			"pro_and_epilogue",	1, PASS_POS_INSERT_AFTER);

	if (!plugin_default_version_check(version, &gcc_version)) {
		error_gcc_version(version);
		return 1;
	}

	for (i = 0; i < argc; ++i) {
		if (!strcmp(argv[i].key, "disable"))
			continue;

		if (!strcmp(argv[i].key, "typecheck")) {
			char *values, *value, *saveptr;

			if (!argv[i].value) {
				error(G_("no value supplied for option '-fplugin-arg-%s-%s'"), plugin_name, argv[i].key);
				continue;
			}

			values = xstrdup(argv[i].value);
			value = strtok_r(values, ",", &saveptr);
			while (value) {
				if (!strcmp(value, "ret"))
					enable_type_ret = true;
				else if (!strcmp(value, "call"))
					enable_type_call = true;
				else if (!strcmp(value, "nospec"))
					enable_type_nospec = true;
				else if (!strcmp(value, "cast"))
					enable_type_cast = true;
				else
					error(G_("unknown value supplied for option '-fplugin-arg-%s-%s=%s'"), plugin_name, argv[i].key, value);
				value = strtok_r(NULL, ",", &saveptr);
			}
			free(values);
			continue;
		}

		if (!strcmp(argv[i].key, "xor")) {
			enable_xor = true;
			continue;
		}

		if (!strcmp(argv[i].key, "retabort")) {
			if (!argv[i].value) {
				error(G_("no value supplied for option '-fplugin-arg-%s-%s'"), plugin_name, argv[i].key);
				continue;
			}

			rap_ret_abort = xstrdup(argv[i].value);
			continue;
		}

		if (!strcmp(argv[i].key, "xorabort")) {
			if (!argv[i].value) {
				error(G_("no value supplied for option '-fplugin-arg-%s-%s'"), plugin_name, argv[i].key);
				continue;
			}

			rap_xor_abort = xstrdup(argv[i].value);
			continue;
		}

		if (!strcmp(argv[i].key, "callabort")) {
			if (!argv[i].value) {
				error(G_("no value supplied for option '-fplugin-arg-%s-%s'"), plugin_name, argv[i].key);
				continue;
			}

			rap_call_abort = xstrdup(argv[i].value);
			continue;
		}

		if (!strcmp(argv[i].key, "hash")) {
			char *values, *value, *saveptr;

			if (!argv[i].value) {
				error(G_("no value supplied for option '-fplugin-arg-%s-%s'"), plugin_name, argv[i].key);
				continue;
			}

			values = xstrdup(argv[i].value);
			value = strtok_r(values, ",", &saveptr);
			while (value) {
				if (!strcmp(value, "abs"))
					enable_abs = enable_abs_finish = true;
				else if (!strcmp(value, "abs-finish"))
					enable_abs_finish = true;
				else if (!strcmp(value, "abs-ops"))
					enable_abs_ops = true;
				else if (!strcmp(value, "abs-attr"))
					enable_abs_attr = true;
//				else if (!strcmp(value, "const"))
//					imprecise_rap_hash_flags.qual_const = 1;
//				else if (!strcmp(value, "volatile"))
//					imprecise_rap_hash_flags.qual_volatile = 1;
				else
					error(G_("unknown value supplied for option '-fplugin-arg-%s-%s=%s'"), plugin_name, argv[i].key, value);
				value = strtok_r(NULL, ",", &saveptr);
			}
			free(values);
			continue;
		}

		if (!strcmp(argv[i].key, "report")) {
			char *values, *value, *saveptr;

			if (!argv[i].value) {
				error(G_("no value supplied for option '-fplugin-arg-%s-%s'"), plugin_name, argv[i].key);
				continue;
			}

			values = xstrdup(argv[i].value);
			value = strtok_r(values, ",", &saveptr);
			while (value) {
				if (!strcmp(value, "func"))
					report_func_hash = true;
				else if (!strcmp(value, "fptr"))
					report_fptr_hash = true;
				else if (!strcmp(value, "abs"))
					report_abs_hash = true;
				else
					error(G_("unknown value supplied for option '-fplugin-arg-%s-%s=%s'"), plugin_name, argv[i].key, value);
				value = strtok_r(NULL, ",", &saveptr);
			}
			free(values);
			continue;
		}

		if (!strcmp(argv[i].key, "include")) {
			if (argv[i].value) {
				error(G_("value supplied for option '-fplugin-arg-%s-%s=%s'"), plugin_name, argv[i].key, argv[i].value);
				continue;
			}

			rap_include = true;
			continue;
		}

		if (!strcmp(argv[i].key, "kernexec_method")) {
#ifdef TARGET_386
			if (TARGET_64BIT == 0)
				continue;

			if (!argv[i].value) {
				error(G_("no value supplied for option '-fplugin-arg-%s-%s'"), plugin_name, argv[i].key);
				continue;
			}

			if (!strcmp(argv[i].value, "bts") || !strcmp(argv[i].value, "\"bts\"")) {
				kernexec_instrument_fptr = kernexec_instrument_fptr_bts;
				kernexec_instrument_retaddr = kernexec_instrument_retaddr_bts;
			} else if (!strcmp(argv[i].value, "or") || !strcmp(argv[i].value, "\"or\"")) {
				kernexec_instrument_fptr = kernexec_instrument_fptr_or;
				kernexec_instrument_retaddr = kernexec_instrument_retaddr_or;
				fix_register("r12", 1, 1);
			} else
#endif
				error(G_("invalid option argument '-fplugin-arg-%s-%s=%s'"), plugin_name, argv[i].key, argv[i].value);
			continue;
		}

		if (!strcmp(argv[i].key, "sls")) {
			enable_sls = true;
			continue;
		}

		if (!strcmp(argv[i].key, "rethunk")) {
			enable_rethunk = true;
			continue;
		}

		if (!strcmp(argv[i].key, "rethooks")) {
			if (!argv[i].value) {
				error(G_("no value supplied for option '-fplugin-arg-%s-%s'"), plugin_name, argv[i].key);
				continue;
			}

			rethooks = atoi(argv[i].value);
			if (rethooks < 0 || rethooks > 2) {
				error(G_("invalid value supplied for option '-fplugin-arg-%s-%s', %d not in [0,2]"), plugin_name, argv[i].key, rethooks);
				continue;
			}
			continue;
		}

		error(G_("unknown option '-fplugin-arg-%s-%s'"), plugin_name, argv[i].key);
	}

	if ((enable_xor || enable_type_ret) && flag_stack_protect)
		flag_stack_protect = 0;

	if (flag_ipa_icf_functions) {
//		warning_at(UNKNOWN_LOCATION, 0, G_("-fipa-icf is incompatible with %s, disabling..."), plugin_name);
//		inform(UNKNOWN_LOCATION, G_("-fipa-icf is incompatible with %s, disabling..."), plugin_name);
		flag_ipa_icf_functions = 0;
	}

#if BUILDING_GCC_VERSION >= 7000
	if (flag_code_hoisting) {
//		warning_at(UNKNOWN_LOCATION, 0, G_("-fcode-hoisting is incompatible with %s, disabling..."), plugin_name);
//		inform(UNKNOWN_LOCATION, G_("-fcode-hoisting is incompatible with %s, disabling..."), plugin_name);
		flag_code_hoisting = 0;
	}
#endif

#if BUILDING_GCC_VERSION >= 8000
	if ((enable_xor || enable_type_call || enable_type_ret) && flag_cf_protection != CF_NONE) {
//		warning_at(UNKNOWN_LOCATION, 0, G_("-fcf-protection is incompatible with %s, disabling..."), plugin_name);
//		inform(UNKNOWN_LOCATION, G_("-fcf-protection is incompatible with %s, disabling..."), plugin_name);
		flag_cf_protection = CF_NONE;
	}

	// FIXME: somehow basic block reordering can leave a stray cfi_restore_state directive around
	// see block/partitions/ldm.c on i386
	if (flag_reorder_blocks_and_partition) {
//		warning_at(UNKNOWN_LOCATION, 0, G_("-flag-reorder-blocks-and-partition is incompatible with %s, disabling..."), plugin_name);
//		inform(UNKNOWN_LOCATION, G_("-freorder-blocks-and-partition is incompatible with %s, disabling..."), plugin_name);
		flag_reorder_blocks_and_partition = 0;
	}
#endif

	if (enable_type_call || enable_type_ret)
		flag_crossjumping = 0;

	if (enable_xor || enable_type_ret)
		rap_reserve_register_hook();

	register_callback(plugin_name, PLUGIN_INFO, NULL, &rap_plugin_info);
	register_callback(plugin_name, PLUGIN_ATTRIBUTES, register_attributes, NULL);

	if (enable_abs) {
		register_callback(plugin_name, PLUGIN_FINISH_DECL, rap_emit_hash_symbols_finish_decl, NULL);

		// this is (ab)used to implement a cross-plugin event mechanism due to https://gcc.gnu.org/bugzilla/show_bug.cgi?id=92217
		register_callback(plugin_name, PLUGIN_PRE_GENERICIZE, rap_emit_hash_symbols_finish_decl, NULL);
	}
	if (enable_abs_ops)
		register_callback(plugin_name, PLUGIN_FINISH_TYPE, rap_emit_hash_symbols_finish_type, NULL);

	if (enable_xor || enable_type_cast || enable_type_call || enable_type_ret) {
#if BUILDING_GCC_VERSION >= 9000
		register_callback(plugin_name, PLUGIN_FINISH_DECL, rap_handle_typedef_main_variant, NULL);
#endif
	}

	if (enable_xor || enable_type_call || enable_type_ret) {
		register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &rap_unignore_pass_info);
		register_callback(plugin_name, PLUGIN_REGISTER_GGC_ROOTS, NULL, (void *)&gt_ggc_r_rap);
		if (enable_abs_finish)
			register_callback(plugin_name, PLUGIN_FINISH_UNIT, rap_finish_unit, NULL);
	}

	if (enable_type_call || enable_type_ret)
		register_callback(plugin_name, PLUGIN_ALL_IPA_PASSES_END, rap_emit_nonweak_hash, NULL);

	if (enable_xor || enable_type_ret) {
		register_callback(plugin_name, PLUGIN_START_UNIT, rap_start_unit_ret, NULL);
		register_callback(plugin_name, PLUGIN_ALL_PASSES_START, rap_all_passes_start, NULL);
		register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &rap_ret_ipa_pass_info);
		register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &rap_ret_pass_info);
		register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &merge_phi_pass_info);
		register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &rap_rewrite_retaddr_check_pass_info);
	}

	if (enable_xor) {
		register_callback(plugin_name, PLUGIN_START_UNIT, rap_start_unit_xor, NULL);
		register_callback(plugin_name, PLUGIN_ALL_IPA_PASSES_END, check_global_register_variables, NULL);
		register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &rap_check_local_register_variables_pass_info);
		register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &rap_reload_pass_info);
		register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &rap_xor_elimination_pass_info);
	}

	if (enable_type_ret) {
		register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &rap_check_tail_calls_pass_info);
		register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &rap_mark_retloc_pass_info);
	}

	if (enable_type_nospec) {
#ifdef TARGET_386
# ifdef ix86_indirect_branch
#  ifndef ix86_indirect_branch_register
#   error unsupported GCC version as it lacks the full retpoline implementation
#  endif
		if (enable_type_call && ix86_indirect_branch != indirect_branch_keep)
			error(G_("only the -mindirect-branch=keep retpoline variant is supported"));
		// NB: if we ever stop replacing the original indirect jmp/call insns
		// then this will need to be enabled to prevent double instrumentation
		// ix86_indirect_branch = indirect_branch_keep;

		if (!ix86_indirect_branch_register)
			error(G_("-mindirect-branch-register must be enabled for retpoline support"));
# else /* !ix86_indirect_branch */
#  if BUILDING_GCC_VERSION < 6000
		flag_jump_tables = 0;
#  endif
		register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &rap_indirect_branch_register_pass_info);
# endif /* ix86_indirect_branch */

		register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &rap_retpoline_pass_info);
#elif defined(SPARC_DEFAULT_CMODEL)
#else
#error unsupported target
#endif
	}

	if (enable_sls) {
		if (!enable_type_nospec)
			rap_sls_pass_info.reference_pass_name = rap_retpoline_pass_info.reference_pass_name;
		register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &rap_sls_pass_info);
	}

	if (enable_rethunk) {
		if (!enable_sls)
			rap_rethunk_pass_info.reference_pass_name = rap_sls_pass_info.reference_pass_name;
		register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &rap_rethunk_pass_info);
	}

	if (enable_type_cast || enable_type_call || enable_type_ret) {
		if (!enable_xor && !enable_type_ret) {
			rap_fptr_pass_info.reference_pass_name = rap_ret_pass_info.reference_pass_name;
			rap_fptr_pass_info.pos_op = rap_ret_pass_info.pos_op;
		}
		register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &rap_fptr_pass_info);
		register_callback(plugin_name, PLUGIN_ALL_IPA_PASSES_START, rap_calculate_fptr_hashes, NULL);
	}

//	if (enable_type_call)
//		register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &rap_fptr_safety_pass_info);

	if (enable_type_cast) {
		register_callback(plugin_name, PLUGIN_ALL_IPA_PASSES_END, check_global_variable_casts, NULL);
		register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &rap_check_local_variable_casts_pass_info);
	}

	if (enable_abs_finish || enable_abs_ops || enable_abs_attr || enable_type_call || enable_type_ret || enable_type_cast) {
		old_builtin_function = lang_hooks.builtin_function;
		old_builtin_function_ext_scope = lang_hooks.builtin_function_ext_scope;
		lang_hooks.builtin_function = rap_builtin_function;
		lang_hooks.builtin_function_ext_scope = rap_builtin_function_ext_scope;
		// this must come last to be the first FINISH_TYPE callback in order to populate the function types with the rap hash
		register_callback(plugin_name, PLUGIN_FINISH_TYPE, rap_hash_finish_type, NULL);
		register_callback(plugin_name, PLUGIN_ALL_PASSES_START, rap_hash_type_cleanup, NULL);
	}

	if (enable_type_cast || enable_type_call || enable_type_ret) {
		register_callback(plugin_name, PLUGIN_FINISH_DECL, rap_hash_compute_attr, NULL);

		// this is (ab)used to implement a cross-plugin event mechanism due to https://gcc.gnu.org/bugzilla/show_bug.cgi?id=92217
		register_callback(plugin_name, PLUGIN_PRE_GENERICIZE, rap_hash_compute_attr, NULL);

		register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &rap_indirectly_called_pass_info);
		register_callback(plugin_name, PLUGIN_ALL_IPA_PASSES_START, rap_calculate_func_hashes, NULL);
		register_callback(plugin_name, PLUGIN_ALL_IPA_PASSES_END, rap_calculate_func_hashes, NULL);
	}

	if (kernexec_instrument_fptr == kernexec_instrument_fptr_or)
		register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &kernexec_reload_pass_info);
	if (kernexec_instrument_fptr)
		register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &kernexec_fptr_pass_info);
	if (kernexec_instrument_retaddr)
		register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &kernexec_retaddr_pass_info);

	register_callback(plugin_name, PLUGIN_START_UNIT, rap_start_unit_common, NULL);

	return 0;
}
