/*
 * Copyright 2012-2024 by PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 *
 * Homepage: http://pax.grsecurity.net/
 */

#include "rap.h"

static bitmap_obstack complex_functions_obstack;
bitmap complex_functions;
static struct cgraph_node_hook_list *function_insertion_hook_holder;
static struct cgraph_node_hook_list *node_removal_hook_holder;
static struct cgraph_2node_hook_list *node_duplication_hook_holder;
static bool after_execute;

struct postorder_stack
{
	cgraph_node_ptr node;
	cgraph_edge_p edge;
	int ref;
};

static bool ipa_reverse_postorder_ignore_node(cgraph_node_ptr node, int pass)
{
	if (node->aux)
		return true;

	if (pass)
		return false;

	if (node->address_taken)
		return true;

	if (NODE_GLOBAL(node)->inlined_to)
		return true;

	if (node->alias || NODE_THUNK(node))
		return true;

	if (cgraph_only_called_directly_p(node))
		return true;

	return false;
}

static int ipa_reverse_postorder_push_node(struct postorder_stack *stack, int stack_size, cgraph_node_ptr node)
{
	++stack_size;
	stack[stack_size].node = node;
	stack[stack_size].edge = node->callers;
	stack[stack_size].ref = 0;
	node->aux = (void *)(size_t)1;
	return stack_size;
}

static int rap_ipa_reverse_postorder_1(cgraph_node_ptr node, cgraph_node_ptr *order, struct postorder_stack *stack, int order_pos, int pass)
{
	int stack_size;

	if (ipa_reverse_postorder_ignore_node(node, pass))
		return order_pos;

	stack_size = ipa_reverse_postorder_push_node(stack, -1, node);
	while (stack_size >= 0) {
		cgraph_node_ptr node2 = NULL;
		struct ipa_ref *ref = NULL;

		while (stack[stack_size].edge && !node2) {
			cgraph_edge_p edge;

			edge = stack[stack_size].edge;
			node2 = edge->caller;

			stack[stack_size].edge = edge->next_caller;
			if (!DECL_DISREGARD_INLINE_LIMITS(edge->caller->decl) && DECL_DISREGARD_INLINE_LIMITS(cgraph_function_node(edge->callee, NULL)->decl))
				node2 = NULL;
		}

		for (; ipa_ref_list_referring_iterate(&stack[stack_size].node->ref_list, stack[stack_size].ref, ref) && !node2; stack[stack_size].ref++)
			if (ref->use == IPA_REF_ALIAS)
				node2 = ipa_ref_referring_cgraph_node(ref);

		if (!node2)
			order[order_pos++] = stack[stack_size--].node;
		else if (!node2->aux)
			stack_size = ipa_reverse_postorder_push_node(stack, stack_size, node2);
	}

	return order_pos;
}

static int rap_ipa_reverse_postorder(cgraph_node_ptr *order)
{
	cgraph_node_ptr node;
	int order_pos = 0;
	struct postorder_stack *stack = XCNEWVEC(struct postorder_stack, cgraph_n_nodes);

	FOR_EACH_FUNCTION(node)
		node->aux = NULL;

	FOR_EACH_FUNCTION(node)
		order_pos = rap_ipa_reverse_postorder_1(node, order, stack, order_pos, 0);

	FOR_EACH_FUNCTION(node)
		order_pos = rap_ipa_reverse_postorder_1(node, order, stack, order_pos, 1);

	FOR_EACH_FUNCTION(node)
		node->aux = NULL;

	free(stack);

	return order_pos;
}

static bool rap_ret_ipa_is_complex_function(void)
{
	basic_block bb;

	if (!rap_cmodel_check())
		return false;

	if (rap_noreturn_function())
		return false;

	if (!rap_is_complex_function(NULL_TREE))
		return false;

	FOR_EACH_BB_FN(bb, cfun) {
		if (rap_ret_gate_bb(bb, ignore_all))
			return true;
	}

	return false;
}

static void rap_add_new_function(cgraph_node_ptr node, void *data __unused)
{
	gcc_assert(!in_lto_p || !after_execute);
	debug_cgraph_node(node);
	gcc_unreachable();
}

static void rap_remove_node_data(cgraph_node_ptr node, void *data __unused)
{
	if (bitmap_bit_p(complex_functions, NODE_UID(node)))
		bitmap_clear_bit(complex_functions, NODE_UID(node));
}

static void rap_duplicate_node_data(cgraph_node_ptr src, cgraph_node_ptr dst, void *data __unused)
{
	gcc_assert(!in_lto_p || !after_execute);
	gcc_assert(!bitmap_bit_p(complex_functions, NODE_UID(dst)));
	gcc_assert(NODE_UID(src) != NODE_UID(dst));

	if (bitmap_bit_p(complex_functions, NODE_UID(src)))
		bitmap_set_bit(complex_functions, NODE_UID(dst));
}

static void rap_ret_register_hooks(void)
{
	static bool init_p = false;

	if (init_p)
		return;

	init_p = true;

	// XXX gcc 5-7 are buggy here, see https://gcc.gnu.org/bugzilla/show_bug.cgi?id=82256
	function_insertion_hook_holder = cgraph_add_function_insertion_hook(&rap_add_new_function, NULL);

	node_removal_hook_holder = cgraph_add_node_removal_hook(&rap_remove_node_data, NULL);
	node_duplication_hook_holder = cgraph_add_node_duplication_hook(&rap_duplicate_node_data, NULL);
}

static void rap_ret_ipa_generate_summary(void)
{
	cgraph_node_ptr node;

	gcc_assert(!complex_functions);

	bitmap_obstack_initialize(&complex_functions_obstack);
	complex_functions = BITMAP_ALLOC(&complex_functions_obstack);
	bitmap_clear(complex_functions);

	rap_ret_register_hooks();

	FOR_EACH_FUNCTION(node) {
		gcc_assert(!NODE_GLOBAL(node)->inlined_to);
		if (cgraph_function_body_availability(node) < AVAIL_INTERPOSABLE) {
			gcc_assert(NODE_DECL(node));
			if (rap_is_complex_function(NODE_DECL(node)))
				bitmap_set_bit(complex_functions, NODE_UID(node));
			continue;
		}

		if (NODE_THUNK(node) || node->alias)
			continue;

		gcc_assert(DECL_STRUCT_FUNCTION(NODE_DECL(node)));
		set_current_function_decl(NODE_DECL(node));
		if (rap_ret_ipa_is_complex_function())
			bitmap_set_bit(complex_functions, NODE_UID(node));
		unset_current_function_decl();
	}
}

static unsigned int rap_ret_ipa_execute(void)
{
	cgraph_node_ptr *order;
	int i, norder;

	after_execute = true;

	order = XCNEWVEC(cgraph_node_ptr, cgraph_n_nodes);
	norder = rap_ipa_reverse_postorder(order);

	for (i = norder - 1; i >= 0; i--) {
		cgraph_node_ptr node = order[i];
		cgraph_edge_p edge;

		if (bitmap_bit_p(complex_functions, NODE_UID(node)))
			continue;

		if (cgraph_function_body_availability(node) < AVAIL_INTERPOSABLE) {
			gcc_assert(!NODE_GLOBAL(node)->inlined_to);
			gcc_assert(NODE_DECL(node));
			if (rap_is_complex_function(NODE_DECL(node)))
				bitmap_set_bit(complex_functions, NODE_UID(node));
			continue;
		}

		if (NODE_THUNK(node) || node->alias)
			continue;

		if (node->indirect_calls) {
			bitmap_set_bit(complex_functions, NODE_UID(node));
			continue;
		}

		if (!node->callees)
			continue;

		for (edge = node->callees; edge; edge = edge->next_callee) {
			if (bitmap_bit_p(complex_functions, NODE_UID(edge->callee))) {
				bitmap_set_bit(complex_functions, NODE_UID(node));
				break;
			}

			gcc_assert(NODE_DECL(edge->callee));
		}

	}

	free(order);
	return 0;
}

#define PASS_NAME rap_ret_ipa
#define NO_READ_SUMMARY
#define NO_WRITE_SUMMARY
#define NO_READ_OPTIMIZATION_SUMMARY
#define NO_WRITE_OPTIMIZATION_SUMMARY
#define NO_STMT_FIXUP
#define NO_FUNCTION_TRANSFORM
#define NO_VARIABLE_TRANSFORM
#define NO_GATE
#include "gcc-generate-ipa-pass.h"
