/*
 * Copyright 2012-2024 by PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 *
 * Homepage: http://pax.grsecurity.net/
 */

#include "rap.h"

bool rap_cmodel_check(void)
{
#ifdef TARGET_386
	if (!TARGET_64BIT || ix86_cmodel != CM_KERNEL)
		return true;

	// early startup code runs at physical addresses, not virtual
	if (has_trait(current_function_decl, "phys"))
		return false;

	return true;
#elif defined(TARGET_ARM) || defined(TARGET_THUMB)
#error unsupported target
	return true;
#elif defined(AARCH64_ABI_DEFAULT)
#error unsupported target
	return true;
#elif defined(SPARC_DEFAULT_CMODEL)
	return true;
#else
#error unsupported target
#endif
}

unsigned int rap_xor_checks[2];
void rap_all_passes_start(void *event_data __unused, void *data __unused)
{
	gcc_assert(current_function_decl);

	rap_xor_checks[0] = 0;
	rap_xor_checks[1] = 0;
}

bool rap_is_complex_function(tree fndecl)
{
	if (fndecl == NULL_TREE)
		fndecl = current_function_decl;
	else if (fndecl == current_function_decl)
		return false;

	if (is_simple_builtin(fndecl)) {
		if (!TREE_ASM_WRITTEN(fndecl))
			return false;
		debug_tree(fndecl);
		gcc_unreachable();
	}

	if (DECL_PURE_P(fndecl))
		return false;

	if (TREE_READONLY(fndecl))
		return false;

	if (TREE_THIS_VOLATILE(fndecl))
		return false;

	return true;
}

static bool rap_noreturn_bb(basic_block bb)
{
	edge e;
	edge_iterator ei;

	gcc_assert(bb);

	if (!EDGE_COUNT(bb->succs))
		return true;

	FOR_EACH_EDGE(e, ei, bb->succs) {
		if (!(e->flags & EDGE_FAKE))
			return false;
	}

	return true;
}

bool rap_calls_complex_function_p(tree fndecl) {
	cgraph_node_ptr cnode;
	cgraph_edge_p edge;

	cnode = cgraph_get_node(fndecl);
	gcc_assert(cnode);

	for (edge = cnode->callees; edge; edge = edge->next_callee) {
		basic_block bb;

		bb = gimple_bb(edge->call_stmt);
		if (bb && rap_noreturn_bb(bb))
			continue;

		gcc_assert(NODE_DECL(edge->callee));
		if (rap_is_complex_function(NODE_DECL(edge->callee)))
			return true;
	}

	return false;
}

static bool rap_is_complex_access(tree arg)
{
	tree ref, offset;
	poly_int64_pod bitsize, bitpos;
	enum machine_mode mode;
	int unsignedp, reversep, volatilep;

	ref = get_inner_reference(arg, &bitsize, &bitpos, &offset, &mode, &unsignedp, &reversep, &volatilep, true);
	switch (TREE_CODE(ref)) {
	default:
		break;

	case RESULT_DECL:
		return false;

	case FUNCTION_DECL:
	case COMPONENT_REF:
	case ARRAY_REF:
	case ADDR_EXPR:
		if (!offset)
			return false;
		break;

	case PARM_DECL:
		if (!offset)
			return false;

		switch (TREE_CODE(offset)) {
		default:
			break;

		case MULT_EXPR:
			return true;
		}
		break;

	case VAR_DECL:
		if (!offset)
			return false;

		switch (TREE_CODE(offset)) {
		default:
			break;

		case MULT_EXPR:
		case PLUS_EXPR:
		case NOP_EXPR:
		case NON_LVALUE_EXPR:
		case SSA_NAME:
			return true;
		}
		break;

	case INDIRECT_REF:
	case TARGET_MEM_REF:
	case MEM_REF:
		return true;
	}

	debug_tree(ref);
	debug_tree(offset);
	gcc_unreachable();
}

static bool rap_is_complex_asm_arg(tree arg, bool input)
{
	switch (TREE_CODE(arg)) {
	default:
		debug_tree(arg);
		gcc_unreachable();

	case SSA_NAME:
		return TREE_CODE(TREE_TYPE(arg)) == POINTER_TYPE;

	case ADDR_EXPR:
		return rap_is_complex_access(arg);

	case ARRAY_REF:
		return !input;

	case PARM_DECL:
	case REAL_CST:
	case RESULT_DECL:
	case VAR_DECL:
	case VECTOR_CST:
		return false;

	case MEM_REF:
	case TARGET_MEM_REF:
		return true;

	case COMPONENT_REF:
		return rap_is_complex_access(arg);

	case INTEGER_CST:
		return false;
	}
}

static bool rap_is_complex_asm(gasm *stmt)
{
	const char *asm_string = gimple_asm_string(stmt);
	size_t ninputs = gimple_asm_ninputs(stmt);
	size_t noutputs = gimple_asm_noutputs(stmt);
	unsigned int i;

	if (!*asm_string)
		return false;

	if (strstr(asm_string, "rap_safe_asm"))
		return false;

	if (!regexec(&asm_call_insn_regex, asm_string, 0, NULL, 0))
		return true;

	if (ninputs == 1 && noutputs == 1) {
		tree input = TREE_VALUE(TREE_PURPOSE(gimple_asm_input_op(stmt, 0)));
		tree output = TREE_VALUE(TREE_PURPOSE(gimple_asm_output_op(stmt, 0)));

		if (!strcmp(TREE_STRING_POINTER(input), "p") && !strcmp(TREE_STRING_POINTER(output), "=r"))
			return false;
	}

	if (gimple_asm_clobbers_memory_p(stmt))
		return true;

	for (i = 0; i < ninputs; i++) {
		tree arg = TREE_VALUE(gimple_asm_input_op(stmt, i));

		if (rap_is_complex_asm_arg(arg, true))
			return true;
	}

	for (i = 0; i < noutputs; i++) {
		tree arg = TREE_VALUE(gimple_asm_output_op(stmt, i));

		if (rap_is_complex_asm_arg(arg, false))
			return true;
	}

	return false;
}

bool rap_ret_gate_bb(basic_block bb, enum rap_ret_gate_ignore_calls ignore_calls)
{
	gimple_stmt_iterator gsi;

	if (rap_noreturn_bb(bb))
		return false;

	for (gsi = gsi_start_bb(bb); !gsi_end_p(gsi); gsi_next(&gsi)) {
		gimple_stmt_ptr stmt;
		tree lhs, rhs1;

		stmt = gsi_stmt(gsi);
		if (is_gimple_call(stmt)) {
			tree fndecl;
			gcall *call_stmt;

			if (ignore_calls == ignore_all)
				continue;

			call_stmt = as_a_gcall(stmt);
			if (ignore_calls == ignore_tail && gimple_call_tail_p(call_stmt))
				continue;

			fndecl = gimple_call_fndecl(call_stmt);
			if (!fndecl)
				return true;

			if (!cgraph_get_node(fndecl)) {
				return true;
			}

			gcc_assert(complex_functions);
			if (bitmap_bit_p(complex_functions, NODE_UID(cgraph_get_node(fndecl))))
				return true;
			continue;
		}

		if (gimple_code(stmt) == GIMPLE_ASM) {
			if (rap_is_complex_asm(as_a_gasm(stmt)))
				return true;
			continue;
		}

		if (!is_gimple_assign(stmt))
			continue;
		if (gimple_clobber_p(stmt))
			continue;
		if (!gimple_store_p(stmt))
			continue;
		if (!gimple_vdef(stmt))
			continue;

		lhs = gimple_get_lhs(stmt);
		rhs1 = gimple_assign_rhs1(stmt);

		if (TREE_CODE(lhs) == PARM_DECL)
			continue;

		if (TREE_CODE(lhs) == RESULT_DECL)
			continue;

		if (VAR_P(lhs) && is_global_var(lhs))
			continue;

		if (VAR_P(lhs) && TREE_CODE(rhs1) == CONSTRUCTOR)
			continue;

		if (rap_is_complex_access(lhs))
			return true;
		continue;
	}

	return false;
}

bool rap_noreturn_function(void)
{
	edge e;
	edge_iterator ei;

	if (TREE_THIS_VOLATILE(current_function_decl))
		return true;

	if (!EDGE_COUNT(EXIT_BLOCK_PTR_FOR_FN(cfun)->preds))
		return true;

	FOR_EACH_EDGE(e, ei, EXIT_BLOCK_PTR_FOR_FN(cfun)->preds) {
		if (!(e->flags & EDGE_FAKE))
			return false;
	}

	return true;
}

bool rap_ret_gate(void)
{
	// handle -fdump-passes
	if (!cfun->cfg)
		return true;

	if (!rap_cmodel_check())
		return false;

	if (rap_noreturn_function())
		return false;

	if (lookup_attribute("naked", DECL_ATTRIBUTES(current_function_decl)))
		return false;

	if (enable_xor || enable_type_ret) {
		gcc_assert(complex_functions);
		return bitmap_bit_p(complex_functions, NODE_UID(cgraph_get_node(current_function_decl)));
	}

	return true;
}

static gimple_stmt_ptr create_assign(tree dst, tree src)
{
	gimple_stmt_ptr stmt;

	dst = make_ssa_name(dst, NULL);
	stmt = gimple_build_assign(dst, src);
	SSA_NAME_DEF_STMT(dst) = stmt;
	return stmt;
}

static bool rap_stmt_ends_bb_p(gimple_stmt_ptr stmt)
{
	if (stmt_ends_bb_p(stmt))
		return true;

	if (!is_gimple_call(stmt))
		return false;

	return gimple_call_tail_p(as_a_gcall(stmt));
}

/*
 * insert the equivalent of
 * unsigned long encoded_retaddr = (unsigned long)__builtin_return_address(0) ^ cookie;
 */
static tree save_encoded_retaddr(gimple_seq *stmts)
{
	gimple_stmt_ptr stmt;
	tree retaddr_ptr, retaddr_ulong, cookie;

	retaddr_ptr = create_tmp_var(ptr_type_node, "rap_retaddr_entry_ptr");
	stmt = gimple_build_call(builtin_decl_implicit(BUILT_IN_RETURN_ADDRESS), 1, integer_zero_node);
	retaddr_ptr = make_ssa_name(retaddr_ptr, stmt);
	gimple_call_set_lhs(stmt, retaddr_ptr);
	gimple_seq_add_stmt(stmts, stmt);

	retaddr_ulong = create_tmp_var(long_unsigned_type_node, "rap_retaddr_entry_ulong");
	stmt = create_assign(retaddr_ulong, fold_convert(long_unsigned_type_node, retaddr_ptr));
	retaddr_ulong = gimple_get_lhs(stmt);
	gimple_seq_add_stmt(stmts, stmt);

	cookie = create_tmp_var(long_unsigned_type_node, "rap_cookie_entry");
	stmt = create_assign(cookie, rap_cookie_decl);
	cookie = gimple_get_lhs(stmt);
	gimple_seq_add_stmt(stmts, stmt);

	stmt = gimple_build_assign_with_ops(BIT_XOR_EXPR, SSA_NAME_VAR(retaddr_ulong), retaddr_ulong, cookie);
	retaddr_ulong = create_tmp_var(long_unsigned_type_node, "rap_retaddr_entry_encoded");
	retaddr_ulong = make_ssa_name(retaddr_ulong, stmt);
	gimple_assign_set_lhs(stmt, retaddr_ulong);
	gimple_seq_add_stmt(stmts, stmt);

	stmt = gimple_barrier(retaddr_ulong, true);
	gimple_seq_add_stmt(stmts, stmt);

	stmt = gimple_barrier(rap_cookie_decl, true);
	gimple_seq_add_stmt(stmts, stmt);

	return retaddr_ulong;
}

/*
 * insert the equivalent of
 * *old_retaddr ^= cookie;
 * return (unsigned long)__builtin_return_address(0);
 */
static tree get_decoded_retaddr(gimple_seq *stmts, tree *old_retaddr)
{
	gimple_stmt_ptr stmt;
	tree retaddr_ptr, retaddr_ulong, cookie;

	stmt = gimple_barrier(NULL_TREE, true);
	gimple_seq_add_stmt(stmts, stmt);

	if (*old_retaddr == NULL_TREE)
		goto skip_decoding;

	cookie = create_tmp_var(long_unsigned_type_node, "rap_cookie_exit");
	stmt = create_assign(cookie, rap_cookie_decl);
	cookie = gimple_get_lhs(stmt);
	gimple_seq_add_stmt(stmts, stmt);

	stmt = gimple_build_assign_with_ops(BIT_XOR_EXPR, SSA_NAME_VAR(*old_retaddr), *old_retaddr, cookie);
	retaddr_ulong = create_tmp_var(long_unsigned_type_node, "rap_retaddr_exit_decoded");
	retaddr_ulong = make_ssa_name(retaddr_ulong, stmt);
	gimple_assign_set_lhs(stmt, retaddr_ulong);
	gimple_seq_add_stmt(stmts, stmt);

	*old_retaddr = create_tmp_var(ptr_type_node, "rap_retaddr_exit_ptr_old");
	stmt = create_assign(*old_retaddr, fold_convert(ptr_type_node, retaddr_ulong));
	*old_retaddr = gimple_get_lhs(stmt);
	gimple_seq_add_stmt(stmts, stmt);

skip_decoding:
	retaddr_ptr = create_tmp_var(ptr_type_node, "rap_retaddr_exit_ptr");
	stmt = gimple_build_call(builtin_decl_implicit(BUILT_IN_RETURN_ADDRESS), 1, integer_zero_node);
	retaddr_ptr = make_ssa_name(retaddr_ptr, stmt);
	gimple_call_set_lhs(stmt, retaddr_ptr);
	gimple_seq_add_stmt(stmts, stmt);

	return retaddr_ptr;
}

static void check_decoded_retaddr(gimple_stmt_iterator gsi, tree old_retaddr, tree new_retaddr, bool tailcall)
{
	gimple_stmt_ptr stmt;
	location_t loc;
	basic_block cond_bb, join_bb, true_bb;
	edge e;
	tree block;

	gcc_assert(!gsi_end_p(gsi));
	loc = gimple_location(gsi_stmt(gsi));
	block = gimple_block(gsi_stmt(gsi));

	gcc_assert(!gsi_end_p(gsi));

	if (old_retaddr != NULL_TREE) {
		stmt = gimple_build_cond(NE_EXPR, old_retaddr, new_retaddr, NULL_TREE, NULL_TREE);
		gimple_set_location(stmt, loc);
		gimple_set_block(stmt, block);
		gsi_insert_after(&gsi, stmt, GSI_CONTINUE_LINKING);

		cond_bb = gimple_bb(gsi_stmt(gsi));
		gcc_assert(!gsi_end_p(gsi));
		gcc_assert(stmt == gsi_stmt(gsi));

		e = split_block(cond_bb, gsi_stmt(gsi));
		gcc_assert(cond_bb == e->src);
		join_bb = e->dest;
		e->flags = EDGE_FALSE_VALUE;
		e->probability -= REG_BR_PROB_VERY_UNLIKELY;

		true_bb = create_empty_bb(join_bb);
		e = make_edge(cond_bb, true_bb, EDGE_TRUE_VALUE | EDGE_PRESERVE);
		e->probability = REG_BR_PROB_VERY_UNLIKELY;
		true_bb->count = edge_count(e);

		gcc_assert(dom_info_available_p(CDI_DOMINATORS));
		set_immediate_dominator(CDI_DOMINATORS, true_bb, cond_bb);
		set_immediate_dominator(CDI_DOMINATORS, join_bb, cond_bb);

		gcc_assert(cond_bb->loop_father == join_bb->loop_father);
		add_bb_to_loop(true_bb, cond_bb->loop_father);
		loops_state_set(LOOPS_NEED_FIXUP);

		gsi = gsi_start_bb(true_bb);

		if (rap_xor_abort) {
			tree input;
			vec<tree, va_gc> *inputs = NULL;

			// asm volatile(rap_xor_abort %0 : : "r"(new_retaddr));
			input = build_tree_list(NULL_TREE, build_const_char_string(2, "r"));
			input = chainon(NULL_TREE, build_tree_list(input, new_retaddr));
			VEC_safe_push(tree, gc, inputs, input);
			stmt = gimple_build_asm_vec(rap_xor_abort, inputs, NULL, NULL, NULL);
			gimple_asm_set_volatile(as_a_gasm(stmt), true);
			gimple_set_location(stmt, loc);
			gimple_set_block(stmt, block);
			gsi_insert_after(&gsi, stmt, GSI_CONTINUE_LINKING);

			stmt = gimple_build_builtin_unreachable(loc);
		} else
			stmt = gimple_build_builtin_trap(loc);

		gimple_set_block(stmt, block);
		gsi_insert_after(&gsi, stmt, GSI_CONTINUE_LINKING);

		gsi = gsi_start_nondebug_after_labels_bb(join_bb);

		if (tailcall)
			rap_xor_checks[1]++;
		else
			rap_xor_checks[0]++;
	}

	if (enable_type_ret) {
		gimple_seq stmts = NULL;
		tree target_hash, computed_hash;
		rap_hash_t hash;

		target_hash = get_rap_hash(&stmts, loc, new_retaddr, tree_to_shwi(RAP_RET_HASH_OFFSET));

		if (gsi_end_p(gsi))
			gsi_insert_seq_after(&gsi, stmts, GSI_CONTINUE_LINKING);
		else {
			if (rap_stmt_ends_bb_p(gsi_stmt(gsi))) {
				gsi_insert_seq_before(&gsi, stmts, GSI_SAME_STMT);
				gsi_prev(&gsi);
			} else
				gsi_insert_seq_after(&gsi, stmts, GSI_CONTINUE_LINKING);
		}

		hash = rap_lookup_imprecise_rap_hash(current_function_decl);
		computed_hash = build_int_cst_type(rap_hash_type_node, -hash.hash);

		stmt = gimple_build_cond(NE_EXPR, target_hash, computed_hash, NULL_TREE, NULL_TREE);
		gimple_set_location(stmt, loc);
		gimple_set_block(stmt, block);
		gsi_insert_after(&gsi, stmt, GSI_CONTINUE_LINKING);

		cond_bb = gimple_bb(gsi_stmt(gsi));
		gcc_assert(!gsi_end_p(gsi));
		gcc_assert(stmt == gsi_stmt(gsi));

		e = split_block(cond_bb, gsi_stmt(gsi));
		gcc_assert(cond_bb == e->src);
		join_bb = e->dest;
		e->flags = EDGE_FALSE_VALUE;
		e->probability -= REG_BR_PROB_VERY_UNLIKELY;

		true_bb = create_empty_bb(join_bb);
		e = make_edge(cond_bb, true_bb, EDGE_TRUE_VALUE | EDGE_PRESERVE);
		e->probability = REG_BR_PROB_VERY_UNLIKELY;
		true_bb->count = edge_count(e);

		gcc_assert(dom_info_available_p(CDI_DOMINATORS));
		set_immediate_dominator(CDI_DOMINATORS, true_bb, cond_bb);
		set_immediate_dominator(CDI_DOMINATORS, join_bb, cond_bb);

		gcc_assert(cond_bb->loop_father == join_bb->loop_father);
		add_bb_to_loop(true_bb, cond_bb->loop_father);
		loops_state_set(LOOPS_NEED_FIXUP);

		gsi = gsi_start_bb(true_bb);

		if (rap_ret_abort) {
			tree input;
			vec<tree, va_gc> *inputs = NULL;

			// asm volatile(rap_ret_abort %0 : : "r"(new_retaddr));
			input = build_tree_list(NULL_TREE, build_const_char_string(2, "r"));
			input = chainon(NULL_TREE, build_tree_list(input, new_retaddr));
			VEC_safe_push(tree, gc, inputs, input);
			stmt = gimple_build_asm_vec(rap_ret_abort, inputs, NULL, NULL, NULL);
			gimple_asm_set_volatile(as_a_gasm(stmt), true);
			gimple_set_location(stmt, loc);
			gimple_set_block(stmt, block);
			gsi_insert_after(&gsi, stmt, GSI_CONTINUE_LINKING);

			stmt = gimple_build_builtin_unreachable(loc);
		} else
			stmt = gimple_build_builtin_trap(loc);

		gimple_set_location(stmt, loc);
		gimple_set_block(stmt, block);
		gsi_insert_after(&gsi, stmt, GSI_CONTINUE_LINKING);

		gsi = gsi_start_nondebug_after_labels_bb(join_bb);
	}
}

/*
 * insert the equivalent of
 * if ((void *)(encoded_retaddr ^ cookie) != __builtin_return_address(0)) abort();
 * if (*(long *)((void *)decoded_retaddr+N) != (long)-function_hash) abort();
 */
static void check_retaddr(gimple_stmt_iterator gsi, tree old_retaddr, bool tailcall)
{
	tree new_retaddr;
	gimple_seq stmts = NULL;
	location_t loc;

	if (gsi_end_p(gsi)) {
		gimple_stmt_ptr reference = NULL;
		basic_block reference_bb = gsi_bb(gsi);

		while (!reference && single_succ_p(reference_bb)) {
			gimple_stmt_iterator gsi2;

			reference_bb = single_succ(reference_bb);
			gsi2 = gsi_start_nondebug_after_labels_bb(reference_bb);
			if (!gsi_end_p(gsi2))
				reference = gsi_stmt(gsi2);
		}
		gcc_assert(reference);
		loc = gimple_location(reference);
	} else
		loc = gimple_location(gsi_stmt(gsi));

	new_retaddr = get_decoded_retaddr(&stmts, &old_retaddr);
	gimple_seq_set_location(stmts, loc);

	if (gsi_end_p(gsi)) {
		gsi = gsi_start_nondebug_after_labels_bb(gsi_bb(gsi));
		gsi_insert_seq_after(&gsi, stmts, GSI_CONTINUE_LINKING);
		gcc_assert(gsi_stmt(gsi) == gsi_stmt(gsi_last_nondebug_bb(gsi_bb(gsi))));
	} else {
		if (rap_stmt_ends_bb_p(gsi_stmt(gsi))) {
			gsi_insert_seq_before(&gsi, stmts, GSI_SAME_STMT);
			gsi_prev(&gsi);
		} else {
			gsi_insert_seq_after(&gsi, stmts, GSI_CONTINUE_LINKING);
		}
	}

	check_decoded_retaddr(gsi, old_retaddr, new_retaddr, tailcall);
}

static gimple_stmt_ptr last_nonreturn_stmt(basic_block bb)
{
	gimple_stmt_iterator gsi;
	gimple_stmt_ptr stmt;

	gsi = gsi_last_nondebug_bb(bb);
	if (gsi_end_p(gsi))
		return NULL;

	stmt = gsi_stmt(gsi);
	if (gimple_code(stmt) != GIMPLE_RETURN)
		return stmt;

	gsi_prev_nondebug(&gsi);
	return gsi_end_p(gsi) ? NULL : gsi_stmt(gsi);
}

static bool is_real_tail_call(tree retval, gcall *call_stmt)
{
	if (retval)
		return true;

	if (TREE_CODE(TREE_TYPE(TREE_TYPE(current_function_decl))) == VOID_TYPE)
		return true;

	if (TREE_TYPE(TREE_TYPE(current_function_decl)) == gimple_call_return_type(call_stmt))
		return true;

	return false;
}

static bool check_tail_calls(basic_block bb_prologue, basic_block bb_epilogue, tree old_retaddr)
{
	basic_block bb;
	bool handled = false;

	FOR_EACH_BB_FN(bb, cfun) {
		gimple_stmt_ptr stmt;
		gimple_stmt_iterator gsi;

		if (!dominated_by_p(CDI_DOMINATORS, bb, bb_prologue))
			continue;
		if (!dominated_by_p(CDI_POST_DOMINATORS, bb, bb_epilogue))
			continue;

		gsi = gsi_last_nondebug_bb(bb);
		if (gsi_end_p(gsi))
			continue;

		stmt = gsi_stmt(gsi);
		if (is_gimple_call(stmt)) {
			gcall *call_stmt;

			call_stmt = as_a_gcall(stmt);
			if (!gimple_call_tail_p(call_stmt))
				continue;
			gcc_assert(is_real_tail_call(gimple_call_lhs(stmt), call_stmt));
		} else {
			greturn *ret_stmt;
			gcall *call_stmt;
			tree retval;

			if (gimple_code(stmt) != GIMPLE_RETURN)
				continue;

			ret_stmt = as_a_greturn(stmt);
			retval = gimple_return_retval(ret_stmt);

			gsi_prev_nondebug(&gsi);
			if (gsi_end_p(gsi))
				continue;

			stmt = gsi_stmt(gsi);
			if (!is_gimple_call(stmt))
				continue;

			call_stmt = as_a_gcall(stmt);
			if (!gimple_call_tail_p(call_stmt))
				continue;

			if (!is_real_tail_call(retval, call_stmt)) {
				gimple_call_set_tail(call_stmt, false);
				gsi_next_nondebug(&gsi);
				gcc_assert(ret_stmt == as_a_greturn(gsi_stmt(gsi)));
			} else if (bb == bb_epilogue)
				gcc_assert(!handled);
		}

		if (bb == bb_epilogue)
			handled = true;
		else if (bb_epilogue == EXIT_BLOCK_PTR_FOR_FN(cfun) && single_succ_p(bb) && single_succ(bb) == bb_epilogue)
			handled = true;

		gcc_assert(!bb_any_loop(bb));
		check_retaddr(gsi, old_retaddr, true);
	}

	return handled;
}

static basic_block get_rap_ret_entry_block(bitmap complex_blocks) {
	basic_block bb = nearest_common_dominator_for_set(CDI_DOMINATORS, complex_blocks);

	while (bb != ENTRY_BLOCK_PTR_FOR_FN(cfun) && bb_any_loop(bb))
		bb = get_immediate_dominator(CDI_DOMINATORS, bb);

	return bb;
}

static basic_block get_rap_ret_exit_block(bitmap complex_blocks) {
	basic_block bb = nearest_common_dominator_for_set(CDI_POST_DOMINATORS, complex_blocks);

	while (bb != EXIT_BLOCK_PTR_FOR_FN(cfun) && bb_any_loop(bb))
		bb = get_immediate_dominator(CDI_POST_DOMINATORS, bb);

	return bb;
}

static bool can_insert_forwarder_p(enum cdi_direction dir, basic_block bb, basic_block dom)
{
	edge e;
	edge_iterator ei;

	if (dominated_by_p(dir, bb, dom))
		return false;

	if (dir == CDI_DOMINATORS) {
		FOR_EACH_EDGE(e, ei, bb->preds) {
			if (dominated_by_p(dir, e->src, dom))
				return true;
		}
	} else {
		FOR_EACH_EDGE(e, ei, bb->succs) {
			if (!dominated_by_p(dir, e->dest, dom))
				return true;
		}
	}

	return false;
}

static basic_block bb_pro;
static bool rap_redirect_epilogue_edge_p(edge e)
{
	return dominated_by_p(CDI_DOMINATORS, e->src, bb_pro);
}

static unsigned int rap_ret_execute(void)
{
	basic_block bb, bb_prologue, bb_epilogue, bb_single_entry, bb_single_exit;
	bitmap complex_blocks;
	tree old_retaddr = NULL_TREE;
	gimple_seq stmts = NULL;
	gimple_stmt_ptr stmt;
	bool changed_prologue, changed_epilogue;

	loop_optimizer_init(LOOPS_NORMAL | LOOPS_HAVE_RECORDED_EXITS);
	gcc_assert(current_loops);

	calculate_dominance_info(CDI_DOMINATORS);
	calculate_dominance_info(CDI_POST_DOMINATORS);

	bb_single_entry = get_single_pred_entry_block();

	bb_single_exit = EXIT_BLOCK_PTR_FOR_FN(cfun);
	if (single_pred_p(bb_single_exit)) {
		bb_single_exit = single_pred(bb_single_exit);
		gcc_assert(single_succ_p(bb_single_exit));
		gcc_assert(!bb_any_loop(bb_single_exit));
	}

	complex_blocks = BITMAP_ALLOC(NULL);
	bitmap_clear(complex_blocks);
	FOR_EACH_BB_FN(bb, cfun) {
		if (rap_ret_gate_bb(bb, ignore_none))
			bitmap_set_bit(complex_blocks, bb->index);
	}

	if (bitmap_empty_p(complex_blocks)) {
		bitmap_clear_bit(complex_functions, NODE_UID(cgraph_get_node(current_function_decl)));
		goto cleanup;
	}

	bb_prologue = get_rap_ret_entry_block(complex_blocks);
	bb_epilogue = get_rap_ret_exit_block(complex_blocks);

	if (bb_epilogue == EXIT_BLOCK_PTR_FOR_FN(cfun))
		bb_prologue = ENTRY_BLOCK_PTR_FOR_FN(cfun);
	else if (can_insert_forwarder_p(CDI_DOMINATORS, bb_epilogue, bb_prologue)) {
		bb_pro = bb_prologue;
		bb_epilogue = make_forwarder_block(bb_epilogue, rap_redirect_epilogue_edge_p, NULL)->src;
		bb_pro = NULL;

		gcc_assert(single_succ_p(bb_epilogue));

		free_dominance_info(CDI_DOMINATORS);
		free_dominance_info(CDI_POST_DOMINATORS);
		calculate_dominance_info(CDI_DOMINATORS);
		calculate_dominance_info(CDI_POST_DOMINATORS);
		gcc_assert(dominated_by_p(CDI_DOMINATORS, bb_epilogue, bb_prologue));

		bb_prologue = get_rap_ret_entry_block(complex_blocks);
		bb_epilogue = get_rap_ret_exit_block(complex_blocks);
	}

	do {
		changed_prologue = false;
		while (bb_prologue != ENTRY_BLOCK_PTR_FOR_FN(cfun) &&
		       (!dominated_by_p(CDI_DOMINATORS, bb_epilogue, bb_prologue) ||
			!dominated_by_p(CDI_POST_DOMINATORS, bb_prologue, bb_epilogue) ||
			bb_any_loop(bb_prologue)))
		{
			bb_prologue = get_immediate_dominator(CDI_DOMINATORS, bb_prologue);
			changed_prologue = true;
		}

		changed_epilogue = false;
		while (bb_epilogue != EXIT_BLOCK_PTR_FOR_FN(cfun) &&
		       (!dominated_by_p(CDI_POST_DOMINATORS, bb_prologue, bb_epilogue) ||
			!dominated_by_p(CDI_DOMINATORS, bb_epilogue, bb_prologue) ||
			bb_any_loop(bb_epilogue)))
		{
			bb_epilogue = get_immediate_dominator(CDI_POST_DOMINATORS, bb_epilogue);
			changed_epilogue = true;
		}
	} while (changed_prologue || changed_epilogue);

	if (bb_prologue == ENTRY_BLOCK_PTR_FOR_FN(cfun))
		gcc_assert(bb_epilogue == EXIT_BLOCK_PTR_FOR_FN(cfun));
	else
		gcc_assert(dominated_by_p(CDI_POST_DOMINATORS, bb_prologue, bb_epilogue));

	if (bb_epilogue == EXIT_BLOCK_PTR_FOR_FN(cfun))
		gcc_assert(bb_prologue == ENTRY_BLOCK_PTR_FOR_FN(cfun));
	else
		gcc_assert(dominated_by_p(CDI_DOMINATORS, bb_epilogue, bb_prologue));

	if (bb_prologue == ENTRY_BLOCK_PTR_FOR_FN(cfun)) {
		bb_prologue = bb_single_entry;
		bb_epilogue = bb_single_exit;
	}

	gcc_assert(!bb_any_loop(bb_prologue));
	gcc_assert(!bb_any_loop(bb_epilogue));

	stmt = last_nonreturn_stmt(bb_epilogue);
	if (stmt && is_gimple_call(stmt)) {
		gcall *call_stmt;

		call_stmt = as_a_gcall(stmt);

		if (gimple_call_fndecl(call_stmt) && TREE_THIS_VOLATILE(gimple_call_fndecl(call_stmt))) {
			bitmap_clear_bit(complex_functions, NODE_UID(cgraph_get_node(current_function_decl)));
			goto cleanup;
		}

		if (bb_prologue == bb_epilogue && gimple_call_tail_p(call_stmt) && !rap_ret_gate_bb(bb_epilogue, ignore_tail)) {
			if (gimple_call_return_type(call_stmt) != void_type_node || TREE_TYPE(TREE_TYPE(current_function_decl)) == void_type_node) {
				bitmap_clear_bit(complex_functions, NODE_UID(cgraph_get_node(current_function_decl)));
				goto cleanup;
			}
		}
	}

	if (enable_xor) {
		gimple_stmt_iterator gsi;

		gsi = gsi_start_nondebug_after_labels_bb(bb_prologue);
		old_retaddr = save_encoded_retaddr(&stmts);
		gsi_insert_seq_before(&gsi, stmts, GSI_CONTINUE_LINKING);
	}

	if (!cfun->tail_call_marked || !check_tail_calls(bb_prologue, bb_epilogue, old_retaddr)) {
		if (bb_epilogue == EXIT_BLOCK_PTR_FOR_FN(cfun)) {
			edge e;
			edge_iterator ei;

			FOR_EACH_EDGE(e, ei, EXIT_BLOCK_PTR_FOR_FN(cfun)->preds) {
				gimple_stmt_iterator gsi;

				gsi = gsi_last_bb(e->src);
				gcc_assert(!gsi_end_p(gsi));
				gcc_assert(gimple_code(gsi_stmt(gsi)) == GIMPLE_RETURN);

				check_retaddr(gsi, old_retaddr, false);
			}
		} else {
			gimple_stmt_iterator gsi;

			gsi = gsi_last_bb(bb_epilogue);
			check_retaddr(gsi, old_retaddr, false);
		}
	}

cleanup:
	BITMAP_FREE(complex_blocks);
	free_dominance_info(CDI_DOMINATORS);
	free_dominance_info(CDI_POST_DOMINATORS);
	loop_optimizer_finalize();
	return 0;
}

#define PASS_NAME rap_ret
#define PROPERTIES_REQUIRED PROP_cfg | PROP_ssa
#define TODO_FLAGS_FINISH TODO_verify_il | TODO_remove_unused_locals | TODO_update_ssa | TODO_cleanup_cfg | TODO_rebuild_cgraph_edges
#include "gcc-generate-gimple-pass.h"

static unsigned int rap_check_tail_calls_execute(void)
{
	basic_block bb;

	FOR_EACH_BB_FN(bb, cfun) {
		gimple_stmt_iterator gsi;

		for (gsi = gsi_start_bb(bb); !gsi_end_p(gsi); gsi_next(&gsi)) {
			gimple_stmt_ptr stmt;
			gcall *call_stmt;
			tree fptr;
			rap_hash_t fptr_hash, func_hash;

			stmt = gsi_stmt(gsi);
			if (!is_gimple_call(stmt))
				continue;

			call_stmt = as_a_gcall(stmt);
			if (!gimple_call_tail_p(call_stmt))
				continue;

			fptr = gimple_call_fn(call_stmt);
			if (!fptr || TREE_CODE(fptr) == INTEGER_CST)
				continue;

			if (TREE_CODE(fptr) == ADDR_EXPR)
				fptr_hash = rap_lookup_imprecise_rap_hash(TREE_OPERAND(fptr, 0));
			else
				fptr_hash = rap_lookup_imprecise_rap_hash(gimple_call_fntype(call_stmt));
			func_hash = rap_lookup_imprecise_rap_hash(current_function_decl);
			if (fptr_hash.hash == func_hash.hash)
				continue;

			gimple_call_set_tail(call_stmt, false);
		}
	}

	return 0;
}

#define PASS_NAME rap_check_tail_calls
#define NO_GATE
#define PROPERTIES_REQUIRED PROP_cfg
#define TODO_FLAGS_FINISH TODO_verify_il | TODO_remove_unused_locals | TODO_update_ssa | TODO_cleanup_cfg
#include "gcc-generate-gimple-pass.h"

static tree __get_reg_target(rtx reg)
{
	df_ref def;

	for (def = DF_REG_DEF_CHAIN(REGNO(reg)); def; def = DF_REF_NEXT_REG(def)) {
		rtx insn, body, dst, src;

		if (DF_REF_IS_ARTIFICIAL(def))
			continue;

		insn = DF_REF_INSN(def);
		body = PATTERN(insn);
		if (GET_CODE(body) == PARALLEL)
			body = XVECEXP(body, 0, 0);
		if (GET_CODE(body) != SET)
			continue;

		dst = SET_DEST(body);
		if (!REG_P(dst))
			continue;
		if (REGNO(dst) != REGNO(reg))
			continue;
		if (ORIGINAL_REGNO(dst) != ORIGINAL_REGNO(reg))
			continue;

		src = SET_SRC(body);
		switch (GET_CODE(src)) {
		default:
			print_rtl_single(stderr, reg);
			print_rtl_single(stderr, insn);
			print_rtl_single(stderr, dst);
			print_rtl_single(stderr, src);
			gcc_unreachable();

		case ASM_OPERANDS:
			if (!strstr(ASM_OPERANDS_TEMPLATE(src), RAP_INDIRECT_BRANCH_TUNNEL))
				continue;

			// retpoline needs to force older gcc to load the fptr into a register
			// trace through the inline asm to find the original fptr
			// see rap_indirect_branch_register_execute for details
			gcc_assert(ASM_OPERANDS_INPUT_LENGTH(src) == 1);
			src = ASM_OPERANDS_INPUT(src, 0);
			gcc_assert(REG_P(src));
			gcc_assert(REG_EXPR(src));
			return REG_EXPR(src);

		case MEM:
			gcc_assert(MEM_EXPR(src));
			return MEM_EXPR(src);
		}
	}

	return NULL_TREE;
}

static tree get_field_at_offset(tree type, poly_int64 offset)
{
	tree field;

	for (field = TYPE_FIELDS(type); field; field = TREE_CHAIN(field)) {
		tree fieldtype;

		fieldtype = TREE_TYPE(field);
		if (RECORD_OR_UNION_TYPE_P(fieldtype)) {
			tree innerfield;

			innerfield = get_field_at_offset(fieldtype, offset - int_byte_position(field));
			if (innerfield)
				return innerfield;
			continue;
		}

		if (!FUNCTION_POINTER_TYPE_P(fieldtype))
			continue;
		if (offset != int_byte_position(field))
			continue;

		return field;
	}

	return NULL_TREE;
}

static tree get_reg_target(rtx_insn *insn, tree origfn, tree fn, poly_int64 offset)
{
	tree fntype, origfntype;

	fntype = TREE_TYPE(fn);
	if (POINTER_TYPE_P(fntype)) {
		fntype = TREE_TYPE(fntype);
		if (TREE_CODE(fntype) == VOID_TYPE) {
			if (origfn == NULL_TREE)
				return NULL_TREE;
			fn = origfn;
			fntype = TREE_TYPE(origfn);
		}
	}

	switch (TREE_CODE(fntype)) {
	default:
#if BUILDING_GCC_VERSION >= 8000
		fprintf(stderr, "CALLTARGET OFF %" HOST_WIDE_INT_PRINT "x ", offset.to_constant());print_rtl_single(stderr, insn);
#else
		fprintf(stderr, "CALLTARGET OFF %" HOST_WIDE_INT_PRINT "x ", offset);print_rtl_single(stderr, insn);
#endif
		debug_tree(fn);
		debug_tree(origfn);
		gcc_unreachable();

	case INTEGER_TYPE:
		gcc_assert(origfn);
		break;

	case RECORD_TYPE: {
		tree field;

		field = get_field_at_offset(fntype, offset);
		if (field == NULL_TREE)
			return NULL_TREE;
		fn = field;
		fntype = TREE_TYPE(TREE_TYPE(field));
		break;
	}

	case UNION_TYPE: {
		tree field;

		for (field = TYPE_FIELDS(fntype); field; field = TREE_CHAIN(field)) {
			if (!FUNCTION_POINTER_TYPE_P(TREE_TYPE(field)))
				continue;
			if (TREE_TYPE(TREE_TYPE(field)) == TREE_TYPE(origfn))
				break;
		}
		gcc_assert(field);
		fn = field;
		fntype = TREE_TYPE(TREE_TYPE(field));
		break;
	}

	case ARRAY_TYPE:
		if (FUNCTION_POINTER_TYPE_P(TREE_TYPE(fntype)))
			fntype = TREE_TYPE(TREE_TYPE(fntype));
		else {
			gcc_assert(origfn);
			fn = origfn;
			fntype = TREE_TYPE(origfn);
		}
		break;

	case FUNCTION_TYPE:
		break;
	}

	if (!origfn)
		return fn;

	origfntype = TREE_TYPE(origfn);
	gcc_assert(TREE_CODE(origfntype) == FUNCTION_TYPE);

	return origfntype == fntype ? fn : origfn;
}

tree get_call_target(rtx_insn *insn)
{
	rtx body;
	tree fn, origfn;

	body = PATTERN(insn);
	if (GET_CODE(body) == PARALLEL)
		body = XVECEXP(body, 0, 0);
	if (GET_CODE(body) == SET)
		body = SET_SRC(body);
	if (GET_CODE(body) != CALL)
		return NULL_TREE;

	body = XEXP(body, 0);
	gcc_assert(GET_CODE(body) == MEM);
	origfn = MEM_EXPR(body);
	body = XEXP(body, 0);

	switch (GET_CODE(body)) {
	default:
		return NULL_TREE;

	case MEM: {
		poly_int64 offset;

		if (origfn)
			return origfn;

		origfn = MEM_EXPR(body);
		if (origfn)
			return origfn;

		body = XEXP(body, 0);
		if (REG_P(body))
			offset = 0;
		else {
			gcc_assert(GET_CODE(body) == PLUS);
			gcc_assert(GET_CODE(XEXP(body, 0)) == REG);
			gcc_assert(GET_CODE(XEXP(body, 1)) == CONST_INT);
			offset = INTVAL(XEXP(body, 1));
			body = XEXP(body, 0);
		}
		gcc_assert(REG_EXPR(body));
		return get_reg_target(insn, origfn, REG_EXPR(body), offset);
	}

	case SYMBOL_REF:
		gcc_assert(SYMBOL_REF_FUNCTION_P(body));
		fn = SYMBOL_REF_DECL(body);
		gcc_assert(!fn || TREE_CODE(fn) == FUNCTION_DECL);
		return fn;

	case REG:
		fn = REG_EXPR(body);
		if (!fn) {
			if (origfn != NULL_TREE)
				return origfn;

			origfn = __get_reg_target(body);
			if (origfn)
				return origfn;

			error_at(INSN_LOCATION(insn), "RAP: fptr variable not found [%u]", ORIGINAL_REGNO(body));
			fflush(stderr);
			print_rtl_single(stderr, insn);
			print_rtl_single(stderr, body);
			gcc_unreachable();
		}

		switch (TREE_CODE(fn)) {
		default:
			error_at(INSN_LOCATION(insn), "RAP: unknown fptr variable");
			fflush(stderr);
			print_rtl_single(stderr, insn);
			print_rtl_single(stderr, body);
			debug_tree(fn);
			debug_tree(origfn);
			gcc_unreachable();

		case ARRAY_REF:
		case COMPONENT_REF:
		case MEM_REF:
		case PARM_DECL:
		case SSA_NAME:
		case TARGET_MEM_REF:
		case VAR_DECL: {
			poly_int64 offset;
			gcc_assert(REG_ATTRS(body));
			offset = REG_ATTRS(body)->offset;

			return get_reg_target(insn, origfn, fn, offset);
		}
		}
	}
}

static tree rap_get_retloc_hash(rtx_insn *insn, rtx_insn *reference)
{
	tree fn, fntype;
	rap_hash_t func_hash;

	fn = get_call_target(insn);
	if (!fn) {
		if (!reference) {
			print_rtl_single(stderr, insn);
			gcc_unreachable();
		}
		fn = get_call_target(reference);
	}

	gcc_assert(fn);
	fntype = strip_array_types(TREE_TYPE(fn));

	switch (TREE_CODE(fn)) {
	default:
		if (DECL_P(fn))
			inform(INSN_LOCATION(insn), "PAX RETLOC DECL %qD", fn);
		else
			inform(INSN_LOCATION(insn), "PAX RETLOC TYPE %qE", fn);
		fflush(stderr);
		print_rtl_single(stderr, insn);
		if (reference) print_rtl_single(stderr, reference);
		debug_tree(fn);
		gcc_unreachable();

	case FUNCTION_DECL:
		if (!lookup_attribute("rap_hash_override", DECL_ATTRIBUTES(fn)))
			break;

		func_hash = rap_lookup_imprecise_rap_hash(fn);
		return build_int_cst_type(rap_hash_type_node, -func_hash.hash);

	case COMPONENT_REF:
	case FIELD_DECL:
	case PARM_DECL:
	case SSA_NAME:
	case TARGET_MEM_REF:
	case VAR_DECL:
		if (TREE_CODE(fntype) != POINTER_TYPE) {
			print_rtl_single(stderr, insn);
			if (reference) print_rtl_single(stderr, reference);
			debug_tree(fn);
			debug_tree(fntype);
			gcc_unreachable();
		}
		fntype = TREE_TYPE(fntype);
		// FALLTHROUGH

	case MEM_REF:
		if (TREE_CODE(fntype) != FUNCTION_TYPE) {
			print_rtl_single(stderr, insn);
			debug_tree(fn);
			debug_tree(fntype);
			gcc_unreachable();
		}
		break;
	}

	if (lookup_attribute("rap_hash", TYPE_ATTRIBUTES(fntype))) {
		func_hash = rap_lookup_imprecise_rap_hash(fntype);
		return build_int_cst_type(rap_hash_type_node, -func_hash.hash);
	}

	gcc_assert(TREE_CODE(fn) == FUNCTION_DECL && DECL_NAME(fn));
	if (!strcmp(DECL_NAME_POINTER(fn), "memcpy") ||
	    !strcmp(DECL_NAME_POINTER(fn), "memset") ||
	    !strcmp(DECL_NAME_POINTER(fn), "__gcov_init") ||
	    !strcmp(DECL_NAME_POINTER(fn), "__gcov_exit")) {
		rap_calculate_fndecl_hash(fn);
		func_hash = rap_lookup_imprecise_rap_hash(fn);
		return build_int_cst_type(rap_hash_type_node, -func_hash.hash);
	}

	print_rtl_single(stderr, insn);
	if (reference) print_rtl_single(stderr, reference);
	debug_tree(fn);
	debug_tree(DECL_NAME(fn));
	debug_tree(fntype);
	return DECL_NAME(fn);
	gcc_unreachable();
}

void rap_mark_retloc(rtx_insn *insn, rtx_insn *reference)
{
	rtvec argvec, constraintvec, labelvec;
	rtx mark;
	rtx_code_label *label;
	tree computed_hash = NULL_TREE;
	unsigned int loc = INSN_LOCATION(insn);
	static const char * const retloc_asm[3] = {
#ifdef TARGET_386
		[4/8] = "rap_retloc hashval=%c0 rethooks=%c1",
		[8/8] = "rap_retloc hashval=%c0 rethooks=%c1",
		[16/8] = "rap_retloc callee=%s rethooks=%%c0",
#elif defined(TARGET_ARM) || defined(TARGET_THUMB)
		[4/8] = "b %l1 ; .long %c0",
#elif defined(AARCH64_ABI_DEFAULT)
		[4/8] = "b %l1 ; .long %c0",
		[8/8] = "b %l1 ; .quad %c0",
#elif defined(SPARC_DEFAULT_CMODEL)
		[4/8] = "jmp %l1 ; nop ; .long %c0",
		[8/8] = "jmp %l1 ; nop ; .quad %c0",
#else
#error unsupported target
#endif
	};

	if (!enable_type_ret)
		return;

	computed_hash = rap_get_retloc_hash(insn, reference);
	gcc_assert(computed_hash);

	switch (TREE_CODE(computed_hash)) {
	default:
		debug_rtx(insn);
		debug_rtx(reference);
		debug_tree(computed_hash);
		gcc_unreachable();

	case IDENTIFIER_NODE: {
		char *retloc_callee;

		argvec = rtvec_alloc(1);
		constraintvec = rtvec_alloc(1);
		labelvec = rtvec_alloc(1);

		gcc_assert(asprintf(&retloc_callee, retloc_asm[2], IDENTIFIER_POINTER(computed_hash)) != -1);
		mark = gen_rtx_ASM_OPERANDS(VOIDmode, ggc_strdup(retloc_callee), empty_string, 0, argvec, constraintvec, labelvec, loc);
		free(retloc_callee);
		MEM_VOLATILE_P(mark) = 1;

		ASM_OPERANDS_INPUT(mark, 0) = expand_expr(has_trait(current_function_decl, "phys") ? integer_zero_node : integer_one_node, NULL_RTX, VOIDmode, EXPAND_INITIALIZER);

		ASM_OPERANDS_INPUT_CONSTRAINT_EXP(mark, 0) = gen_rtx_ASM_INPUT_loc(DImode, ggc_strdup("i"), loc);
		break;
	}

	case INTEGER_CST:
		argvec = rtvec_alloc(2);
		constraintvec = rtvec_alloc(2);
		labelvec = rtvec_alloc(1);

		mark = gen_rtx_ASM_OPERANDS(VOIDmode, ggc_strdup(retloc_asm[UNITS_PER_WORD / 8]), empty_string, 0, argvec, constraintvec, labelvec, loc);
		MEM_VOLATILE_P(mark) = 1;

		ASM_OPERANDS_INPUT(mark, 0) = expand_expr(computed_hash, NULL_RTX, VOIDmode, EXPAND_INITIALIZER);
		ASM_OPERANDS_INPUT(mark, 1) = expand_expr(has_trait(current_function_decl, "phys") ? integer_zero_node : integer_one_node, NULL_RTX, VOIDmode, EXPAND_INITIALIZER);

		ASM_OPERANDS_INPUT_CONSTRAINT_EXP(mark, 0) = gen_rtx_ASM_INPUT_loc(DImode, ggc_strdup("i"), loc);
		ASM_OPERANDS_INPUT_CONSTRAINT_EXP(mark, 1) = gen_rtx_ASM_INPUT_loc(DImode, ggc_strdup("i"), loc);
		break;
	}

	label = gen_label_rtx();
	ASM_OPERANDS_LABEL(mark, 0) = label;

#if BUILDING_GCC_VERSION < 8000
	do {
		insn = NEXT_INSN(insn);
	} while (GET_CODE(insn) == NOTE && NOTE_KIND(insn) == NOTE_INSN_CALL_ARG_LOCATION);
	insn = PREV_INSN(insn);
#endif

	insn = emit_insn_after_setloc(mark, insn, loc);

	emit_label_after(label, insn);
	LABEL_NUSES(label)++;
}

static unsigned int rap_mark_retloc_execute(void)
{
	rtx_insn *insn;

	for (insn = get_insns(); insn; insn = NEXT_INSN(insn)) {
		rtx body;

		if (INSN_DELETED_P(insn))
			continue;

		// rtl match (call_insn (set (reg) (call (mem))))
		if (!CALL_P(insn))
			continue;

		if (SIBLING_CALL_P(insn))
			continue;

		if (find_reg_note(insn, REG_NORETURN, 0))
			continue;

		body = PATTERN(insn);
		if (GET_CODE(body) == SET)
			body = SET_SRC(body);
		gcc_assert(GET_CODE(body) == CALL);

		body = XEXP(body, 0);
		gcc_assert(MEM_P(body));

		body = XEXP(body, 0);

		if (GET_CODE(body) == SYMBOL_REF) {
			// TODO: skip direct calls to non-complex functions that won't emit a return check anyway
			gcc_assert(SYMBOL_REF_FUNCTION_P(body));
		} else if (enable_type_nospec && rap_retpoline_gate()) {
			// skip indirect calls here that retpoline will handle later
			continue;
		}

		rap_mark_retloc(insn, NULL);
	}

	return 0;
}

#define PASS_NAME rap_mark_retloc
#define NO_GATE
#define TODO_FLAGS_FINISH TODO_verify_il | TODO_df_verify | TODO_df_finish
#include "gcc-generate-rtl-pass.h"
