/*
 * Copyright 2012-2024 by PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 *
 * Homepage: http://pax.grsecurity.net/
 */

#include "rap.h"

#ifdef TARGET_386
static bool rap_match_leaveq(rtx insn)
{
	rtx body;

	if (!NONJUMP_INSN_P(insn))
		return false;

	body = PATTERN(insn);
	if (GET_CODE(body) != PARALLEL)
		return false;

	return rtx_equal_p(XVECEXP(leaveq_rtx, 0, 0), XVECEXP(body, 0, 0)) && rtx_equal_p(XVECEXP(leaveq_rtx, 0, 1), XVECEXP(body, 0, 1));
}

static bool rap_match_pop_frame_ptr(rtx insn)
{
	rtx body;

	if (!NONJUMP_INSN_P(insn))
		return false;

	body = PATTERN(insn);
	if (GET_CODE(body) != SET)
		return false;

	return rtx_equal_p(pop_rbp_rtx, body);
}

static bool rap_rewrite_pop_frame_ptr(rtx_insn *insn)
{
	if (!rap_match_pop_frame_ptr(insn))
		return false;

	PATTERN(insn) = copy_rtx(leaveq_rtx);
	df_insn_rescan(insn);
	return true;
}
#endif

// ensure that the check of the retaddr uses the same stack slot as retq itself
static unsigned int rap_rewrite_retaddr_check_execute(void)
{
#ifdef TARGET_386
	df_ref use, def;
	bool has_leaveq = false, has_pop_rbp = false;

	if (!cfun->machine->fs.sp_valid || cfun->calls_alloca) {
		rtx_insn *insn;
		edge e;
		edge_iterator ei;

		if (!cfun->machine->fs.fp_valid && !lookup_attribute("naked", DECL_ATTRIBUTES(current_function_decl))) {
			debug_tree(current_function_decl);
			gcc_unreachable();
		}

		FOR_EACH_EDGE(e, ei, EXIT_BLOCK_PTR_FOR_FN(cfun)->preds) {
			edge e2;
			edge_iterator ei2;

			if (e->flags & EDGE_FAKE)
				continue;

			insn = BB_END(e->src);

			if (JUMP_P(insn))
				gcc_assert(JUMP_LABEL(insn) ? ANY_RETURN_P(JUMP_LABEL(insn)) : ANY_RETURN_P(PATTERN(insn)));
			else if (CALL_P(insn))
				gcc_assert(SIBLING_CALL_P(insn) && e->flags & EDGE_SIBCALL);
			else
				continue;

			insn = PREV_INSN(insn);
			if (rap_rewrite_pop_frame_ptr(insn)) {
				if (cfun->tail_call_marked)
					continue;
				return 0;
			}

			FOR_EACH_EDGE(e2, ei2, e->src->preds) {
				gcc_assert(!(e2->flags & EDGE_FAKE));

				FOR_BB_INSNS_REVERSE(e2->src, insn) {
					if (!NONJUMP_INSN_P(insn))
						continue;

					rap_rewrite_pop_frame_ptr(insn);
					break;
				}
			}
		}
		return 0;
	}

	for (def = DF_REG_DEF_CHAIN(REGNO(hard_frame_pointer_rtx)); def; def = DF_REF_NEXT_REG(def)) {
		rtx insn;

		if (DF_REF_IS_ARTIFICIAL(def))
			continue;

		insn = DF_REF_INSN(def);
		if (rap_match_pop_frame_ptr(insn))
			has_pop_rbp = true;
		else if (rap_match_leaveq(insn))
			has_leaveq = true;
	}

	gcc_assert(!has_pop_rbp || !has_leaveq);

	for (use = DF_REG_USE_CHAIN(REGNO(hard_frame_pointer_rtx)); use; use = DF_REF_NEXT_REG(use)) {
		rtx_insn *insn;
		rtx body, dst, src, src1, src2, bp, off;
		tree rap_retaddr;

		if (DF_REF_IS_ARTIFICIAL(use))
			continue;

		insn = DF_REF_INSN(use);
		if (!NONJUMP_INSN_P(insn))
			continue;

		body = PATTERN(insn);

		if (GET_CODE(body) != SET)
			continue;

		dst = SET_DEST(body);
		if (!REG_P(dst))
			continue;

		src = SET_SRC(body);
		if (GET_CODE(src) != COMPARE)
			continue;
		src1 = XEXP(src, 0);
		if (!REG_P(src1) || !REG_ATTRS(src1))
			continue;

		rap_retaddr = REG_EXPR(src1);
		if (!rap_retaddr)
			continue;
		if (!VAR_P(rap_retaddr))
			continue;

		if (!DECL_NAME(rap_retaddr) || strncmp(DECL_NAME_POINTER(rap_retaddr), "rap_retaddr_exit_ptr", sizeof("rap_retaddr_exit_ptr") - 1))
			continue;

		src2 = XEXP(src, 1);
		if (!MEM_P(src2))
			continue;
		body = XEXP(src2, 0);
		if (GET_CODE(body) != PLUS)
			continue;
		bp = XEXP(body, 0);
		gcc_assert(GET_CODE(bp) == REG);
		if (REGNO(bp) != REGNO(hard_frame_pointer_rtx))
			continue;
		off = XEXP(body, 1);
		if (!CONST_INT_P(off))
			continue;

		gcc_assert(!ix86_static_chain_on_stack);

		if (cfun->machine->fs.realigned) {
			inform(DECL_SOURCE_LOCATION(current_function_decl), "PAX: rewrite retaddr cmpq stack_realign_drap:%u stack_realign_fp:%u for %qE", stack_realign_drap, stack_realign_fp, current_function_decl);
			gcc_unreachable();
		} else {
			if (has_leaveq && !has_pop_rbp)
				return 0;
			XEXP(body, 0) = stack_pointer_rtx;
			XEXP(body, 1) = GEN_INT(cfun->machine->fs.sp_offset - cfun->machine->fs.fp_offset + INTVAL(off));
		}

		use = DF_REF_PREV_REG(use);
		df_insn_rescan(insn);
		if (DF_REF_NEXT_REG(use))
			use = DF_REF_NEXT_REG(use);
	}
	return 0;
#endif
	return 0;
}

static bool rap_rewrite_retaddr_check_gate(void)
{
	return rap_ret_gate();
}

#define PASS_NAME rap_rewrite_retaddr_check
#define TODO_FLAGS_FINISH TODO_verify_il | TODO_df_verify | TODO_df_finish
#include "gcc-generate-rtl-pass.h"
