/*
 * Copyright 2011-2024 by Emese Revfy <re.emese@gmail.com>
 * Licensed under the GPL v2
 *
 * Documentation:
 * http://forums.grsecurity.net/viewtopic.php?f=7&t=3043
 *
 * This plugin recomputes expressions of function arguments marked by a size_overflow attribute
 * with double integer precision (DImode/TImode for 32/64 bit integer types).
 * The recomputed argument is checked against TYPE_MAX and an event is logged on overflow and the triggering process is killed.
 *
 * Usage:
 * $ make
 * $ make run
 */

#include "size_overflow.h"

__visible int plugin_is_GPL_compatible;

GTY() tree size_overflow_type_HI;
GTY() tree size_overflow_type_SI;
GTY() tree size_overflow_type_DI;
GTY() tree size_overflow_type_TI;

bool check_fields, check_fns, check_fnptrs, check_vars, search_missing_fns;
bool enable_ipa_transform;

static struct plugin_info size_overflow_plugin_info = {
	.version	= "20231114",
	.help		= "no-size-overflow\tturn off size overflow checking\n",
};

static tree handle_size_overflow_attribute(tree *node, tree __unused name, tree args, int __unused flags, bool *no_add_attrs)
{
	unsigned int arg_count;
	enum tree_code code = TREE_CODE(*node);

	switch (code) {
	case FUNCTION_DECL:
		arg_count = type_num_arguments(TREE_TYPE(*node));
		break;
	case FUNCTION_TYPE:
	case METHOD_TYPE:
		arg_count = type_num_arguments(*node);
		break;
	default:
		*no_add_attrs = true;
		debug_tree(*node);
		fflush(stderr);
		error_at(DECL_P(*node) ? DECL_SOURCE_LOCATION(*node) : UNKNOWN_LOCATION, "%s: %qE attribute only applies to functions", __func__, name);
		return NULL_TREE;
	}

	for (; args; args = TREE_CHAIN(args)) {
		int cur_val;
		tree position = TREE_VALUE(args);

		if (TREE_CODE(position) != INTEGER_CST) {
			error_at(DECL_P(*node) ? DECL_SOURCE_LOCATION(*node) : UNKNOWN_LOCATION, "%s: parameter isn't an integer", __func__);
			fflush(stderr);
			debug_tree(args);
			*no_add_attrs = true;
			return NULL_TREE;
		}

		cur_val = tree_to_shwi(position);
		if (cur_val < 0 || arg_count < (unsigned int)cur_val) {
			error_at(DECL_P(*node) ? DECL_SOURCE_LOCATION(*node) : UNKNOWN_LOCATION, "%s: parameter %d is outside range.", __func__, cur_val);
			*no_add_attrs = true;
			return NULL_TREE;
		}
	}
	return NULL_TREE;
}

static tree handle_intentional_overflow_attribute(tree *node, tree __unused name, tree args, int __unused flags, bool *no_add_attrs)
{
	unsigned int arg_count;
	HOST_WIDE_INT s_first_arg;
	enum tree_code code = TREE_CODE(*node);

	switch (code) {
	case FUNCTION_DECL:
		arg_count = type_num_arguments(TREE_TYPE(*node));
		break;
//	case FUNCTION_TYPE:
//	case METHOD_TYPE:
//		arg_count = type_num_arguments(*node);
//		break;
	case VAR_DECL:
	case FIELD_DECL:
		return NULL_TREE;
	default:
		*no_add_attrs = true;
		debug_tree(*node);
		fflush(stderr);
		error_at(DECL_SOURCE_LOCATION(*node), "%qE attribute only applies to functions, fields or vars", name);
		return NULL_TREE;
	}

	s_first_arg = tree_to_shwi(TREE_VALUE(args));
	if (s_first_arg == -1)
		return NULL_TREE;
	if (s_first_arg < -1)
		error_at(DECL_SOURCE_LOCATION(*node), "%s: parameter %d is outside range.", __func__, (int)s_first_arg);

	for (; args; args = TREE_CHAIN(args)) {
		unsigned int cur_val;

		if (TREE_CODE(TREE_VALUE(args)) != INTEGER_CST) {
			error_at(DECL_P(*node) ? DECL_SOURCE_LOCATION(*node) : UNKNOWN_LOCATION, "%s: parameter isn't an integer", __func__);
			fflush(stderr);
			debug_tree(args);
			*no_add_attrs = true;
			return NULL_TREE;
		}

		cur_val = (unsigned int)tree_to_uhwi(TREE_VALUE(args));
		if (cur_val > arg_count ) {
			error_at(DECL_SOURCE_LOCATION(*node), "%s: parameter %u is outside range. (arg_count: %u)", __func__, cur_val, arg_count);
			*no_add_attrs = true;
			return NULL_TREE;
		}
	}
	return NULL_TREE;
}

static struct attribute_spec size_overflow_attr = ATTRIBUTE_SPEC_INIT("size_overflow", 1, -1, true, false, false, handle_size_overflow_attribute, false, NULL);
static struct attribute_spec intentional_overflow_attr = ATTRIBUTE_SPEC_INIT("intentional_overflow", 1, -1, true, false, false, handle_intentional_overflow_attribute, false, NULL);

static void register_attributes(void __unused *event_data, void __unused *data)
{
	register_attribute(&size_overflow_attr);
	register_attribute(&intentional_overflow_attr);
}

static tree create_typedef(tree type, const char* ident)
{
	tree new_type, decl;

	new_type = build_variant_type_copy(type);
	decl = build_decl(BUILTINS_LOCATION, TYPE_DECL, get_identifier(ident), new_type);
	DECL_ORIGINAL_TYPE(decl) = type;
	TYPE_NAME(new_type) = decl;
	return new_type;
}

// Create the size_overflow_type_* trees.
static void size_overflow_start_unit(void __unused *gcc_data, void __unused *user_data)
{
	size_overflow_type_HI = create_typedef(intHI_type_node, "size_overflow_type_HI");
	size_overflow_type_SI = create_typedef(intSI_type_node, "size_overflow_type_SI");
	size_overflow_type_DI = create_typedef(intDI_type_node, "size_overflow_type_DI");
	size_overflow_type_TI = create_typedef(intTI_type_node, "size_overflow_type_TI");
}

#if BUILDING_GCC_VERSION >= 12000
// undo the unwanted frontend UBSAN instrumentations caused by gcc commit 9c1a633d96926357155d4702b66f8a0ec856a81f
static unsigned int disable_ubsan_si_overflow_execute(void)
{
	basic_block bb;

	FOR_EACH_BB_FN(bb, cfun) {
		gimple_stmt_iterator gsi;

		for (gsi = gsi_start_bb(bb); !gsi_end_p(gsi); gsi_next(&gsi)) {
			gimple_stmt_ptr stmt;
			gcall *call_stmt;
			gcond *cond_stmt;
			tree fndecl;
			edge e;
			basic_block cond_bb;

			// gimple match:  __builtin___ubsan_handle_divrem_overflow (&*.Lubsan_data47, __x.8_10, _9);
			stmt = gsi_stmt(gsi);
			if (!is_gimple_call(stmt))
				continue;
			call_stmt = as_a_gcall(stmt);

			fndecl = gimple_call_fndecl(call_stmt);
			if (fndecl == NULL_TREE)
				continue;
			if (!fndecl_built_in_class_p(fndecl, BUILT_IN_NORMAL))
				continue;
			if (DECL_FUNCTION_CODE(fndecl) != BUILT_IN_UBSAN_HANDLE_DIVREM_OVERFLOW)
				continue;

			// DCE the instrumentation by faking the controlling condition to never execute it
			gcc_assert(single_succ_p(bb));
			cond_bb = single_succ(bb);
			gcc_assert(EDGE_COUNT(cond_bb->preds) == 2);
			e = EDGE_I(cond_bb->preds, 0)->src == bb ? EDGE_I(cond_bb->preds, 1) : EDGE_I(cond_bb->preds, 0);
			stmt = gimple_seq_last_nondebug_stmt(bb_seq(e->src));
			gcc_assert(gimple_code(stmt) == GIMPLE_COND);
			cond_stmt = as_a_gcond(stmt);
			if (e->flags & EDGE_TRUE_VALUE)
				gimple_cond_make_true(cond_stmt);
			else {
				gcc_assert(e->flags & EDGE_FALSE_VALUE);
				gimple_cond_make_false(cond_stmt);
			}
			update_stmt(cond_stmt);
		}
	}

	return 0;
}
#endif

static bool disable_ubsan_si_overflow_gate(void)
{
	flag_sanitize &= ~SANITIZE_SI_OVERFLOW;
	return true;
}

#define PASS_NAME disable_ubsan_si_overflow
#if BUILDING_GCC_VERSION < 12000
#define NO_EXECUTE
#define TODO_FLAGS_FINISH TODO_verify_il | TODO_remove_unused_locals | TODO_update_ssa_no_phi | TODO_cleanup_cfg
#endif
#include "gcc-generate-gimple-pass.h"

/*
 * convert this:
 *
 * cdev->int_params.in.num_vectors = (u8) ((signed char) cdev->int_params.in.num_vectors + (signed char) sb_cnt_info.cnt);
 *
 * into this:
 *
 * cdev->int_params.in.num_vectors = (cdev->int_params.in.num_vectors + (u8) sb_cnt_info.cnt);
 *
 *
 */
static bool handle_useless_sign_casts(tree rhs)
{
	tree rhs_ltype, rhs_rtype;
	tree rhs1, rhs1_ltype, rhs1_rtype;
	tree rhs2, rhs2_ltype, rhs2_rtype;

	rhs_ltype = TREE_TYPE(rhs);
	if (!INTEGRAL_TYPE_P(rhs_ltype))
		return false;
	if (!TYPE_UNSIGNED(rhs_ltype))
		return false;

	rhs = TREE_OPERAND(rhs, 0);
	rhs_rtype = TREE_TYPE(rhs);
	gcc_assert(INTEGRAL_TYPE_P(rhs_ltype));
	if (TYPE_UNSIGNED(rhs_rtype))
		return false;

	switch (TREE_CODE(rhs)) {
	default:
		return false;

	case MINUS_EXPR:
	case PLUS_EXPR:
		if (TYPE_SIZE(rhs_ltype) != TYPE_SIZE(rhs_rtype))
			return false;
		break;
	}

	rhs1 = TREE_OPERAND(rhs, 0);
	if (!CONVERT_EXPR_P(rhs1))
		return false;
	rhs1_ltype = TREE_TYPE(rhs1);
	if (TYPE_SIZE(rhs1_ltype) < TYPE_SIZE(rhs_rtype))
		return false;
	rhs1_rtype = TREE_TYPE(TREE_OPERAND(rhs1, 0));

	rhs2 = TREE_OPERAND(rhs, 1);
	if (!CONVERT_EXPR_P(rhs2))
		return false;
	rhs2_ltype = TREE_TYPE(rhs2);
	if (TYPE_SIZE(rhs2_ltype) < TYPE_SIZE(rhs_rtype))
		return false;
	rhs2_rtype = TREE_TYPE(TREE_OPERAND(rhs2, 0));

	if (INTEGRAL_TYPE_P(rhs1_rtype) && INTEGRAL_TYPE_P(rhs2_rtype) && useless_type_conversion_p(rhs1_rtype, rhs2_rtype)) {
		TREE_TYPE(rhs) = TYPE_MAIN_VARIANT(rhs1_rtype);
		TREE_TYPE(rhs1) = rhs1_rtype;
		TREE_TYPE(rhs2) = rhs2_rtype;
	} else if (TREE_CODE(rhs) != MINUS_EXPR) {
		TREE_TYPE(rhs) = rhs_ltype;
		TREE_TYPE(rhs1) = rhs_ltype;
		TREE_TYPE(rhs2) = rhs_ltype;
	}

	return true;
}

static bool handle_unsafe_assignments(tree stmt)
{
	tree rhs;

	gcc_assert(TREE_CODE(stmt) == MODIFY_EXPR);
	rhs = TREE_OPERAND(stmt, 1);
	switch (TREE_CODE(rhs)) {
	default:
		return false;

	case CONVERT_EXPR:
		return false;

	case NOP_EXPR:
		return handle_useless_sign_casts(rhs);
	}
}

static bool is_unsigned_integer_type(tree op)
{
	tree type = TREE_TYPE(op);

	if (TREE_CODE(type) != INTEGER_TYPE)
		return false;
	return TYPE_UNSIGNED(type);
}

/*
 * convert this:
 * e.g. include/linux/usb.h unsigned int __create_pipe()
 * int devnum;
 * unsigned int endpoint;
 *
 * return (dev->devnum << 8) | (endpoint << 15);
 * return (unsigned int) (SAVE_EXPR <dev->devnum> << 8) | SAVE_EXPR <endpoint> << 15;
 *
 * into this:
 *
 * return ((unsigned int)dev->devnum << 8) | (endpoint << 15);
 * return SAVE_EXPR <(unsigned int) dev->devnum> << 8 | SAVE_EXPR <endpoint> << 15;
 */

static void handle_unsafe_returns(tree stmt)
{
	tree comp_ref, save, cast_old, lshift, ior, result, cast_new;
	tree modify = TREE_OPERAND(stmt, 0);

	if (modify == NULL_TREE)
		return;
	if (TREE_CODE(modify) != MODIFY_EXPR)
		return;
	if (!is_unsigned_integer_type(modify))
		return;

	result = TREE_OPERAND(modify, 0);
	if (TREE_CODE(result) != RESULT_DECL)
		return;
	if (!is_unsigned_integer_type(result))
		return;

	ior = TREE_OPERAND(modify, 1);
	if (TREE_CODE_CLASS(TREE_CODE(ior)) != tcc_binary)
		return;
	if (!is_unsigned_integer_type(ior))
		return;

	cast_old = TREE_OPERAND(ior, 0);
	if (TREE_CODE(cast_old) != NOP_EXPR)
		return;
	if (!is_unsigned_integer_type(cast_old))
		return;

	lshift = TREE_OPERAND(cast_old, 0);
	if (TREE_CODE(lshift) != LSHIFT_EXPR)
		return;
	if (is_unsigned_integer_type(lshift))
		return;

	save = TREE_OPERAND(lshift, 0);
	if (TREE_CODE(save) != SAVE_EXPR)
		return;
	if (is_unsigned_integer_type(save))
		return;

	comp_ref = TREE_OPERAND(save, 0);
	if (TREE_CODE(comp_ref) != COMPONENT_REF)
		return;
	if (is_unsigned_integer_type(comp_ref))
		return;

	gcc_assert(TREE_CODE(TREE_OPERAND(lshift, 1)) == INTEGER_CST);

	TREE_OPERAND(ior, 0) = lshift;
	TREE_TYPE(lshift) = TREE_TYPE(cast_old);

	cast_new = build_c_cast(UNKNOWN_LOCATION, TREE_TYPE(cast_old), comp_ref);
	TREE_OPERAND(save, 0) = cast_new;
}

static tree handle_unsafe_canonicalizations(tree *tp, int *walk_subtrees, void *data)
{
	tree stmt = *tp;

	*walk_subtrees = 0;

	if (TYPE_P(stmt))
		return NULL_TREE;

	switch (TREE_CODE(stmt)) {
	default:
		break;

	case RETURN_EXPR:
		handle_unsafe_returns(stmt);
		break;

	case MODIFY_EXPR:
		if (handle_unsafe_assignments(stmt))
			break;
		// FALLTHROUGH

	case BIND_EXPR:
	case COMPOUND_EXPR:
	case COND_EXPR:
	case CONVERT_EXPR:
	case DECL_EXPR:
	case FOR_STMT:
	case NOP_EXPR:
	case SAVE_EXPR:
	case STATEMENT_LIST:
	case SWITCH_STMT:
	case TARGET_EXPR:
	case TRY_FINALLY_EXPR:
	case VAR_DECL:
	case WHILE_STMT:
		*walk_subtrees = 1;
		break;
	}

	return NULL_TREE;
}

static void undo_unsafe_canonicalizations(void *event_data __unused, void *data __unused)
{
	tree fndecl = (tree)event_data, stmts;

	stmts = DECL_SAVED_TREE(fndecl);
	if (stmts == NULL_TREE)
		return;

	if (TREE_CODE(stmts) == BIND_EXPR)
		stmts = BIND_EXPR_BODY(stmts);
	walk_tree_without_duplicates(&stmts, handle_unsafe_canonicalizations, NULL);
}

__visible int plugin_init(struct plugin_name_args *plugin_info, struct plugin_gcc_version *version)
{
	int i;
	const char * const plugin_name = plugin_info->base_name;
	const int argc = plugin_info->argc;
	const struct plugin_argument * const argv = plugin_info->argv;
	bool enable = true;

	static const struct ggc_root_tab gt_ggc_r_size_overflow[] = {
		INIT_GGC_ROOT_TAB_TREE(size_overflow_type_HI),
		INIT_GGC_ROOT_TAB_TREE(size_overflow_type_SI),
		INIT_GGC_ROOT_TAB_TREE(size_overflow_type_DI),
		INIT_GGC_ROOT_TAB_TREE(size_overflow_type_TI),
		LAST_GGC_ROOT_TAB
	};

	PASS_INFO(insert_size_overflow_asm, "ssa", 1, PASS_POS_INSERT_AFTER);
	PASS_INFO(size_overflow, "inline", 1, PASS_POS_INSERT_AFTER);
	PASS_INFO(size_overflow_transform, "slsr", 1, PASS_POS_INSERT_AFTER);
	PASS_INFO(disable_ubsan_si_overflow, "ubsan", 1, PASS_POS_REPLACE);

	if (!plugin_default_version_check(version, &gcc_version)) {
		error_gcc_version(version);
		return 1;
	}

	for (i = 0; i < argc; ++i) {
		if (!strcmp(argv[i].key, "no-size-overflow")) {
			enable = false;
			continue;
		}

		if (!strcmp(argv[i].key, "check-fields")) {
			check_fields = true;
			continue;
		}

		if (!strcmp(argv[i].key, "check-fns")) {
			check_fns = true;
			continue;
		}

		if (!strcmp(argv[i].key, "check-fptrs")) {
			check_fnptrs = true;
			continue;
		}

		if (!strcmp(argv[i].key, "check-vars")) {
			check_vars = true;
			continue;
		}

		if (!strcmp(argv[i].key, "search-missing-fns")) {
			search_missing_fns = true;
			continue;
		}

		if (!strcmp(argv[i].key, "enable-ipa-transform")) {
			enable_ipa_transform = true;
			continue;
		}

		error(G_("unknown option '-fplugin-arg-%s-%s'"), plugin_name, argv[i].key);
	}

	register_callback(plugin_name, PLUGIN_INFO, NULL, &size_overflow_plugin_info);
	if (enable) {
		struct opt_pass *reference_pass;

		if (flag_sanitize & SANITIZE_SI_OVERFLOW) {
			error(G_("ubsan SANITIZE_SI_OVERFLOW option is unsupported"));
			return 1;
		}

		register_callback(plugin_name, PLUGIN_PRE_GENERICIZE, undo_unsafe_canonicalizations, NULL);
		register_callback(plugin_name, PLUGIN_START_UNIT, size_overflow_start_unit, NULL);
		register_callback(plugin_name, PLUGIN_REGISTER_GGC_ROOTS, NULL, (void *)&gt_ggc_r_size_overflow);
		flag_sanitize |= SANITIZE_SI_OVERFLOW;
		register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &disable_ubsan_si_overflow_pass_info);
		register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &insert_size_overflow_asm_pass_info);

#if BUILDING_GCC_VERSION >= 13000
		reference_pass = plugin_find_pass("respectre");
		if (!reference_pass)
#endif
			reference_pass = plugin_find_pass("autoslab");
		if (reference_pass)
			size_overflow_transform_pass_info.reference_pass_name = reference_pass->name;

		if (!enable_ipa_transform)
			register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &size_overflow_transform_pass_info);
		register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &size_overflow_pass_info);
	}
	register_callback(plugin_name, PLUGIN_ATTRIBUTES, register_attributes, NULL);

	return 0;
}
