/*
 * Copyright 2022,2023,2024 by Open Source Security, Inc.
 *
 * Licensed under the GPL v2
 *
 * Implementation of the 'canary(mode[, count=1])' struct member attribute,
 * __builtin_canary_value(obj[, index=1]) and __builtin_copy_canary(dst, src)
 * built-in functions.
 *
 * The attribute needs to be attached to a struct member that is designated to
 * be a type's "canary" member which is used to verify an object's integrity at
 * runtime.
 *
 * The __builtin_canary_value() function shall be used to calculate an object's
 * canary value for object initialization or manual verification.
 *
 * The __builtin_copy_canary() function should be used to inherit an object's
 * canary value when creating an object copy, e.g. via memcpy(). It should be
 * preferred over using __builtin_canary_value() to not blindly make the new
 * object valid, in case the source object's canary is corrupted.
 *
 * Code making use of canary typed objects will get the first memory read
 * operation to such an object instrumented and the embedded canary member
 * checked to still have its "magic" value. If it doesn't, a call to
 * __builtin_trap() will happen.
 *
 * Write operations to canary typed objects get ignored -- on purpose, to not
 * require special wrappers for canary initialization for dynamically allocated
 * objects.
 *
 * Canaries come in two forms: 'static' and 'dynamic'. The former uses a fixed,
 * compile time generated random value. The latter does the same, but XORs an
 * object's address on top for a more diverse magic value at a slightly higher
 * runtime cost.
 *
 * Objects may have multiple valid canary values. The exact number gets set by
 * the 'canary' attribute's second argument, which needs to be an integer
 * between 1 and 16 (which is an arbitrary but sensible maximum for the current
 * usage). By default, i.e. if no explicit number is specified, only a single
 * canary value will be generated and used.
 *
 * A higher than 1 canary count is only needed when an object's life span has
 * multiple valid stages that need to be explicitly accounted for, e.g. an
 * object that was marked for release but might still be seen via RCU lookups.
 * Such "half dead" objects should no longer be seen in operational use, but
 * shouldn't fail the canary checks either. The transitional state should
 * therefore be marked differently than for live objects and using a different
 * canary value might just suit that need.
 *
 * Objects with multiple canary values don't get automatically instrumented and
 * require the 'noinstr' flag to be specified in the 'mode' argument of the
 * canary attribute. The plugin simply cannot know which values are allowed for
 * a given code sequence and therefore requires manual canary management and
 * verification making use of the __builtin_canary_value() helper function.
 *
 * Simple objects with only a binary life cycle (valid <-> invalid) just need a
 * single canary value which can be explicitly invalidated by assigning its
 * inverse to the canary member, e.g. as done for 'cred' objects:
 *
 *	cred->canary = ~__canary_value(cred);
 *
 * An example of a canary typed object making use of multiple value is 'struct
 * file' using three distinct values to mark the different stages of an
 * object's live cycle:
 *
 * 	# define FILE_MAGIC_ALIVE(f)   __canary_value((f), 1)
 * 	# define FILE_MAGIC_DYING(f)   __canary_value((f), 2)
 * 	# define FILE_MAGIC_DEAD(f)    __canary_value((f), 3)
 *
 *
 * Implementation details:
 *
 * We propagate the existence of a 'canary' member to the surrounding structure
 * type by attaching a '*has_canary' attribute to it. This attribute's
 * TREE_VALUE() holds the FIELD_DECL of the structure's canary member (which
 * may nest to another canary typed structure). The attribute's TREE_PURPOSE()
 * holds a TREE_VEC of the "canary control data" (generated from the 'canary'
 * attribute and flattened to a single integer_cst, followed by the canary
 * values). This propagation bubbles up all the way to the outer-most type to
 * ease the decision if a pointer dereference needs to be instrumented or not
 * and to build up the COMPONENT_REF for the inner-most canary member for a
 * given type.
 *
 * Attributes:
 * - 'canary':
 *   - takes one or two arguments:
 *     - first argument is the mode including flags (string)
 *     - second is the canary count (integer, default: 1)
 *   - only used to annotate the in-source canary member and set mode & count
 * - '*has_canary':
 *   - gets attached to RECORD_TYPE nodes that embed (nested) canary fields
 *   - has a single argument:
 *     - purpose: FIELD_DECL of the canary member (might be a RECORD_TYPE too)
 *     - value: TREE_VEC with 1 + 'canary count' elements (first is flattened
 *         canary_control_data (INTEGER_CST), followed by 'canary count' magic
 *         values, already typed to match the canary field_decl's type)
 *
 * TODO:
 * - add canary member automatically (annotate the type only)
 * - IPA pass to handle ADDR_EXPR call arguments
 * - don't instrument canary member reads
 */

#include "gcc-common.h"

static struct plugin_info type_canary_plugin_info = {
	.version	= "20230315",
	.help		= "verbose\t\tverbose mode\n"
};

__visible int plugin_is_GPL_compatible;

static int verbose;

enum canary_mode { invalid_mode, static_canary, dynamic_canary } __attribute__((packed));

struct canary_control_data {
	enum canary_mode mode;
	unsigned char count;
	unsigned char noinstr : 1;
};

#define CANARY_VALUES_MAX	16	// must fit struct canary_control_data::count!
STATIC_ASSERT(CANARY_VALUES_MAX <= UCHAR_MAX);

union canary_control {
	struct canary_control_data ccd;
	unsigned long _raw;
};
STATIC_ASSERT(sizeof(union canary_control) == sizeof(unsigned long));
STATIC_ASSERT(sizeof(union canary_control) <= sizeof(HOST_WIDE_INT));

/*** <from randomize_layout_plugin.c> ***/
#define ORIG_TYPE_NAME(node) \
	(TYPE_NAME(TYPE_MAIN_VARIANT(node)) != NULL_TREE ? \
	 ((const unsigned char *)TYPE_NAME_POINTER(TYPE_MAIN_VARIANT(node))) : \
	 (const unsigned char *)" anonymous ")

/*
 * 64bit variant of Bob Jenkins' public domain PRNG
 * 256 bits of internal state
 */

typedef unsigned long long u64;

typedef struct ranctx { u64 a; u64 b; u64 c; u64 d; } ranctx;

#define rot(x,k) (((x)<<(k))|((x)>>(64-(k))))
static u64 ranval(ranctx *x) {
	u64 e = x->a - rot(x->b, 7);
	x->a = x->b ^ rot(x->c, 13);
	x->b = x->c + rot(x->d, 37);
	x->c = x->d + e;
	x->d = e + x->a;
	return x->d;
}

static void raninit(ranctx *x, u64 *seed) {
	int i;

	x->a = seed[0];
	x->b = seed[1];
	x->c = seed[2];
	x->d = seed[3];

	for (i=0; i < 30; ++i)
		(void)ranval(x);
}

/*
 * TYPE_CANARY_SEED_INIT is from include/generated/type_canary.h, which gets
 * sourced via commandline.
 */
static u64 shuffle_seed[4] = TYPE_CANARY_SEED_INIT;

/* from old Linux dcache.h */
static inline unsigned long
partial_name_hash(unsigned long c, unsigned long prevhash)
{
	return (prevhash + (c << 4) + (c >> 4)) * 11;
}

static inline unsigned int
name_hash(const unsigned char *name)
{
	unsigned long hash = 0;
	unsigned int len = strlen((const char *)name);
	while (len--)
		hash = partial_name_hash(*name++, hash);
	return (unsigned int)hash;
}
/*** </from randomize_layout_plugin.c> ***/

static tree handle_canary_attr(tree *node, tree name, tree args, int flags, bool *no_add_attrs)
{
	const_tree decl = *node;
	location_t loc;

	*no_add_attrs = true;

	gcc_assert(DECL_P(decl));
	loc = DECL_SOURCE_LOCATION(decl);

	if (TREE_CODE(decl) != FIELD_DECL) {
		error_at(loc, "attribute %qE applies to structure members only", name);
		return NULL_TREE;
	}

	if (TREE_CODE(TREE_TYPE(decl)) != INTEGER_TYPE) {
		error_at(loc, "attribute %qE applies to integer types only", name);
		return NULL_TREE;
	}

	/*
	 * Further checks like verifying the upper type is a RECORD_TYPE (instead
	 * of a UNION_TYPE) need to be deferred to finish_type(), when the type is
	 * complete. This is where we check and parse the attribute's arguments as
	 * well.
	 */
	*no_add_attrs = false;

	return NULL_TREE;
}

static struct attribute_spec canary_attr_spec = ATTRIBUTE_SPEC_INIT("canary", 1, 2, true, false, false, handle_canary_attr, true, NULL);

static void register_attributes(void *event_data, void *data)
{
	register_attribute(&canary_attr_spec);
}

static bool canary_type_p(const_tree type)
{
	gcc_assert(TYPE_P(type));
	return !!lookup_attribute("*has_canary", TYPE_ATTRIBUTES(type));
}

/*
 * Return the tree_list of a type's '*has_canary' attribute, if it has one,
 * NULL_TREE otherwise.
 *
 * Its TREE_PURPOSE() is the canary member's FIELD_DECL, its TREE_VALUE() is
 * yet another tree_list with a TREE_VEC holding the canary control data (see
 * get_canary_attr_control()).
 */
static tree get_canary_attr_list(const_tree type)
{
	tree attr;

	gcc_assert(TYPE_P(type));
	attr = lookup_attribute("*has_canary", TYPE_ATTRIBUTES(type));
	if (!attr)
		return NULL_TREE;

	return TREE_VALUE(attr);
}

/* The FIELD_DECL for the (possibly nested) canary member */
static tree get_canary_attr_field(const_tree type)
{
	tree attr;

	gcc_assert(canary_type_p(type));

	attr = get_canary_attr_list(type);
	gcc_assert(attr != NULL_TREE && TREE_CODE(attr) == TREE_LIST);

	return TREE_PURPOSE(attr);
}

/*
 * The canary control data is a TREE_VEC with the following layout:
 * [0]: flattened 'struct canary_control_data', stuffed in an INTEGER_CST by
 *      set_canary_attr()
 * [1..]: canary magic values (up to ccd.count)
 */
static tree get_canary_attr_control(const_tree type)
{
	tree attr;

	gcc_assert(canary_type_p(type));

	attr = get_canary_attr_list(type);
	gcc_assert(attr != NULL_TREE && TREE_CODE(attr) == TREE_LIST);

	return TREE_VALUE(attr);
}

static struct canary_control_data canary_control_2_ccd(const_tree ctrl)
{
	union canary_control ctl = { };

	gcc_assert(ctrl != NULL_TREE && TREE_CODE(ctrl) == TREE_VEC);
	gcc_assert(TREE_VEC_LENGTH(ctrl) >= 1);

	ctrl = TREE_VEC_ELT(ctrl, 0);
	gcc_assert(ctrl != NULL_TREE && TREE_CODE(ctrl) == INTEGER_CST);

	ctl._raw = tree_to_uhwi(ctrl);

	return ctl.ccd;
}

static struct canary_control_data get_canary_ccd(const_tree type)
{
	return canary_control_2_ccd(get_canary_attr_control(type));
}

static bool canary_type_needs_instrumentation(const_tree type)
{
	struct canary_control_data ccd = get_canary_ccd(type);
	return !ccd.noinstr;
}

static bool set_canary_mode(location_t loc, const_tree name, struct canary_control_data *ccd, enum canary_mode mode)
{
	if (ccd->mode == invalid_mode) {
		ccd->mode = mode;
		return true;
	}

	if (ccd->mode == mode) {
		warning_at(loc, OPT_Wattributes, "duplicated mode in %qE attribute argument", name);
		return true;
	}

	error_at(loc, "mode already set for %qE attribute", name);
	return false;
}

static bool parse_canary_mode(location_t loc, const_tree name, const_tree mode_str, struct canary_control_data *ccd)
{
	const char *str, *tok;
	bool res = true;
	int len;

	gcc_assert(TREE_CODE(mode_str) == STRING_CST);
	str = TREE_STRING_POINTER(mode_str);

	do {
		tok = str;
#ifdef _GNU_SOURCE
		str = strchrnul(tok, ',');
#else
		str = strchr(tok, ',') ?: tok + strlen(tok);
#endif
		len = str - tok;

		switch (len) {
		case 0:
			break;

		case 6:
			if (memcmp(tok, "static", len) == 0)
				res = set_canary_mode(loc, name, ccd, static_canary) && res;
			else
				goto unknown_flag;
			break;

		case 7:
			if (memcmp(tok, "dynamic", len) == 0)
				res = set_canary_mode(loc, name, ccd, dynamic_canary) && res;
			else if (memcmp(tok, "noinstr", len) == 0) {
				if (ccd->noinstr)
					warning_at(loc, OPT_Wattributes, "duplicated %<noinstr%> in %qE attribute argument", name);
				ccd->noinstr = 1;
			} else
				goto unknown_flag;
			break;

		default:
		unknown_flag:
			error_at(loc, "unknown option %<%.*s%> for %qE attribute", len, tok, name);
			res = false;
		}

		while (*str == ',')
			str++;
	} while (*str);

	if (ccd->mode == invalid_mode) {
		error_at(loc, "missing mode for %qE attribute", name);
		res = false;
	}

	return res;
}

static tree generate_canary_control(const_tree type, const_tree decl, const_tree attr)
{
	union canary_control ctl = { };
	tree canary_type, canary_vec;
	const_tree name, args, arg;
	unsigned int i, count = 1;
	ranctx prng_state;
	location_t loc;
	u64 seed[4];

	loc = DECL_SOURCE_LOCATION(decl);
	name = TREE_PURPOSE(attr);

	canary_type = TREE_TYPE(decl);
	gcc_assert(TREE_CODE(canary_type) == INTEGER_TYPE);

	/* Verify and parse 'canary' attribute arguments */
	args = TREE_VALUE(attr);
	gcc_assert(args && TREE_CODE(args) == TREE_LIST);

	/* required 'mode' argument */
	arg = TREE_VALUE(args);
	if (arg == NULL_TREE || TREE_CODE(arg) != STRING_CST) {
		error_at(loc, "attribute %qE takes a string constant as its first argument", name);
		return NULL_TREE;
	}

	if (!parse_canary_mode(loc, name, arg, &ctl.ccd))
		return NULL_TREE;

	/* optional 'count' argument */
	arg = TREE_CHAIN(args) ? TREE_VALUE(TREE_CHAIN(args)) : NULL_TREE;
	if (arg) {
		if (TREE_CODE(arg) != INTEGER_CST) {
			error_at(loc, "attribute %qE takes an integer constant as its second argument", name);
			return NULL_TREE;
		}

		if (( TYPE_UNSIGNED(TREE_TYPE(arg)) && (tree_to_uhwi(arg) < 1 || tree_to_uhwi(arg) > CANARY_VALUES_MAX)) ||
		    (!TYPE_UNSIGNED(TREE_TYPE(arg)) && (tree_to_shwi(arg) < 1 || tree_to_shwi(arg) > CANARY_VALUES_MAX)))
		{
			error_at(loc, "second argument of attribute %qE must be a positive integer less than %d", name, CANARY_VALUES_MAX);
			return NULL_TREE;
		}
		count = tree_to_uhwi(arg);
	}

	gcc_assert(count >= 1 && count <= CANARY_VALUES_MAX);

	if (count > 1 && !ctl.ccd.noinstr) {
		error_at(loc, "canary count above 1 requires %<noinstr%> flag for %qE attribute", name);
		return NULL_TREE;
	}

	ctl.ccd.count = count;
	canary_vec = make_tree_vec(count + 1);

	/* flatten and store canary control structure as first element */
	TREE_VEC_ELT(canary_vec, 0) = build_int_cstu(long_unsigned_type_node, ctl._raw);

	/* generate canary values */
	STATIC_ASSERT(ARRAY_SIZE(seed) == ARRAY_SIZE(shuffle_seed));
	for (i = 0; i < ARRAY_SIZE(seed); i++) {
		seed[i] = shuffle_seed[i];
		seed[i] ^= name_hash(ORIG_TYPE_NAME(type));
	}
	raninit(&prng_state, seed);

	for (i = 0; i < count; i++) {
		TREE_VEC_ELT(canary_vec, i + 1) = build_int_cst_type(canary_type, ranval(&prng_state));
		if (verbose >= 3)
			inform(loc, "%qT canary value %d: %qE", type, i+1, TREE_VEC_ELT(canary_vec, i + 1));
	}

	return canary_vec;
}

static void set_canary_attr(tree type, tree decl)
{
	tree canary_control = NULL_TREE;
	const_tree attr;

	gcc_assert(TYPE_P(type) && TREE_CODE(decl) == FIELD_DECL);
	gcc_assert(!lookup_attribute("*has_canary", TYPE_ATTRIBUTES(type)));

	attr = lookup_attribute("canary", DECL_ATTRIBUTES(decl));

	/*
	 * 'attr' will only be non-NULL for the real canary field. We use this
	 * as a trigger to do the remaining attribute parsing and canary value
	 * generation.
	 *
	 * If 'attr' is NULL, we take it from the '*has_canary' attribute a
	 * previous call to this function generated for decl's type.
	 */
	if (attr) {
		canary_control = generate_canary_control(type, decl, attr);
		if (canary_control == NULL_TREE)
			return;
	} else {
		canary_control = get_canary_attr_control(TREE_TYPE(decl));
	}

	gcc_assert(canary_control != NULL_TREE && TREE_CODE(canary_control) == TREE_VEC);

	add_type_attr(type, "*has_canary", tree_cons(decl, canary_control, NULL_TREE));
}

static void type_check_canary_attr(tree type)
{
	tree canary_field = NULL_TREE;
	tree field;

	/*
	 * We have to allow unions here too, to verify none of its members has the
	 * 'canary' attribute.
	 */
	if (!RECORD_OR_UNION_TYPE_P(type))
		return;

	/*
	 * Most of the sanity checks have already been done by handle_canary_attr().
	 * However, it only saw the annotated member itself, out of context. Do the
	 * remaining context checks here:
	 * - verify a given type has only one canary member at max and
	 * - a canary member isn't part of a union (even if it has only a single
	 *   member, to keep things simple).
	 */
	for (field = TYPE_FIELDS(type); field; field = TREE_CHAIN(field)) {
		tree field_type = TREE_TYPE(field);
		tree inner_canary = NULL_TREE;

		if (RECORD_OR_UNION_TYPE_P(field_type)) {
			if (!canary_type_p(field_type))
				continue;

			inner_canary = get_canary_attr_field(field_type);
		} else {
			if (!lookup_attribute("canary", DECL_ATTRIBUTES(field)))
				continue;

			gcc_assert(!AGGREGATE_TYPE_P(field_type));
			inner_canary = field;
		}

		gcc_assert(TREE_CODE(inner_canary) == FIELD_DECL);

		if (TREE_CODE(type) != RECORD_TYPE) {
			location_t loc = DECL_SOURCE_LOCATION(field);
			bool nested = inner_canary != field;
			const_tree name = type_name(type);

			if (name)
				error_at(loc, "%s%<canary%> %s cannot be part of %<union %E%>", nested ? "nested " : "", nested ? "type" : "member", name);
			else
				error_at(loc, "%s%<canary%> %s cannot be part of anonymous union", nested ? "nested " : "", nested ? "type" : "member");

			if (nested) {
				tree tmp;

				tmp = inner_canary;
				while (canary_type_p(TREE_TYPE(tmp))) {
					tmp = get_canary_attr_field(TREE_TYPE(tmp));
					inner_canary = tmp;
				}

				inform(DECL_SOURCE_LOCATION(inner_canary), "nested %<canary%> member is from %qT:", DECL_FIELD_CONTEXT(inner_canary));
			}

			continue;
		}

		if (canary_field) {
			error_at(DECL_SOURCE_LOCATION(field), "a type can have only one %<canary%> member");
			inform(DECL_SOURCE_LOCATION(canary_field), "previous use was here:");
			continue;
		}

		canary_field = field;
	}

	/*
	 * The inner-most canary field will automatically trigger the generation of
	 * all required canary values, as, by then, all sanity checks have passed.
	 */
	if (canary_field)
		set_canary_attr(type, canary_field);
}

/*
 * We could have used an unused C type flag like TYPE_LANG_FLAG_6 (_4 and _5
 * already taken by CONSTIFY and STRUCTLEAK) but that doesn't really scale with
 * the rate we keep on writing new plugins. ;)  So we use a bitmap instead!
 */
static TYPEMAP(canary_typemap);

static void finish_type(void *event_data, void *data)
{
	tree type = (tree)event_data;

	if (type == NULL_TREE || type == error_mark_node)
		return;

	gcc_assert(TYPE_P(type));

	/* Ignore forward declarations. */
	if (!COMPLETE_TYPE_P(type))
		return;

	/* We see types multiple times, just walk them once. */
	if (typemap_has(canary_typemap, type))
		return;

	type_check_canary_attr(type);

	/* Skip this type next time we see it. */
	typemap_add(canary_typemap, type);
}

static void typemap_cleanup(void *event_data, void *data)
{
	typemap_free(&canary_typemap);
}

static tree build_canary_ref(location_t loc, tree type, tree ref)
{
	tree field;

	/*
	 * Build the access chain to reach the possibly nested canary member from
	 * the top level object type.
	 */
	while (canary_type_p(type)) {
		field = get_canary_attr_field(type);
		type = TREE_TYPE(field);
		ref = build3_loc(loc, COMPONENT_REF, type, ref, field, NULL_TREE);
	}

	return ref;
}

static tree canary_value_builtin(location_t loc, tree decl, void *_args)
{
	vec<tree, va_gc> *args = static_cast<vec<tree, va_gc> *>(_args);
	unsigned int arg_cnt = args ? args->length() : 0;
	struct canary_control_data ccd;
	unsigned int canary_slot = 1;
	tree obj, type, canary;
	const_tree canary_ctl;

	if (arg_cnt < 1) {
		error_at(loc, "too few arguments to function %qD", decl);
		return NULL_TREE;
	}

	if (arg_cnt > 2) {
		error_at(loc, "too many arguments to function %qD", decl);
		return NULL_TREE;
	}

	obj = VEC_index(tree, args, 0);
	type = TREE_TYPE(obj);
	if (!POINTER_TYPE_P(type) || !canary_type_p(TREE_TYPE(type))) {
		loc = EXPR_LOCATION(obj) ?: loc;
		error_at(loc, "argument 1 of %qD must be a %<canary%> object pointer", decl);
		return NULL_TREE;
	}

	/* strip off pointer type */
	type = TREE_TYPE(type);

	canary_ctl = get_canary_attr_control(type);
	ccd = canary_control_2_ccd(canary_ctl);

	if (arg_cnt == 2) {
		const_tree cnt;

		cnt = VEC_index(tree, args, 1);
		if (TREE_CODE(cnt) != INTEGER_CST ||
			tree_to_shwi(cnt) < 1 || tree_to_uhwi(cnt) > ccd.count)
		{
			loc = EXPR_LOC_OR_LOC(cnt, loc);
			error_at(loc, "argument 2 of %qD must be a positive integer less than %u", decl, ccd.count);
			return NULL_TREE;
		}
		canary_slot = tree_to_uhwi(cnt);
	}

	gcc_assert(canary_slot >= 1 && canary_slot <= ccd.count);
	canary = TREE_VEC_ELT(canary_ctl, canary_slot);

	switch (ccd.mode) {
	case static_canary:
		/* fixed canary value, not taking obj's address into account */
		return canary;

	case dynamic_canary: {
		tree addr;

		/*
		 * Build the equivalent of the following C expression:
		 *
		 *   (typeof CANARY_MAGIC)(unsigned long)&obj->foo.bar.canary ^ CANARY_MAGIC;
		 */
		addr = build_indirect_ref(loc, obj, RO_ARROW);	// obj ptr deref
		addr = build_canary_ref(loc, type, addr);	// component_ref for canary member
		addr = build1_loc(loc, ADDR_EXPR, TYPE_POINTER_TO(TREE_TYPE(addr)), addr); // take address
		addr = build1_loc(loc, NOP_EXPR, long_unsigned_type_node, addr);	// cast to unsigned long
		addr = build1_loc(loc, NOP_EXPR, TREE_TYPE(canary), addr);		// cast to canary member type

		return build_binary_op(loc, BIT_XOR_EXPR, addr, canary, false);	// addr ^ canary
		}

	default:
		break;
	}

	gcc_unreachable();
	return NULL_TREE;
}

static tree canary_copy_builtin(location_t loc, tree decl, void *_args)
{
	vec<tree, va_gc> *args = static_cast<vec<tree, va_gc> *>(_args);
	unsigned int arg_cnt = args ? args->length() : 0;
	tree src, dst, src_type, dst_type, src_canary, dst_canary;
	struct canary_control_data ccd;

	if (arg_cnt != 2) {
		error_at(loc, "too %s arguments to function %qD, need 2", arg_cnt < 2 ? "few" : "many", decl);
		return NULL_TREE;
	}

	dst = VEC_index(tree, args, 0);
	dst_type = TREE_TYPE(dst);
	if (!POINTER_TYPE_P(dst_type) || !canary_type_p(TREE_TYPE(dst_type))) {
		loc = EXPR_LOCATION(dst) ?: loc;
		error_at(loc, "argument 1 of %qD must be a %<canary%> object pointer", decl);
		return NULL_TREE;
	}
	dst_type = TREE_TYPE(dst_type);

	src = VEC_index(tree, args, 1);
	src_type = TREE_TYPE(src);
	if (!POINTER_TYPE_P(src_type) || !canary_type_p(TREE_TYPE(src_type))) {
		loc = EXPR_LOCATION(src) ?: loc;
		error_at(loc, "argument 2 of %qD must be a %<canary%> object pointer", decl);
		return NULL_TREE;
	}
	src_type = TREE_TYPE(src_type);

	if (TYPE_MAIN_VARIANT(dst_type) != TYPE_MAIN_VARIANT(src_type)) {
		error_at(loc, "cannot copy between incompatible %<canary%> types for %qD (%qT and %qT)", decl, TREE_TYPE(dst), TREE_TYPE(src));
		return NULL_TREE;
	}

	// src->canary
	src_canary = build_indirect_ref(loc, src, RO_ARROW);
	src_canary = build_canary_ref(loc, src_type, src_canary);

	// dst->canary
	dst_canary = build_indirect_ref(loc, dst, RO_ARROW);
	dst_canary = build_canary_ref(loc, dst_type, dst_canary);

	ccd = get_canary_ccd(dst_type);
	switch (ccd.mode) {
	case static_canary:
		/* dst->canary = src->canary; */
		return fold_build2_loc(loc, MODIFY_EXPR, TREE_TYPE(dst_canary), dst_canary, src_canary);

	case dynamic_canary: {
		tree xop;

		/*
		 * Build the equivalent of the following C expression:
		 *
		 *   dst->canary = src->canary ^
		 *   	(typeof src->canary)(unsigned long)&src->canary ^
		 *   	(typeof dst->canary)(unsigned long)&dst->canary;
		 */
		src = build1_loc(loc, ADDR_EXPR, TYPE_POINTER_TO(TREE_TYPE(src_canary)), src_canary); // take address
		src = build1_loc(loc, NOP_EXPR, long_unsigned_type_node, src);	// cast to unsigned long
		src = build1_loc(loc, NOP_EXPR, TREE_TYPE(src_canary), src);	// cast to canary member type

		dst = build1_loc(loc, ADDR_EXPR, TYPE_POINTER_TO(TREE_TYPE(dst_canary)), dst_canary); // take address
		dst = build1_loc(loc, NOP_EXPR, long_unsigned_type_node, dst);	// cast to unsigned long
		dst = build1_loc(loc, NOP_EXPR, TREE_TYPE(dst_canary), dst);	// cast to canary member type

		xop = build_binary_op(loc, BIT_XOR_EXPR, src, dst, false);	// src ^ dst
		xop = build_binary_op(loc, BIT_XOR_EXPR, xop, src_canary, false); // diff ^ canary

		return fold_build2_loc(loc, MODIFY_EXPR, TREE_TYPE(dst_canary), dst_canary, xop);
		}

	default:
		break;
	}

	gcc_unreachable();
	return NULL_TREE;
}

static GTY(()) tree canary_copy_builtin_decl;
static GTY(()) tree canary_value_builtin_decl;
static tree (*orig_resolve_overloaded_builtin)(unsigned int, tree, void *);

static tree canary_resolve_overloaded_builtin(unsigned int loc, tree decl, void *args)
{
	if (decl == canary_value_builtin_decl)
		return canary_value_builtin(loc, decl, args);

	if (decl == canary_copy_builtin_decl)
		return canary_copy_builtin(loc, decl, args);

	if (orig_resolve_overloaded_builtin)
		return orig_resolve_overloaded_builtin(loc, decl, args);

	return NULL_TREE;
}

static void register_builtins(void *event_data, void *data)
{
	tree fntype;
	tree attr;

	/*
	 * Create a stub prototype for our builtin to please the parser:
	 *
	 *   unsigned long __builtin_canary_value(const void *obj_ptr[, index])
	 *		__attribute__((nothrow,const,leaf));
	 *
	 * The return type isn't necessarily 'unsigned long' but will be
	 * inherited from obj's canary member type during code generation.
	 *
	 * The attributes aren't really needed either, as we provide the
	 * implementation already during parsing -- i.e. the call won't be
	 * even visible to later passes -- but for completeness sake.
	 */
	fntype = build_varargs_function_type_list(long_unsigned_type_node, const_ptr_type_node, NULL_TREE);

	attr = tree_cons(get_identifier("nothrow"), NULL_TREE, NULL_TREE);
	attr = tree_cons(get_identifier("leaf"), NULL_TREE, attr);
	attr = tree_cons(get_identifier("const"), NULL_TREE, attr);
	// "type generic", too?

	canary_value_builtin_decl = add_builtin_function_ext_scope("__builtin_canary_value", fntype, BUILT_IN_CANARY_VALUE, BUILT_IN_MD, NULL, attr);

	/* void __builtin_canary_copy(void *dst, const void *src); */
	fntype = build_function_type_list(void_type_node, ptr_type_node, const_ptr_type_node, NULL_TREE);

	// "type generic" as attr?
	canary_copy_builtin_decl = add_builtin_function_ext_scope("__builtin_canary_copy", fntype, BUILT_IN_CANARY_COPY, BUILT_IN_MD, NULL, NULL_TREE);

	orig_resolve_overloaded_builtin = targetm.resolve_overloaded_builtin;
	targetm.resolve_overloaded_builtin = canary_resolve_overloaded_builtin;

	/* allocate type bitmap */
	canary_typemap = typemap_init();
}

static tree canary_for_ref(tree mem_ref, location_t loc)
{
	tree cref, type;

	gcc_assert(TREE_CODE(mem_ref) == MEM_REF);

	cref = unshare_expr(mem_ref);
	type = TREE_TYPE(mem_ref);

	/*
	 * find_uninstrumented_ptr() might run into narrowed MEM_REFs that use a
	 * narrow type.
	 * Detect these and upcast the unshared MEM_REF to the canary type for
	 * building the canary member chain.
	 */
	if (!canary_type_p(type)) {
		tree op, ptype;

		op = TREE_OPERAND(cref, 0);
		gcc_assert(TREE_CODE(op) == SSA_NAME);

		ptype = TREE_TYPE(op);
		gcc_assert(POINTER_TYPE_P(ptype));

		// change the mem_ref to have the canary type
		type = TREE_TYPE(ptype);
		TREE_TYPE(cref) = type;
		TYPE_READONLY(type) = 1;

		// change the offset to zero for building up the field_decl chain below
		TREE_OPERAND(cref, 1) = build_int_cst(ptype, 0);
	}

//	fprintf(stderr, ">>> mref = "); debug_tree(mem_ref);
//	fprintf(stderr, ">>> cref = "); debug_tree(cref);

	gcc_assert(canary_type_p(type));

	return build_canary_ref(loc, type, cref);
}

static void gsi_insert_loc(gimple_stmt_iterator *gsi, gimple_stmt_ptr stmt, location_t loc, tree block, enum gsi_iterator_update gsi_link_mode, bool after)
{
	gimple_set_location(stmt, loc);
	gimple_set_block(stmt, block);
	if (after)
		gsi_insert_after(gsi, stmt, gsi_link_mode);
	else
		gsi_insert_before(gsi, stmt, gsi_link_mode);
	update_stmt(stmt);
}

static gimple_stmt_ptr build_canary_check_assign(gimple_stmt_iterator *gsi, tree var, location_t loc, tree block, enum gsi_iterator_update gsi_link_mode, bool after)
{
	gimple_stmt_ptr stmt;
	tree tmp_var;

	tmp_var = make_temp_ssa_name(TREE_TYPE(var), NULL, "canary_check");
	stmt = gimple_build_assign(tmp_var, var);
	gsi_insert_loc(gsi, stmt, loc, block, gsi_link_mode, after);

	return stmt;
}

static void canary_add_check(gimple_stmt_iterator *gsi, tree ref)
{
	struct canary_control_data ccd;
	tree canary, magic_val, tmp_var;
	gimple_stmt_ptr stmt, canary_check;
	basic_block cond_bb, join_bb, abort_bb;
	const_tree canary_ctl;
	location_t loc;
	tree block;
	edge e;

	stmt = gsi_stmt(*gsi);
	loc = EXPR_LOC_OR_LOC(ref, gimple_location(stmt));
	block = gimple_block(stmt);

	if (verbose)
		inform(loc, "instrumenting %<canary%> type'd read%s", loc != UNKNOWN_LOCATION ? " at:" : "");

//	fprintf(stderr, "%s: stmt: ", __func__); debug_gimple_stmt(stmt);

	/*
	 * gsi / stmt points to an instruction dereferencing a canary typed object.
	 * Insert a canary check in front of it:
	 *
	 * - for 'static' canaries:
	 *   > typeof(var->canary) tmp = var->canary;
	 *   > if (unlikely(tmp != CANARY_MAGIC))
	 *   > 	__buildin_trap();
	 *
	 * - for 'dynamic' canaries:
	 *   > typeof(var->canary) tmp = var->canary;
	 *   > if (unlikely(tmp != CANARY_MAGIC ^ (ulong)&var->canary))
	 *   > 	__buildin_trap();
	 */
	canary = canary_for_ref(ref, loc);

	gcc_assert(TREE_CODE(canary) == COMPONENT_REF);
	gcc_assert(TREE_CODE(TREE_OPERAND(canary, 1)) == FIELD_DECL);
	gcc_assert(TREE_CODE(TREE_TYPE(canary)) == INTEGER_TYPE);

	canary_ctl = get_canary_attr_control(DECL_FIELD_CONTEXT(TREE_OPERAND(canary, 1)));
	ccd = canary_control_2_ccd(canary_ctl);

	gcc_assert(!ccd.noinstr);
	gcc_assert(ccd.count == 1);

	magic_val = TREE_VEC_ELT(canary_ctl, 1);
	gcc_assert(TREE_CODE(TREE_TYPE(magic_val)) == INTEGER_TYPE);

	canary_check = build_canary_check_assign(gsi, canary, loc, block, GSI_CONTINUE_LINKING, false);
//	fprintf(stderr, "%s: canary_check(load) = ", __func__); debug_gimple_stmt(canary_check);

	/*
	 * Instead of computing '(typeof MAGIC)(unsigned long)&foo->cred ^ MAGIC'
	 * and comparing it against 'foo->cred' for the 'dynamic' mode, we compute
	 * 'foo->cred ^ (typeof foo->cred)(unsigned long)&foo->cred' and compare it
	 * against MAGIC. It leads to the same result, as 'typeof MAGIC' is the
	 * same as 'typeof foo->cred', but allows us to share some code between
	 * 'dynamic' and 'static' modes.
	 */
	if (ccd.mode == dynamic_canary) {
		gimple_stmt_ptr canary_addr;
		tree addr;

		canary = unshare_expr(canary);

		// take address
		addr = build_fold_addr_expr_loc(loc, canary);
		canary_addr = build_canary_check_assign(gsi, addr, loc, block, GSI_CONTINUE_LINKING, true);
//		fprintf(stderr, "%s: canary_check(addr) = ", __func__); debug_gimple_stmt(canary_addr);

		// cast to unsigned long
		addr = fold_convert_loc(loc, long_unsigned_type_node, gimple_assign_lhs(canary_addr));
		canary_addr = build_canary_check_assign(gsi, addr, loc, block, GSI_CONTINUE_LINKING, true);
//		fprintf(stderr, "%s: canary_check(addr) = ", __func__); debug_gimple_stmt(canary_addr);

		// cast to canary member type
		addr = fold_convert_loc(loc, TREE_TYPE(canary), gimple_assign_lhs(canary_addr));
		canary_addr = build_canary_check_assign(gsi, addr, loc, block, GSI_CONTINUE_LINKING, true);
//		fprintf(stderr, "%s: canary_check(addr) = ", __func__); debug_gimple_stmt(canary_addr);

		// XOR canary value with its address, store in 'canary_check'
		tmp_var = make_temp_ssa_name(TREE_TYPE(magic_val), NULL, "canary_check");
		canary_check = gimple_build_assign_with_ops(BIT_XOR_EXPR, tmp_var, gimple_assign_lhs(canary_addr), gimple_assign_lhs(canary_check));
		gsi_insert_loc(gsi, canary_check, loc, block, GSI_CONTINUE_LINKING, true);
//		fprintf(stderr, "%s: canary_check(xor) = ", __func__); debug_gimple_stmt(canary_check);
	}

	canary_check = gimple_build_cond(NE_EXPR, gimple_assign_lhs(canary_check), magic_val, NULL_TREE, NULL_TREE);
	gsi_insert_loc(gsi, canary_check, loc, block, GSI_NEW_STMT, true);
//	fprintf(stderr, "%s: canary_check(cmp) = ", __func__); debug_gimple_stmt(canary_check);

	cond_bb = gimple_bb(gsi_stmt(*gsi));
	gcc_assert(!gsi_end_p(*gsi));
	gcc_assert(canary_check == gsi_stmt(*gsi));

	e = split_block(cond_bb, gsi_stmt(*gsi));
	gcc_assert(cond_bb == e->src);
	join_bb = e->dest;
	e->flags = EDGE_FALSE_VALUE;
	e->probability -= REG_BR_PROB_VERY_UNLIKELY;

	abort_bb = create_empty_bb(join_bb);
	e = make_edge(cond_bb, abort_bb, EDGE_TRUE_VALUE);
	e->probability = REG_BR_PROB_VERY_UNLIKELY;
	abort_bb->count = edge_count(e);

	gcc_assert(dom_info_available_p(CDI_DOMINATORS));
	set_immediate_dominator(CDI_DOMINATORS, abort_bb, cond_bb);
	set_immediate_dominator(CDI_DOMINATORS, join_bb, cond_bb);

	gcc_assert(cond_bb->loop_father == join_bb->loop_father);
	add_bb_to_loop(abort_bb, cond_bb->loop_father);
	loops_state_set(LOOPS_NEED_FIXUP);

	*gsi = gsi_start_bb(abort_bb);

	canary_check = gimple_build_call(builtin_decl_implicit(BUILT_IN_TRAP), 0);
	gsi_insert_loc(gsi, canary_check, loc, block, GSI_CONTINUE_LINKING, true);
//	fprintf(stderr, "%s: canary_check(trap) = ", __func__); debug_gimple_stmt(canary_check);

	*gsi = gsi_start_nondebug_after_labels_bb(join_bb);
}

static tree find_uninstrumented_ptr(tree *tp, int *walk_subtrees, void *data)
{
	tree res = NULL_TREE;
	tree op = *tp;

	/* Look for MEM_REFs to canary'ed types. */
	switch (TREE_CODE(op)) {
	default:
		error_at(EXPR_LOCATION(op), "canary:%s(): unhandled tree type %qs", __func__, get_tree_code_name(TREE_CODE(op)));
		debug_tree(op);
		break;

	case MEM_REF:
	case TARGET_MEM_REF: {
		tree arg, type;

		type = TREE_TYPE(op);
		if (canary_type_p(type)) {
			// TODO: exclude canary member reads as well
			if (!canary_type_needs_instrumentation(type))
				break;

			res = op;
			break;
		}

		/*
		 * Look for a possibly offset MEM_REF using a narrowed type but based
		 * on an SSA_NAME that needs instrumentation.
		 */

		arg = TREE_OPERAND(op, 0);
		if (TREE_CODE(arg) != SSA_NAME)
			break;

		type = TREE_TYPE(arg);
		if (!POINTER_TYPE_P(type))
			break;

		type = TREE_TYPE(type);
		if (!canary_type_p(type))
			break;

		// TODO: exclude canary member reads as well
		if (!canary_type_needs_instrumentation(type))
			break;

		/* A narrowed MEM_REF -- will handle the upcast in canary_for_ref(). */
//		inform(EXPR_LOCATION(op), "narrowed %<canary%> MEM_REF found");
//		fprintf(stderr, "op: "); debug_tree(op);
//		fprintf(stderr, "type: "); debug_tree(type);
		res = op;
		break;
	}

	case ARRAY_REF:
	case BIT_FIELD_REF:
	case COMPONENT_REF:
	case POINTER_PLUS_EXPR:
	case VIEW_CONVERT_EXPR:
		/* nodes we need to dig in -- the first operand only, though */
		break;

	case ADDR_EXPR:
		/*
		 * Taking the address of an object only requires instrumentation if
		 * that address gets dereferenced for reading later on. Within the
		 * same function we'd see those as MEM_REFs, which will be handled.
		 * The only two exceptions would be call arguments passed to other
		 * functions and asm statements.
		 *
		 * We could handle the latter but these would likely only catch
		 * atomic ops and there's no real point to, as we care about "real"
		 * uses of members -- not the refcounting of the object.
		 *
		 * Handling call arguments would need an IPA pass to know if the
		 * passed address gets read or written to. So far we have no need
		 * for it for the object types handled by the 'canary' attribute.
		 */
		/* fall-through */
	case CONSTRUCTOR:
	case EQ_EXPR:
	case NE_EXPR:
	case GE_EXPR:
	case GT_EXPR:
	case LE_EXPR:
	case LT_EXPR:
	case VAR_DECL:
	case PARM_DECL:
	case RESULT_DECL:
	case FUNCTION_DECL:
	case FIELD_DECL:
	case LABEL_DECL:
	case CASE_LABEL_EXPR:
	case INTEGER_CST:
	case STRING_CST:
	case VECTOR_CST:
#if BUILDING_GCC_VERSION > 8000
	case POLY_INT_CST:
#endif
	case FIXED_CST:
	case REAL_CST:
	case SSA_NAME:
	case REALPART_EXPR:	// (ab)used by __builtin_*_overflow()
	case IMAGPART_EXPR:	//          -- " --
	case COMPLEX_CST:	//          -- " --
		/* uninteresting trees */
		*walk_subtrees = 0;
		break;
	}

	return res;
}

static bool canary_object_p(tree obj, bool ptr)
{
	switch (TREE_CODE(obj)) {
	case MEM_REF:
	case TARGET_MEM_REF:
	case SSA_NAME:
		break;

	default:
		return false;
	}

	gcc_assert(!ptr || POINTER_TYPE_P(TREE_TYPE(obj)));

	return canary_type_p(ptr ? TREE_TYPE(TREE_TYPE(obj)) : TREE_TYPE(obj));

}

static bool canary_ref_use_p(const_tree ref, const_tree ssa, bool is_canary_field)
{
	tree type, off;

	gcc_assert(TREE_CODE(ssa) == SSA_NAME);

	if (TREE_CODE(ref) != COMPONENT_REF)
		return false;

	off = TREE_OPERAND(ref, 1);
	ref = TREE_OPERAND(ref, 0);

	if (TREE_CODE(ref) != MEM_REF)
		return false;

	type = TREE_TYPE(ref);
	if (!canary_type_p(type))
		return false;

	ref = TREE_OPERAND(ref, 0);
	gcc_assert(TREE_CODE(ref) == SSA_NAME);

	if (ref != ssa)
		return false;

	if (TREE_CODE(off) != FIELD_DECL)
		return false;

	if (is_canary_field && off != get_canary_attr_field(type))
		return false;

	return true;
}

static tree get_ref_ssa_name(tree ref)
{
	switch (TREE_CODE(ref)) {
	case MEM_REF:
	case TARGET_MEM_REF:
		ref = TREE_OPERAND(ref, 0);
		break;
	default:
		break;
	}
	return ref;
}

static void check_canary_copy(gimple_stmt_iterator *gsi, tree lhs, tree rhs, bool is_memcpy)
{
	gimple_stmt_ptr copy_stmt, canary_stmt = NULL;
	bool saw_problematic_use, warned;
	struct canary_control_data ccd;
	location_t loc, loc_use;
	imm_use_iterator iter;
	use_operand_p use_p;
	tree type;

	if (!canary_object_p(lhs, is_memcpy) || !canary_object_p(rhs, is_memcpy))
		return;

	/*
	 * lhs and rhs types need to be the same or gcc would have complained
	 * already, so it doesn't matter which one we're looking at.
	 */
	type = is_memcpy ? TREE_TYPE(TREE_TYPE(lhs)) : TREE_TYPE(lhs);
	ccd = get_canary_ccd(type);

	/* fixed canary values need no update */
	if (ccd.mode == static_canary)
		return;

	/*
	 * look for uses of 'lhs' following the copy and warn if there is no canary
	 * update dominating it.
	 */
	lhs = get_ref_ssa_name(lhs);
//	debug_tree(lhs);
	gcc_assert(TREE_CODE(lhs) == SSA_NAME);

	/*
	 * Look for a possible canary update first to be able to check dominance
	 * relation later on.
	 */
	copy_stmt = gsi_stmt(*gsi);
	FOR_EACH_IMM_USE_FAST(use_p, iter, lhs) {
		gimple_stmt_ptr stmt = USE_STMT(use_p);

		if (stmt == copy_stmt)
			continue;

		if (!gimple_references_memory_p(stmt))
			continue;

		if (!stmt_dominates_stmt_p(copy_stmt, stmt))
			continue;

		switch (gimple_code(stmt)) {
		default:
			/* uninteresting statement */
			break;

		case GIMPLE_ASSIGN:
			if (canary_ref_use_p(gimple_assign_lhs(stmt), lhs, true)) {
				/*
				 * If we already have a different canary assignment statement,
				 * check if the new one dominates it and if so, use that one
				 * instead.
				 */
				if (!canary_stmt || stmt_dominates_stmt_p(stmt, canary_stmt)) {
					canary_stmt = stmt;
				 } else {
					/*
					 * We don't support (or need) canary updates in disjoint
					 * branches. Bail out loud if we ever see such.
					 */
					gcc_assert(stmt_dominates_stmt_p(canary_stmt, stmt));
				}
			}

			break;
		}
	}

	if (verbose && canary_stmt)
		inform(gimple_location(canary_stmt), "found %<canary%> member update following object copy");

	/*
	 * Look for reads of 'lhs' following the copy operation and preceding a
	 * possible canary update. Warn that this will trap!
	 */
	saw_problematic_use = false;
	loc_use = UNKNOWN_LOCATION;
	FOR_EACH_IMM_USE_FAST(use_p, iter, lhs) {
		gimple_stmt_ptr stmt = USE_STMT(use_p);

		if (stmt == copy_stmt)
			continue;

		if (!gimple_references_memory_p(stmt))
			continue;

		if (!stmt_dominates_stmt_p(copy_stmt, stmt))
			continue;

		if (canary_stmt && stmt_dominates_stmt_p(canary_stmt, stmt))
			continue;

//		print_gimple_stmt(stderr, stmt, 0, TDF_RAW);
		switch (gimple_code(stmt)) {
		default:
			/* "complex" statement */
			loc_use = gimple_location(stmt);
			/* Continue the search, we might find a more serve use. */
			break;

		case GIMPLE_ASM:
			/* XXX: Assume inline asm behaves -- as in only be used for atomic ops. */
			break;

		case GIMPLE_ASSIGN: {
			unsigned int i;

			// check RHSs only
			for (i = 1; i < gimple_num_ops(stmt); i++) {
				const_tree op = gimple_op(stmt, i);
				if (canary_ref_use_p(op, lhs, false)) {
					gcc_assert(TREE_CODE(op) == COMPONENT_REF);
					saw_problematic_use = true;
					loc_use = EXPR_LOC_OR_LOC(TREE_OPERAND(op, 0), gimple_location(stmt));
					goto break_loop;
				}
			}

			break;
			}
		}
	}

	/* If we saw a canary update and no problematic use inbetween, get out. */
	if (canary_stmt && !saw_problematic_use && loc_use == UNKNOWN_LOCATION)
		return;

break_loop:
	/*
	 * If we reached the end of the function (or saw a complex statement we
	 * didn't handle) without any further reference to 'lhs', emit a behaved
	 * warning only as it might be a false positive.
	 */
	rhs = get_ref_ssa_name(rhs);
	loc = gimple_location(gsi_stmt(*gsi));
	warned = warning_at(loc, OPT_Wattributes, "object copy %s %<canary%> update, missing call to %<%D(%E, %E)%>?",
	                    saw_problematic_use ? "needs" : "might lack", canary_copy_builtin_decl, lhs, rhs);
	if (warned && loc_use != UNKNOWN_LOCATION)
		inform(loc_use, "%<canary%> will be invalid when %s here:", saw_problematic_use ? "checked" : "used");
}

static bool canary_handle_stmt(gimple_stmt_iterator *gsi)
{
	gimple_stmt_ptr stmt = gsi_stmt(*gsi);
	bool changed = false;
	int i = -1;

	switch (gimple_code(stmt)) {
	default:
		error_at(gimple_location(stmt), "canary:%s(): unhandled statement %qs", __func__, gimple_code_name[gimple_code(stmt)]);
		debug_gimple_stmt(stmt);
		gcc_unreachable();
		break;

	case GIMPLE_ASM:
		/* XXX: Assume inline asm behaves -- as in only be used for atomic ops. */
		break;

	case GIMPLE_ASSIGN:
		// check if this is a canary object copy
		if (gimple_assign_single_p(stmt))
			check_canary_copy(gsi, gimple_assign_lhs(stmt), gimple_assign_rhs1(stmt), false);

		i = 1;	// check rhs operand(s)
		break;

	case GIMPLE_CALL: {
		tree fndecl;

		fndecl = gimple_call_fndecl(stmt);
		if (fndecl != NULL_TREE && gimple_call_num_args(stmt) == 3 &&
		    (!strcmp(DECL_NAME_POINTER(fndecl), "memcpy") || gimple_call_builtin_p(stmt, BUILT_IN_MEMCPY)))
		{
			// XXX: Ensure that arg3 covers the canary member to not warn
			// XXX: about constructs like "memcpy(new, old, 0);"
			check_canary_copy(gsi, gimple_call_arg(stmt, 0), gimple_call_arg(stmt, 1), true);
		}

		i = 3;	// check call arguments
		break;
		}

	case GIMPLE_RETURN:
		i = 0;	// check return value
		break;
	}

	if (i >= 0) {
		tree ptr, *op_ptr;

		/* Walk the operands and instrument where necessary. */
		for (; i < gimple_num_ops(stmt); i++) {
			op_ptr = gimple_op_ptr(stmt, i);
			ptr = walk_tree_without_duplicates(op_ptr, find_uninstrumented_ptr, NULL);
			if (ptr != NULL_TREE) {
				canary_add_check(gsi, ptr);
				changed = true;
			}
		}
	}

	return changed;
}

/*
 * This pass looks for statements involving memory references that are part of
 * or are based on a '*has_canary' marked type (which was propagated from the
 * initial in-source 'canary' attribute marking). If such a reference gets
 * dereferenced for a read, it gets instrumented with a canary check.
 *
 * We don't try to optimize multiple reads of the same memory reference and
 * rely on later passes to DCE unneeded ones. So far, this is working quite
 * well, as later passes are able to reason which statements may modify memory
 * and even move canary checks out of loops.
 *
 * Considering the following function and assuming that 'struct cred' has a
 * canary member:
 *
 *  10 bool test_creds(struct cred *cred)
 *  11 {
 *  12     const struct cred *new, *old;
 *  13     kuid_t uid;
 *  14     bool res;
 *  15
 *  16     new = prepare_kernel_cred(NULL);
 *  17     old = override_creds(new);
 *  18     revert_creds(old);
 *  18
 *  20     if (userns_may_setgroups(cred->user_ns))
 *  21         return true;
 *  22
 *  23     uid = cred->uid;
 *  24     res = uid_eq(uid, GLOBAL_ROOT_UID) && !(cred->securebits & SECBIT_NOROOT);
 *  25
 *  26     return res;
 *  27 }
 *
 * None of the 'cred' objects in lines 16-18 will be instrumented, as they do
 * not get dereferenced in this function -- only pointers get passed around.
 * If they are dereferenced in either override_creds() or revert_creds(), those
 * accesses will be instrumented in the corresponding function instead.
 *
 * Lines 20, 23 and 24 dereference the 'cred' object ('->user_ns' for the call
 * argument, '->uid' for the variable assignment and '->securebits' for the bit
 * test, resp.). All of these get instrumented, as can be seen in the following
 * gimple dump of the pass (slightly stripped down for readability):
 *
 * 100 test_creds (struct cred * cred)
 * 101 {
 * 102   unsigned int canary_check.1130;
 * 103   unsigned int canary_addr.1129;
 * 104   long unsigned int canary_addr.1128;
 * 105   unsigned int * canary_addr.1127;
 * 106   unsigned int canary_check.1126;
 * 107   unsigned int canary_check.1125;
 * 108   unsigned int canary_addr.1124;
 * 109   long unsigned int canary_addr.1123;
 * 110   unsigned int * canary_addr.1122;
 * 111   unsigned int canary_check.1121;
 * 112   unsigned int canary_check.1120;
 * 113   unsigned int canary_addr.1119;
 * 114   long unsigned int canary_addr.1118;
 * 115   unsigned int * canary_addr.1117;
 * 116   unsigned int canary_check.1116;
 * 117   unsigned int uid$val;
 * 118   bool res;
 * 119   const struct cred * old;
 * 120   const struct cred * new;
 * 121   struct user_namespace * _1;
 * 122   _Bool _2;
 * 123   unsigned int _3;
 * 124   unsigned int _4;
 * 125   bool _5;
 * 126   int iftmp.3_6;
 * 127
 * 128   <bb 2> [local count: 1073741824]:
 * 129   new_9 = prepare_kernel_cred (0B);
 * 130   old_11 = override_creds (new_9);
 * 131   revert_creds (old_11);
 * 132   canary_check.1116_18 = cred_13(D)->canary;
 * 133   canary_addr.1117_19 = &cred_13(D)->canary;
 * 134   canary_addr.1118_20 = (long unsigned int) canary_addr.1117_19;
 * 135   canary_addr.1119_21 = (unsigned int) canary_addr.1118_20;
 * 136   canary_check.1120_22 = canary_addr.1119_21 ^ canary_check.1116_18;
 * 137   if (canary_check.1120_22 != 1461575499)
 * 138     goto <bb 9>; [0.05%]
 * 139   else
 * 140     goto <bb 8>; [99.95%]
 * 141
 * 142   <bb 8> [local count: 1073741824]:
 * 143   _1 = cred_13(D)->user_ns;
 * 144   _2 = userns_may_setgroups (_1);
 * 145   if (_2 != 0)
 * 146     goto <bb 7>; [34.00%]
 * 147   else
 * 148     goto <bb 3>; [66.00%]
 * 149
 * 150   <bb 9> [local count: 536864]:
 * 151   __builtin_trap ();
 * 152
 * 153   <bb 3> [local count: 708669601]:
 * 154   canary_check.1121_24 = cred_13(D)->canary;
 * 155   canary_addr.1122_25 = &cred_13(D)->canary;
 * 156   canary_addr.1123_26 = (long unsigned int) canary_addr.1122_25;
 * 157   canary_addr.1124_27 = (unsigned int) canary_addr.1123_26;
 * 158   canary_check.1125_28 = canary_addr.1124_27 ^ canary_check.1121_24;
 * 159   if (canary_check.1125_28 != 1461575499)
 * 160     goto <bb 11>; [0.05%]
 * 161   else
 * 162     goto <bb 10>; [99.95%]
 * 163
 * 164   <bb 10> [local count: 708669601]:
 * 165   uid$val_17 = cred_13(D)->uid.val;
 * 166   if (uid$val_17 == 0)
 * 167     goto <bb 4>; [50.00%]
 * 168   else
 * 169     goto <bb 5>; [50.00%]
 * 170
 * 171   <bb 11> [local count: 354330]:
 * 172   __builtin_trap ();
 * 173
 * 174   <bb 4> [local count: 354334800]:
 * 175   canary_check.1126_30 = cred_13(D)->canary;
 * 176   canary_addr.1127_31 = &cred_13(D)->canary;
 * 177   canary_addr.1128_32 = (long unsigned int) canary_addr.1127_31;
 * 178   canary_addr.1129_33 = (unsigned int) canary_addr.1128_32;
 * 179   canary_check.1130_34 = canary_addr.1129_33 ^ canary_check.1126_30;
 * 180   if (canary_check.1130_34 != 1461575499)
 * 181     goto <bb 13>; [0.05%]
 * 182   else
 * 183     goto <bb 12>; [99.95%]
 * 184
 * 185   <bb 12> [local count: 354334800]:
 * 186   _3 = cred_13(D)->securebits;
 * 187   _4 = _3 & 1;
 * 188   if (_4 == 0)
 * 189     goto <bb 6>; [50.00%]
 * 190   else
 * 191     goto <bb 5>; [50.00%]
 * 192
 * 193   <bb 13> [local count: 177165]:
 * 194   __builtin_trap ();
 * 195
 * 196   <bb 5> [local count: 531502200]:
 * 197
 * 198   <bb 6> [local count: 708669601]:
 * 199   # iftmp.3_6 = PHI <1(12), 0(5)>
 * 200   res_15 = (bool) iftmp.3_6;
 * 201
 * 202   <bb 7> [local count: 1073741824]:
 * 203   # _5 = PHI <1(8), res_15(6)>
 * 204   return _5;
 * 205 }
 *
 * Lines 132-140 are the canary check for the '->user_ns' dereference of line
 * 20 in the source code. The canary value for the 'cred' object is calculated
 * by taking the value of the '->canary' member and XORing it against its
 * address. This value is then checked against a compile time random value
 * (which is 1461575499 in the above dump) to test the object's integrity.
 *
 * Lines 154-162 contain the check for the '->uid.val' load of the already
 * inlined and optimized call to uid_eq(). It's followed by yet another canary
 * check in lines 175-183 for the '->securebits' load.
 *
 * Even though the last check seems superfluous as there is no memory write
 * operation between the '->uid' and '->securebits' loads that could modify the
 * 'cred' object, there's also no need to do sophisticated data flow analysis.
 * Unneeded checks will be optimized away by later passes, as can be seen in
 * the below gimple dump of the 'optimized' pass:
 *
 * 100 test_creds (struct cred * cred)
 * 101 {
 * 102   unsigned int canary_check.1125;
 * 103   unsigned int canary_check.1121;
 * 104   unsigned int canary_check.1120;
 * 105   unsigned int canary_addr.1119;
 * 106   long unsigned int canary_addr.1118;
 * 107   unsigned int * canary_addr.1117;
 * 108   unsigned int canary_check.1116;
 * 109   unsigned int uid$val;
 * 110   const struct cred * old;
 * 111   const struct cred * new;
 * 112   struct user_namespace * _1;
 * 113   _Bool _2;
 * 114   unsigned int _3;
 * 115   unsigned int _4;
 * 116   bool _5;
 * 117   unsigned int _6;
 * 118   _Bool prephitmp_25;
 * 119   _Bool _29;
 * 120
 * 121   <bb 2> [local count: 1073741824]:
 * 122   new_9 = prepare_kernel_cred (0B);
 * 123   old_11 = override_creds (new_9);
 * 124   revert_creds (old_11);
 * 125   canary_check.1116_18 = cred_13(D)->canary;
 * 126   canary_addr.1117_19 = &cred_13(D)->canary;
 * 127   canary_addr.1118_20 = (long unsigned int) canary_addr.1117_19;
 * 128   canary_addr.1119_21 = (unsigned int) canary_addr.1118_20;
 * 129   canary_check.1120_22 = canary_check.1116_18 ^ canary_addr.1119_21;
 * 130   if (canary_check.1120_22 != 1461575499)
 * 131     goto <bb 4>; [0.05%]
 * 132   else
 * 133     goto <bb 3>; [99.95%]
 * 134
 * 135   <bb 3> [local count: 1073741824]:
 * 136   _1 = cred_13(D)->user_ns;
 * 137   _2 = userns_may_setgroups (_1);
 * 138   if (_2 != 0)
 * 139     goto <bb 9>; [34.00%]
 * 140   else
 * 141     goto <bb 5>; [66.00%]
 * 142
 * 143   <bb 4> [local count: 891194]:
 * 144   __builtin_trap ();
 * 145
 * 146   <bb 5> [local count: 708669601]:
 * 147   canary_check.1121_24 = cred_13(D)->canary;
 * 148   canary_check.1125_28 = canary_addr.1119_21 ^ canary_check.1121_24;
 * 149   if (canary_check.1125_28 != 1461575499)
 * 150     goto <bb 4>; [0.05%]
 * 151   else
 * 152     goto <bb 6>; [99.95%]
 * 153
 * 154   <bb 6> [local count: 708669601]:
 * 155   uid$val_17 = cred_13(D)->uid.val;
 * 156   if (uid$val_17 == 0)
 * 157     goto <bb 7>; [50.00%]
 * 158   else
 * 159     goto <bb 8>; [50.00%]
 * 160
 * 161   <bb 7> [local count: 354334800]:
 * 162   _3 = cred_13(D)->securebits;
 * 163   _4 = _3 & 1;
 * 164   _6 = 1 - _4;
 * 165   _29 = (_Bool) _6;
 * 166
 * 167   <bb 8> [local count: 708669601]:
 * 168   # prephitmp_25 = PHI <_29(7), _2(6)>
 * 169
 * 170   <bb 9> [local count: 1073741824]:
 * 171   # _5 = PHI <prephitmp_25(8), _2(3)>
 * 172   return _5;
 * 173 }
 *
 * Only two checks remain: the initial one for the '->user_ns' dereference
 * and the one preceding the '->uid' and '->securebits' loads.
 *
 * Even more, taking the address of the canary member and casting it to the
 * canary's type is also done only once and the value reused for the second
 * check (lines 126-128, reused in line 148).
 *
 * Even though the amount of instructions needed to do the canary checks looks
 * ample at first, it boils down to only two or three instructions, highlighted
 * below:
 *
 * ffffffff811c4bd0 <test_creds>:
 * ffffffff811c4bd0:              push   %rbp
 * ffffffff811c4bd1:              push   %rbx
 * ffffffff811c4bd2:              mov    %rdi,%rbx
 * ffffffff811c4bd5:              xor    %edi,%edi
 * ffffffff811c4bd7:              xor    %eax,%eax
 * ffffffff811c4bd9:              callq  ffffffff811d5550 <prepare_kernel_cred>
 * ffffffff811c4bde: (1)          lea    0x28(%rbx),%rbp
 * ffffffff811c4be2:              mov    %rax,%rdi
 * ffffffff811c4be5:              xor    %eax,%eax
 * ffffffff811c4be7:              callq  ffffffff811d4a60 <override_creds>
 * ffffffff811c4bec:              mov    %rax,%rdi
 * ffffffff811c4bef:              callq  ffffffff811d4e90 <revert_creds>
 * ffffffff811c4bf4: (2a)         mov    0x28(%rbx),%eax
 * ffffffff811c4bf7: (2b)         xor    %ebp,%eax
 * ffffffff811c4bf9: (2c)         cmp    $0x571ddf4b,%eax
 * ffffffff811c4bfe:       /----- jne    ffffffff811c4c34 <test_creds+0x64>
 * ffffffff811c4c00:       |      mov    0x88(%rbx),%rdi
 * ffffffff811c4c07:       |      xor    %eax,%eax
 * ffffffff811c4c09:       |      callq  ffffffff81280f90 <userns_may_setgroups>
 * ffffffff811c4c0e:       |      test   %al,%al
 * ffffffff811c4c10:       |  /-- jne    ffffffff811c4c2c <test_creds+0x5c>
 * ffffffff811c4c12: (3a)  |  |   xor    0x28(%rbx),%ebp
 * ffffffff811c4c15: (3b)  |  |   cmp    $0x571ddf4b,%ebp
 * ffffffff811c4c1b:       +--|-- jne    ffffffff811c4c34 <test_creds+0x64>
 * ffffffff811c4c1d:       |  |   mov    0x4(%rbx),%edx
 * ffffffff811c4c20:       |  |   test   %edx,%edx
 * ffffffff811c4c22:       |  +-- jne    ffffffff811c4c2c <test_creds+0x5c>
 * ffffffff811c4c24:       |  |   mov    0x24(%rbx),%eax
 * ffffffff811c4c27:       |  |   not    %eax
 * ffffffff811c4c29:       |  |   and    $0x1,%eax
 * ffffffff811c4c2c:       |  \-> pop    %rbx
 * ffffffff811c4c2d:       |      pop    %rbp
 * ffffffff811c4c2e:       |      jmpq   ffffffff81fe2580 <__x86_return_thunk>
 * ffffffff811c4c33:       |      int3
 * ffffffff811c4c34:       \----> ud2
 *
 * At (1) the address of the 'cred->canary' member gets loaded into RBP.
 * Instructions (2a-c) are the 'cred->canary' load (a), the XOR with its
 * address (b) and (c) the compare against the compile time constant.
 * As can be seen, the casts of the canary member address, which took two
 * instructions in GIMPLE, have all been folded into the XOR operation by
 * making the instruction execute on 32-bit wide registers.
 * This got even further optimized for the check in (3a-b), which combines
 * the 'cred->canary' load, cast to 'unsigned int' and XOR with its address
 * into a single instruction (3a).
 *
 * This is even less for the 'static' canary mode which doesn't take an
 * object's address into account and only checks the runtime value against
 * a fixed compile time random value.
 *
 * Temporarily changing 'struct cred' to make use of the 'static' mode by
 * annotating it with __canary("static") instead leads to instrumentation
 * like below:
 *
 * ffffffff811c49b0 <test_creds>:
 * ffffffff811c49b0:              push   %rbx
 * ffffffff811c49b1:              mov    %rdi,%rbx
 * ffffffff811c49b4:              xor    %edi,%edi
 * ffffffff811c49b6:              xor    %eax,%eax
 * ffffffff811c49b8:              callq  ffffffff811d5110 <prepare_kernel_cred>
 * ffffffff811c49bd:              mov    %rax,%rdi
 * ffffffff811c49c0:              xor    %eax,%eax
 * ffffffff811c49c2:              callq  ffffffff811d46c0 <override_creds>
 * ffffffff811c49c7:              mov    %rax,%rdi
 * ffffffff811c49ca:              callq  ffffffff811d4ac0 <revert_creds>
 * ffffffff811c49cf: (1)          cmpl   $0x571ddf4b,0x28(%rbx)
 * ffffffff811c49d6:       /----- jne    ffffffff811c4a09 <test_creds+0x59>
 * ffffffff811c49d8:       |      mov    0x88(%rbx),%rdi
 * ffffffff811c49df:       |      xor    %eax,%eax
 * ffffffff811c49e1:       |      callq  ffffffff81280970 <userns_may_setgroups>
 * ffffffff811c49e6:       |      test   %al,%al
 * ffffffff811c49e8:       |  /-- jne    ffffffff811c4a02 <test_creds+0x52>
 * ffffffff811c49ea: (2)   |  |   cmpl   $0x571ddf4b,0x28(%rbx)
 * ffffffff811c49f1:       +--|-- jne    ffffffff811c4a09 <test_creds+0x59>
 * ffffffff811c49f3:       |  |   mov    0x4(%rbx),%edx
 * ffffffff811c49f6:       |  |   test   %edx,%edx
 * ffffffff811c49f8:       |  +-- jne    ffffffff811c4a02 <test_creds+0x52>
 * ffffffff811c49fa:       |  |   mov    0x24(%rbx),%eax
 * ffffffff811c49fd:       |  |   not    %eax
 * ffffffff811c49ff:       |  |   and    $0x1,%eax
 * ffffffff811c4a02:       |  \-> pop    %rbx
 * ffffffff811c4a03:       |      jmpq   ffffffff81fe1580 <__x86_return_thunk>
 * ffffffff811c4a08:       |      int3
 * ffffffff811c4a09:       \----> ud2
 *
 * The 'cred->canary' member load and compare against the magic value is folded
 * into a single instruction (1), (2) respectively.
 *
 * However, 'static' mode uses only fixed (compile time random) values, making
 * it easier for an attacker to forge fake objects as soon as the magic value
 * leaked. In 'dynamic' mode an attacker must also be aware of the forged
 * object's address to be able to provide a valid canary value, making it
 * harder to exploit use-after-free bugs targeting canary protected types.
 */
static unsigned int type_canary_execute(void)
{
	basic_block bb;

	loop_optimizer_init(LOOPS_NORMAL | LOOPS_HAVE_RECORDED_EXITS);
	gcc_assert(current_loops);

	calculate_dominance_info(CDI_DOMINATORS);
	calculate_dominance_info(CDI_POST_DOMINATORS);

	/* Evaluate all nodes and instrument, where needed. */
	FOR_EACH_BB_FN(bb, cfun) {
		gimple_stmt_iterator gsi;

		for (gsi = gsi_start_bb(bb); !gsi_end_p(gsi); gsi_next(&gsi)) {
			gimple_stmt_ptr stmt = gsi_stmt(gsi);

			if (!gimple_references_memory_p(stmt))
				continue;

			if (canary_handle_stmt(&gsi)) {
				/*
				 * Instrumenting a statement hijacks its BB requiring us to
				 * update the 'bb' pointer to avoid seeing (and instrumenting)
				 * statements twice. We will, however, see the BB for the call
				 * to __builtin_trap(). But that's benign, as neither will it
				 * match our search pattern nor is it big enough to cause any
				 * measurable compile time slowdown.
				 */
				bb = gsi_bb(gsi);
			}

			gcc_assert(gsi_stmt(gsi) == stmt);
		}
	}

	free_dominance_info(CDI_DOMINATORS);
	free_dominance_info(CDI_POST_DOMINATORS);
	loop_optimizer_finalize();

	return 0;
}

static bool type_canary_gate(void)
{
	if (lookup_attribute("no_instrument_function", DECL_ATTRIBUTES(current_function_decl)))
		return false;

	if (has_trait(current_function_decl, "nocanary"))
		return false;

	return true;
}

#define PASS_NAME type_canary
#define PROPERTIES_REQUIRED PROP_cfg
#define TODO_FLAGS_FINISH TODO_verify_il | TODO_update_ssa | TODO_update_address_taken

#include "gcc-generate-gimple-pass.h"

__visible int plugin_init(struct plugin_name_args *plugin_info, struct plugin_gcc_version *version)
{
	static const struct ggc_root_tab gt_ggc_r_canary[] = {
		INIT_GGC_ROOT_TAB_TREE(canary_copy_builtin_decl),
		INIT_GGC_ROOT_TAB_TREE(canary_value_builtin_decl),
		LAST_GGC_ROOT_TAB
	};
	const char *const plugin_name = plugin_info->base_name;
	const struct plugin_argument *argv = plugin_info->argv;
	const int argc = plugin_info->argc;
	int i;

	PASS_INFO(type_canary, "sra", 0, PASS_POS_INSERT_AFTER);

	if (!plugin_default_version_check(version, &gcc_version)) {
		error_gcc_version(version);
		return 1;
	}

	register_callback(plugin_name, PLUGIN_INFO, NULL, &type_canary_plugin_info);

	for (i = 0; i < argc; ++i) {
		if (!strcmp(argv[i].key, "verbose")) {
			verbose++;
			continue;
		}

		error(G_("unknown option '-fplugin-arg-%s-%s'"), plugin_name, argv[i].key);
	}

	register_callback(plugin_name, PLUGIN_ATTRIBUTES, register_attributes, NULL);
	register_callback(plugin_name, PLUGIN_START_UNIT, register_builtins, NULL);
	register_callback(plugin_name, PLUGIN_FINISH_TYPE, finish_type, NULL);
	register_callback(plugin_name, PLUGIN_ALL_PASSES_START, typemap_cleanup, NULL);
	register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &type_canary_pass_info);
	register_callback(plugin_name, PLUGIN_REGISTER_GGC_ROOTS, NULL, (void *)&gt_ggc_r_canary);

	return 0;
}
