/*
 * Copyright 2024 by PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 *
 * implement the 'late_inline' function attribute which allows inlining
 * such functions after the normal inline passes (early inline and ipa-inline).
 *
 * NOTE: only such functions will be late inlined (recursively), everything else
 * will not be, even if they have been marked inline or even always_inline (they
 * should not survive after ipa-inline but if they do you will get linker errors
 * since the out-of-line calls won't resolve).
 *
 * on gcc 12 and older the late_inline function must be used before ipa-inline
 * on gcc 13 and newer the late_inline function can be used after ipa-inline
 *
 * attribute usage rules:
 * - the late_inline function must be declared 'static inline'
 * - the late_inline function must not be declared 'always_inline' or 'noinline'
 *
 * plugin rules:
 * - must declare for its pass:
 *   #define TODO_FLAGS_START TODO_rebuild_cgraph_edges
 *   - this will create all missing cgraph edges (some gcc passes are sloppy...)
 * - it must only define a gate if skipped functions cannot have late_inline calls
 *   - otherwise all or parts of the gate filtering must be moved to the execute callback
 *     - e.g., if some calls can originate from src
 * - the execute callback must process all of its own late_inline calls at the end by calling
 *   clone_late_inlined_nodes(callee_edge);
 *   - this will prepare new cgraph nodes for inlining
 * - the execute callback must return with
 *   return optimize_inline_calls(current_function_decl);
 *   - this will do the actual work of cloning/inlining the gimple stmts
 *   - this is an expensive operation, should only be called if there's any actual work to do
 *   - see respectre_execute for an example
 */

#include "utilities.h"

static tree handle_late_inline_attribute(tree *node, tree name, tree args, int flags, bool *no_add_attrs)
{
	tree decl = *node;
	location_t loc = DECL_SOURCE_LOCATION(decl);

	*no_add_attrs = true;

	if (TREE_CODE(decl) != FUNCTION_DECL) {
		error_at(loc, "%qE attribute only applies to functions", name);
		return NULL_TREE;
	}

	if (TREE_PUBLIC(decl)) {
		error_at(loc, "function %qD with attribute %qE must be static", decl, name);
		return NULL_TREE;
	}

	if (!DECL_DECLARED_INLINE_P(decl)) {
		error_at(loc, "function %qD with attribute %qE must be inline", decl, name);
		return NULL_TREE;
	}

	if (DECL_DISREGARD_INLINE_LIMITS(decl)) {
		error_at(loc, "function %qD with attribute %qE must not be always_inline", decl, name);
		return NULL_TREE;
	}

	TREE_PUBLIC(decl) = 1;

#if BUILDING_GCC_VERSION >= 13000
	DECL_ABSTRACT_P(decl) = 1; // do not emit debug information
#endif

	*no_add_attrs = false;
	return NULL_TREE;
}

#if BUILDING_GCC_VERSION < 8000
struct attribute_spec late_inline_attr = ATTRIBUTE_SPEC_INIT("late_inline", 0, 0, true, false, false, handle_late_inline_attribute, false, NULL);
#else
static const struct attribute_spec::exclusions attr_late_inline_exclusions[] = {
	{ "always_inline", true, true, true },
	{ "noinline", true, true, true },
	{ NULL, false, false, false },
};

struct attribute_spec late_inline_attr = ATTRIBUTE_SPEC_INIT("late_inline", 0, 0, true, false, false, handle_late_inline_attribute, false, attr_late_inline_exclusions);
#endif

#if BUILDING_GCC_VERSION >= 13000
static const bool allow_inline = false;
#else
static bool allow_inline = false;
#endif

static bool (*orig_late_inline_can_inline_p)(tree, tree);

static bool late_inline_can_inline_p(tree caller, tree callee)
{
	if (lookup_attribute("late_inline", DECL_ATTRIBUTES(callee))) {
		DECL_DISREGARD_INLINE_LIMITS(callee) = allow_inline;
		return allow_inline;
	}

	return orig_late_inline_can_inline_p(caller, callee);
}

void late_inline_register_hook(void)
{
	orig_late_inline_can_inline_p = targetm.target_option.can_inline_p;
	targetm.target_option.can_inline_p = late_inline_can_inline_p;
}

#if BUILDING_GCC_VERSION >= 13000
void late_inline_delay_inlines(void *event_data __unused, void *data __unused)
{
	cgraph_node_ptr node;

	FOR_EACH_FUNCTION(node) {
		tree fndecl;

		fndecl = NODE_DECL(node);
		gcc_assert(fndecl);
		if (!lookup_attribute("late_inline", DECL_ATTRIBUTES(fndecl)))
			continue;

		node->gc_candidate = 1;
		node->no_reorder = false;
	}
}
#else
static unsigned int late_inline_execute(void)
{
	struct cgraph_node *node;

	// turn 'late_inline' into 'always_inline'
	gcc_assert(!allow_inline);
	allow_inline = true;

	FOR_EACH_DEFINED_FUNCTION(node) {
		tree fndecl = NODE_DECL(node);

		gcc_assert(fndecl);
		if (!lookup_attribute("late_inline", DECL_ATTRIBUTES(fndecl)))
			continue;

		// emit proper debug information
		DECL_POSSIBLY_INLINED(fndecl) = 1;

		// trick mark_functions_to_output to not emit it
		node->process = 0;
		TREE_ASM_WRITTEN(fndecl) = 1;
		DECL_EXTERNAL(fndecl) = 1;
	}

	return 0;
}

#define PASS_NAME late_inline
#define NO_WRITE_SUMMARY
#define NO_GENERATE_SUMMARY
#define NO_READ_SUMMARY
#define NO_READ_OPTIMIZATION_SUMMARY
#define NO_WRITE_OPTIMIZATION_SUMMARY
#define NO_STMT_FIXUP
#define NO_FUNCTION_TRANSFORM
#define NO_VARIABLE_TRANSFORM
#define NO_GATE

#include "gcc-generate-ipa-pass.h"
#endif
