/*
 * Copyright 2021-2024 by PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 *
 * implement the 'maybe_used' function attribute that preserves the function
 * at least until its potential use in pre-inline instrumentation passes.
 *
 * functions not used otherwise after inlining can be freely omitted by gcc.
 *
 * NOTE: functions marked with 'maybe_used' MUST also be marked with 'used'
 */

#include "utilities.h"

static tree handle_maybe_used_attribute(tree *node, tree name, tree args, int flags, bool *no_add_attrs)
{
	tree decl;
	location_t loc;

	*no_add_attrs = true;
	decl = *node;
	loc = DECL_SOURCE_LOCATION(decl);

	if (TREE_CODE(decl) != FUNCTION_DECL) {
		warning_at(loc, OPT_Wattributes, "%qE attribute only applies to functions", name);
		return NULL_TREE;
	}

	if (!lookup_attribute("used", DECL_ATTRIBUTES(decl))) {
		error_at(loc, "%qE attribute requires attribute 'used' first", name);
		return NULL_TREE;
	}

	gcc_assert(TREE_USED(decl));
	gcc_assert(DECL_PRESERVE_P(decl));

	*no_add_attrs = false;
	return NULL_TREE;
}

struct attribute_spec maybe_used_attr = ATTRIBUTE_SPEC_INIT("maybe_used", 0, 0, true, false, false, handle_maybe_used_attribute, false, NULL);

unsigned int unmark_maybe_used_functions_execute(void)
{
	cgraph_node_ptr node;

	FOR_EACH_FUNCTION(node) {
		tree fndecl;

		if (NODE_THUNK(node) || node->alias)
			continue;
		gcc_assert(!NODE_GLOBAL(node)->inlined_to);

		fndecl = NODE_DECL(node);
		if (DECL_EXTERNAL(fndecl))
			continue;
		if (!lookup_attribute("maybe_used", DECL_ATTRIBUTES(fndecl)))
			continue;

		gcc_assert(TREE_USED(fndecl));
		TREE_USED(fndecl) = 0;

		gcc_assert(DECL_PRESERVE_P(fndecl));
		DECL_PRESERVE_P(fndecl) = 0;

		gcc_assert((DECL_ABSTRACT_ORIGIN(fndecl) && DECL_ABSTRACT_ORIGIN(fndecl) != fndecl) || node->force_output);
		node->force_output = 0;
	}

	return 0;
}

#define PASS_NAME unmark_maybe_used_functions
#define NO_GATE
#include "gcc-generate-simple_ipa-pass.h"
