/*
 * Copyright 2024 by PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 *
 * implement the x86 'no_caller_saved_registers' type attribute for gcc 6 and older (only bits needed by Linux on amd64!)
 *
 * strategy that tries to follow the gcc 7+ implementation (gcc commit f8071c056663dad2458c96705b459561cd5ea130):
 * - ix86_conditional_register_usage
 *   - hooked via targetm.conditional_register_usage
 *   - re-executed as necessary via hooking targetm.set_current_function and calling reinit_regs
 * - ix86_epilogue_uses
 *   - unhandled (cannot be hooked)
 * - ix86_hard_regno_scratch_ok
 *   - hooked via targetm.hard_regno_scratch_ok
 * - ix86_save_reg
 *   - cannot be hooked
 *   - handled in utilities_conditional_register_usage by taking into account the registers used for argument passing
 * - find_drap_reg
 *   - handled by hooking targetm.calls.get_drap_rtx
 * - ix86_function_ok_for_sibcall
 *   - handled by hooking targetm.function_ok_for_sibcall
 * - ix86_expand_call
 *   - handled by hooking targetm.gen_call and targetm.gen_call_value
 *
 * TODO: implement its full semantics, see https://gcc.gnu.org/onlinedocs/gcc/x86-Function-Attributes.html#index-no_005fcaller_005fsaved_005fregisters-function-attribute_002c-x86
 */

#include "utilities.h"

#if defined(TARGET_386) && BUILDING_GCC_VERSION < 7000
static void (*old_conditional_register_usage)(void);
static void (*old_set_current_function)(tree);
static rtx (*old_get_drap_rtx)(void);
static bool (*old_hard_regno_scratch_ok)(unsigned int);
static bool (*old_function_ok_for_sibcall)(tree, tree);
#if BUILDING_GCC_VERSION >= 6000
static rtx_insn * (*old_gen_call)(rtx, rtx, rtx, rtx);
static rtx_insn * (*old_gen_call_value)(rtx, rtx, rtx, rtx, rtx);
#endif

static tree handle_no_caller_saved_registers_attribute(tree *node, tree name, tree args, int flags, bool *no_add_attrs)
{
	tree type;

	*no_add_attrs = true;
	type = *node;

	if (!FUNC_OR_METHOD_TYPE_P(type)) {
		warning_at(UNKNOWN_LOCATION, OPT_Wattributes, "%qE attribute only applies to function types", name);
		return NULL_TREE;
	}

	*no_add_attrs = false;
	return NULL_TREE;
}

struct attribute_spec no_caller_saved_registers_attr = ATTRIBUTE_SPEC_INIT("no_caller_saved_registers", 0, 0, false, true, true, handle_no_caller_saved_registers_attribute, true, NULL);

static void utilities_conditional_register_usage(void)
{
	unsigned int i;
	tree fntype;
	char c_mask = CALL_USED_REGISTERS_MASK(false);
	static const char ix86_call_used_regs[] = CALL_USED_REGISTERS;

	old_conditional_register_usage();
	if (current_function_decl == NULL_TREE)
		return;

	fntype = TREE_TYPE(current_function_decl);
	if (!lookup_attribute("no_caller_saved_registers", TYPE_ATTRIBUTES(fntype)))
		return;

	for (i = 0; i < FIRST_PSEUDO_REGISTER; i++) {
		if (fixed_regs[i])
			continue;

		// this could be targetm.calls.function_value_regno_p(i) but we'll need to control ix86_save_reg as well
		if (i == AX_REG && !VOID_TYPE_P(TREE_TYPE(TREE_TYPE(current_function_decl))))
			continue;
		if (ix86_call_used_regs[i] != 1 && !(ix86_call_used_regs[i] & c_mask))
			continue;

		call_used_regs[i] = 0;
		if (TEST_HARD_REG_BIT(reg_class_contents[(int)GENERAL_REGS], i))
			CLEAR_HARD_REG_BIT(reg_class_contents[(int)CLOBBERED_REGS], i);
	}
}

static void utilities_set_current_function(tree fndecl)
{
	old_set_current_function(fndecl);
	// TODO: this is strictly only necessary when switching between functions of different no_caller_saved_registers state
	reinit_regs();
}

static rtx utilities_get_drap_rtx(void)
{
	rtx drap, arg_ptr, drap_vreg, body;
	rtx_insn *insn, *seq;

	drap = old_get_drap_rtx();
	if (!drap)
		return drap;

	if (!TARGET_64BIT)
		return drap;

	if (!lookup_attribute("no_caller_saved_registers", TYPE_ATTRIBUTES(TREE_TYPE(current_function_decl))))
		return drap;

	insn = NEXT_INSN(entry_of_function());
	body = PATTERN(insn);
	gcc_assert(GET_CODE(body) == SET);
	gcc_assert(GET_CODE(SET_SRC(body)) == REG);
	if (REGNO(SET_SRC(body)) == R13_REG)
		return drap;

	gcc_assert(REGNO(SET_SRC(body)) == R10_REG);
	arg_ptr = gen_rtx_REG(Pmode, R13_REG);
	crtl->drap_reg = arg_ptr;

	start_sequence();
	drap_vreg = copy_to_reg(arg_ptr);
	seq = get_insns();
	end_sequence();

	insn = emit_insn_before(seq, insn);
	if (!optimize) {
		add_reg_note(insn, REG_CFA_SET_VDRAP, drap_vreg);
		RTX_FRAME_RELATED_P(insn) = 1;
	}

	delete_insn(NEXT_INSN(insn));

	return drap;
}

static bool utilities_hard_regno_scratch_ok(unsigned int regno)
{
	if (!lookup_attribute("no_caller_saved_registers", TYPE_ATTRIBUTES(TREE_TYPE(current_function_decl))))
		return true;

	return !epilogue_completed && df_regs_ever_live_p(regno);
}

static bool utilities_function_ok_for_sibcall(tree decl, tree exp)
{
	if (lookup_attribute("no_caller_saved_registers", TYPE_ATTRIBUTES(TREE_TYPE(current_function_decl))))
		return false;

	return old_function_ok_for_sibcall(decl, exp);
}

static rtx_insn * utilities_gen_call_common(rtx_insn *call)
{
	rtx body;
	tree fndecl, fntype;
	unsigned int i;
	char c_mask = CALL_USED_REGISTERS_MASK(false);
	static const char ix86_call_used_regs[] = CALL_USED_REGISTERS;

	if (!CALL_P(call))
		return call;
	if (SIBLING_CALL_P(call))
		return call;

	body = PATTERN(call);
	if (GET_CODE(body) == SET)
		body = SET_SRC(body);
	gcc_assert(GET_CODE(body) == CALL);

	body = XEXP(body, 0);
	gcc_assert(MEM_P(body));
	body = XEXP(body, 0);
	if (GET_CODE(body) != SYMBOL_REF)
		return call;

	gcc_assert(SYMBOL_REF_FUNCTION_P(body));
	fndecl = SYMBOL_REF_DECL(body);
	fntype = TREE_TYPE(fndecl);
	if (!lookup_attribute("no_caller_saved_registers", TYPE_ATTRIBUTES(TREE_TYPE(current_function_decl))) && !lookup_attribute("no_caller_saved_registers", TYPE_ATTRIBUTES(fntype)))
		return call;

	for (i = 0; i < FIRST_PSEUDO_REGISTER; i++) {
		if (STACK_REGNO_P(i))
			continue;
		if (MMX_REGNO_P(i))
			continue;
		if (fixed_regs[i])
			continue;
		if (ix86_call_used_regs[i] != 1 && !(ix86_call_used_regs[i] & c_mask))
			continue;

		clobber_reg(&CALL_INSN_FUNCTION_USAGE(call), gen_rtx_REG(GET_MODE(regno_reg_rtx[i]), i));
	}

	return call;
}

#if BUILDING_GCC_VERSION < 6000
static unsigned int no_caller_saved_registers_gen_call_execute(void)
{
	rtx_insn *insn;

	for (insn = get_insns(); insn; insn = NEXT_INSN(insn)) {
		if (INSN_DELETED_P(insn))
			continue;

		utilities_gen_call_common(insn);
	}

	return 0;
}

#define PASS_NAME no_caller_saved_registers_gen_call
#define NO_GATE
#include "gcc-generate-rtl-pass.h"
#else
static rtx_insn * utilities_gen_call(rtx op0, rtx op1, rtx op2, rtx op3)
{
	return utilities_gen_call_common(old_gen_call(op0, op1, op2, op3));
}

static rtx_insn * utilities_gen_call_value(rtx op0, rtx op1, rtx op2, rtx op3, rtx op4)
{
	return utilities_gen_call_common(old_gen_call_value(op0, op1, op2, op3, op4));
}
#endif

void no_caller_saved_registers_start_unit(void __unused *gcc_data, void __unused *user_data)
{
	old_set_current_function = targetm.set_current_function;
	targetm.set_current_function = utilities_set_current_function;

	old_conditional_register_usage = targetm.conditional_register_usage;
	targetm.conditional_register_usage = utilities_conditional_register_usage;

	old_get_drap_rtx = targetm.calls.get_drap_rtx;
	targetm.calls.get_drap_rtx = utilities_get_drap_rtx;

	old_hard_regno_scratch_ok = targetm.hard_regno_scratch_ok;
	targetm.hard_regno_scratch_ok = utilities_hard_regno_scratch_ok;

	old_function_ok_for_sibcall = targetm.function_ok_for_sibcall;
	targetm.function_ok_for_sibcall = utilities_function_ok_for_sibcall;

#if BUILDING_GCC_VERSION >= 6000
	old_gen_call = targetm.gen_call;
	targetm.gen_call = utilities_gen_call;

	old_gen_call_value = targetm.gen_call_value;
	targetm.gen_call_value = utilities_gen_call_value;
#endif
}
#endif
