/*
 * Copyright 2020-2024 by PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 *
 * allow DSE on READ_ONCE/WRITE_ONCE accesses
 */

#include "utilities.h"

bool report_null_deref = false;

static tree handle_once_attribute(tree *node, tree name, tree args, int flags, bool *no_add_attrs)
{
	tree *type, orig_type, decl;

	decl = *node;
	gcc_assert(DECL_P(decl));
	gcc_assert(TREE_CODE(decl) == TYPE_DECL);
	gcc_assert(DECL_ORIGINAL_TYPE(decl) == NULL_TREE);
	gcc_assert(!(flags & (int)ATTR_FLAG_TYPE_IN_PLACE));

	type = &TREE_TYPE(decl);
	orig_type = TREE_TYPE(decl);
	if (lookup_attribute(IDENTIFIER_POINTER(name), TYPE_ATTRIBUTES(orig_type)))
		return NULL_TREE;

	if (RECORD_OR_UNION_TYPE_P(orig_type)) {
		gcc_assert(TYPE_FIELDS(orig_type));
		if (TREE_CHAIN(TYPE_FIELDS(orig_type))) {
//			inform(DECL_SOURCE_LOCATION(decl), "invalid ONCE access to %qD of type %qT: more than one field", decl, orig_type);
			return NULL_TREE;
		}
		if (tree_int_cst_lt(TYPE_SIZE_UNIT(long_unsigned_type_node), TYPE_SIZE_UNIT(orig_type))) {
//			inform(DECL_SOURCE_LOCATION(decl), "invalid ONCE access to %qD of type %qT: size too big", decl, orig_type);
			return NULL_TREE;
		}
	}

	if (POINTER_TYPE_P(orig_type)) {
		if (POINTER_TYPE_P(TREE_TYPE(orig_type))) {
			if (POINTER_TYPE_P(TREE_TYPE(TREE_TYPE(orig_type)))) {
				gcc_assert(!POINTER_TYPE_P(TREE_TYPE(TREE_TYPE(TREE_TYPE(orig_type)))));
				*type = build_pointer_type(build_pointer_type(build_pointer_type(build_variant_type_copy(TREE_TYPE(TREE_TYPE(TREE_TYPE(orig_type)))))));
			} else
				*type = build_pointer_type(build_pointer_type(build_variant_type_copy(TREE_TYPE(TREE_TYPE(orig_type)))));
		} else
			*type = build_pointer_type(build_variant_type_copy(TREE_TYPE(orig_type)));
	} else
		*type = build_variant_type_copy(orig_type);

	TYPE_ATTRIBUTES(*type) = tree_cons(name, NULL_TREE, TYPE_ATTRIBUTES(orig_type));
	*no_add_attrs = false;
	return NULL_TREE;
}

struct attribute_spec read_once_attr = ATTRIBUTE_SPEC_INIT("read_once", 0, 0, true, false, false, handle_once_attribute, false, NULL);
struct attribute_spec write_once_attr = ATTRIBUTE_SPEC_INIT("write_once", 0, 0, true, false, false, handle_once_attribute, false, NULL);

static inline void gimple_set_write_once(gimple_stmt_ptr stmt, bool val_p)
{
	gimple_set_plf(stmt, GF_PLF_2, val_p);
}

static inline bool gimple_write_once_p(gimple_stmt_ptr stmt)
{
	return gimple_plf(stmt, GF_PLF_2);
}

static inline void gimple_set_read_once(gimple_stmt_ptr stmt, bool val_p)
{
	stmt->pad = val_p;
}

static inline bool gimple_read_once_p(gimple_stmt_ptr stmt)
{
	return stmt->pad == 1;
}

static void undo_volatile(tree mem)
{
	tree type;
	int quals;

	TREE_THIS_VOLATILE(mem) = 0;
	TREE_SIDE_EFFECTS(mem) = 0;

	type = TREE_TYPE(mem);
	quals = TYPE_QUALS(type) & ~TYPE_QUAL_VOLATILE;
	type = build_qualified_type(type, quals);
	TREE_TYPE(mem) = type;
}

static void redo_volatile(tree mem)
{
	tree type;
	int quals;

	TREE_THIS_VOLATILE(mem) = 1;
	TREE_SIDE_EFFECTS(mem) = 1;

	type = TREE_TYPE(mem);
	quals = TYPE_QUALS(type) | TYPE_QUAL_VOLATILE;
	type = build_qualified_type(type, quals);
	TREE_TYPE(mem) = type;
}

static tree once_do_volatile(tree *tp, int *walk_subtrees, void *data)
{
	tree var = *tp;

	*walk_subtrees = 0;

	switch (TREE_CODE(var)) {
	default:
		debug_tree(var);
		gcc_unreachable();

	case ADDR_EXPR:
	case FIELD_DECL:
	case INTEGER_CST:
	case SSA_NAME:
	case VAR_DECL:
		break;

	case ARRAY_REF:
	case COMPONENT_REF:
	case INDIRECT_REF:
		gcc_assert(TREE_THIS_VOLATILE(var) == TREE_SIDE_EFFECTS(var));
		gcc_assert(!lookup_attribute("write_once", TYPE_ATTRIBUTES(TREE_TYPE(var))) && !lookup_attribute("read_once", TYPE_ATTRIBUTES(TREE_TYPE(var))));
		*walk_subtrees = 1;
		break;

	case MEM_REF:
		gcc_assert(TREE_THIS_VOLATILE(var) == TREE_SIDE_EFFECTS(var));
		*walk_subtrees = 1;
		break;
	}

	return NULL_TREE;
}

static void remove_ssa_names(tree var)
{
	tree ssaname;
	size_t i;

	FOR_EACH_SSA_NAME(i, ssaname, cfun) {
		imm_use_iterator imm_iter;
		gimple_stmt_ptr stmt;

		if (virtual_operand_p(ssaname))
			continue;

		if (SSA_NAME_VAR(ssaname) != var)
			continue;

		FOR_EACH_IMM_USE_STMT(stmt, imm_iter, ssaname) {
			use_operand_p use;

			if (gimple_code(stmt) == GIMPLE_PHI) {
				gimple_stmt_iterator gsi;

				gsi = gsi_for_stmt(stmt);
				remove_phi_node(&gsi, false);
				continue;
			}

			FOR_EACH_IMM_USE_ON_STMT(use, imm_iter) {
				SET_USE(use, var);
			}
			update_stmt(stmt);
		}

		stmt = SSA_NAME_DEF_STMT(ssaname);
		if (SSA_NAME_IS_DEFAULT_DEF(ssaname))
			set_ssa_default_def(cfun, var, NULL_TREE);
		release_ssa_name(ssaname);
		if (!is_gimple_assign(stmt))
			continue;
		gimple_assign_set_lhs(stmt, var);
		update_stmt(stmt);
	}
}

static bool once_null_deref(gimple_stmt_ptr stmt, tree arg)
{
	location_t loc;
	tree ptr;
	gimple_stmt_ptr def_stmt;
	gphi *phi_stmt;
	unsigned int i, num_args;

	loc = gimple_location(stmt);

	if (TREE_CODE(arg) != COMPONENT_REF)
		return false;

	ptr = get_inner_arg(arg);
	if (TREE_CODE(ptr) != MEM_REF)
		return false;

	ptr = TREE_OPERAND(ptr, 0);
	if (TREE_CODE(ptr) == INTEGER_CST) {
		if (report_null_deref)
			inform(loc, "%s.%u: null pointer dereference", current_pass->name, current_pass->static_pass_number);
		return true;
	}

	gcc_assert(TREE_CODE(ptr) == SSA_NAME);
	def_stmt = SSA_NAME_DEF_STMT(ptr);
	if (gimple_code(def_stmt) != GIMPLE_PHI)
		return false;

	phi_stmt = as_a_gphi(def_stmt);

	for (i = 0, num_args =  gimple_phi_num_args(phi_stmt); i < num_args; i++) {
		tree phi_arg = gimple_phi_arg_def(phi_stmt, i);

		if (TREE_CODE(phi_arg) != INTEGER_CST)
			continue;
		if (report_null_deref)
			inform(loc, "%s.%u: null pointer dereference", current_pass->name, current_pass->static_pass_number);
		return true;
	}

	return false;
}

static bool write_once_p(gimple_stmt_ptr stmt, bool sideeffects)
{
	tree lhs;

	if (!gimple_assign_copy_p(stmt))
		return false;
	if (gimple_clobber_p(stmt))
		return false;

	lhs = gimple_assign_lhs(stmt);

	if (sideeffects && !gimple_has_volatile_ops(stmt))
		return false;

	switch (TREE_CODE(lhs)) {
	default:
		if (lookup_attribute("write_once", TYPE_ATTRIBUTES(TREE_TYPE(lhs))))
			return true;
		lhs = get_inner_ref(lhs);
		if (lookup_attribute("write_once", TYPE_ATTRIBUTES(TREE_TYPE(lhs))))
			return true;
		if (TREE_CODE(lhs) != MEM_REF)
			return false;
		lhs = TREE_OPERAND(lhs, 1);
		gcc_assert(TREE_CODE(lhs) == INTEGER_CST);
		lhs = TREE_TYPE(lhs);
		gcc_assert(POINTER_TYPE_P(lhs));
		return lookup_attribute("write_once", TYPE_ATTRIBUTES(TREE_TYPE(lhs))) != NULL_TREE;

	case VAR_DECL:
		if (sideeffects)
			return lookup_attribute("write_once", TYPE_ATTRIBUTES(TREE_TYPE(lhs))) != NULL_TREE;
		return true;

	case SSA_NAME:
		if (sideeffects)
			return false;
		gcc_assert(SSA_NAME_VAR(lhs) != NULL_TREE);
		return true;
	}
}

static bool write_once(gimple_stmt_ptr stmt, bool sideeffects)
{
	tree lhs, rhs;

	if (!write_once_p(stmt, sideeffects))
		return false;

	lhs = gimple_assign_lhs(stmt);
	rhs = gimple_assign_rhs1(stmt);

	gcc_assert(TREE_THIS_VOLATILE(lhs) == TREE_SIDE_EFFECTS(lhs));
	// when gcc emits a NULL deref, it can lose TREE_THIS_VOLATILE
	if (once_null_deref(stmt, lhs))
		return true;
	gcc_assert(TREE_THIS_VOLATILE(lhs) == sideeffects);
	gcc_assert(TREE_CODE(rhs) != MEM_REF || !lookup_attribute("write_once", TYPE_ATTRIBUTES(TREE_TYPE(rhs))));

	switch (TREE_CODE(lhs)) {
	default:
		debug_gimple_stmt(stmt);
		debug_tree(lhs);
		gcc_unreachable();

	case COMPONENT_REF:
		if (sideeffects)
			undo_volatile(lhs);
		else
			redo_volatile(lhs);
		walk_tree_without_duplicates(&TREE_OPERAND(lhs, 0), once_do_volatile, sideeffects ? lhs : NULL);
		break;

	case ARRAY_REF:
	case MEM_REF:
		if (sideeffects)
			undo_volatile(lhs);
		else
			redo_volatile(lhs);
		walk_tree_without_duplicates(&TREE_OPERAND(lhs, 0), once_do_volatile, sideeffects ? lhs : NULL);
		break;

	case SSA_NAME:
	case VAR_DECL: {
		location_t loc;
		tree var;

		gcc_assert(!sideeffects);

		loc = gimple_location(stmt);
		if (TREE_CODE(lhs) == SSA_NAME) {
			gcc_assert(SSA_NAME_VAR(lhs));
			var = SSA_NAME_VAR(lhs);
		} else
			var = lhs;
		TREE_ADDRESSABLE(var) = 1;
		lhs = build1_loc(loc, ADDR_EXPR, TYPE_POINTER_TO(TREE_TYPE(var)), var);
		lhs = build2_loc(loc, MEM_REF, TREE_TYPE(TREE_TYPE(lhs)), lhs, build_int_cst(TREE_TYPE(lhs), 0));
		redo_volatile(lhs);
		gimple_assign_set_lhs(stmt, lhs);
		gimple_set_write_once(stmt, sideeffects);
		update_stmt(stmt);

		remove_ssa_names(var);
		return true;
	}
	}

	gimple_set_has_volatile_ops(stmt, !sideeffects);
	gimple_set_write_once(stmt, sideeffects);
	update_stmt(stmt);
	return true;
}

static bool read_once_p(gimple_stmt_ptr stmt, bool sideeffects)
{
	tree lhs, rhs;

	if (!gimple_assign_single_p(stmt))
		return false;
	if (gimple_clobber_p(stmt))
		return false;

	if (sideeffects && !gimple_has_volatile_ops(stmt))
		return false;

	rhs = gimple_assign_rhs1(stmt);

	if (sideeffects && TREE_THIS_VOLATILE(get_inner_arg(rhs)))
		return false;

	if (!sideeffects && TREE_CODE(rhs) == SSA_NAME) {
		gcc_assert(SSA_NAME_VAR(rhs) != NULL_TREE);
		return true;
	}

	if (lookup_attribute("read_once", TYPE_ATTRIBUTES(TREE_TYPE(rhs))) != NULL_TREE)
		return true;

	lhs = gimple_assign_lhs(stmt);
	return lookup_attribute("read_once", TYPE_ATTRIBUTES(TREE_TYPE(lhs))) != NULL_TREE;
}

static bool read_once(gimple_stmt_ptr stmt, bool sideeffects)
{
	tree lhs, rhs;

	if (!read_once_p(stmt, sideeffects))
		return false;

	lhs = gimple_assign_lhs(stmt);
	rhs = gimple_assign_rhs1(stmt);

	gcc_assert(TREE_THIS_VOLATILE(rhs) == TREE_SIDE_EFFECTS(rhs));

	// when gcc emits a NULL deref, it can lose TREE_THIS_VOLATILE
	if (once_null_deref(stmt, rhs))
		return true;
	if (sideeffects && TREE_CODE(rhs) == SSA_NAME)
		return true;
	gcc_assert(TREE_THIS_VOLATILE(rhs) == sideeffects);
	gcc_assert(TREE_CODE(lhs) != MEM_REF || !lookup_attribute("read_once", TYPE_ATTRIBUTES(TREE_TYPE(lhs))));

	switch (TREE_CODE(rhs)) {
	default:
		debug_gimple_stmt(stmt);
		debug_tree(rhs);
		gcc_unreachable();

	case COMPONENT_REF:
		if (sideeffects)
			undo_volatile(rhs);
		else
			redo_volatile(rhs);
		walk_tree_without_duplicates(&TREE_OPERAND(rhs, 0), once_do_volatile, sideeffects ? rhs : NULL);
		break;

	case ARRAY_REF:
	case INDIRECT_REF:
	case MEM_REF:
		if (sideeffects)
			undo_volatile(rhs);
		else
			redo_volatile(rhs);
		walk_tree_without_duplicates(&TREE_OPERAND(rhs, 0), once_do_volatile, sideeffects ? rhs : NULL);
		break;

	case SSA_NAME:
		gcc_assert(SSA_NAME_VAR(rhs));
		rhs = SSA_NAME_VAR(rhs);
		// FALLTHROUGH

	case PARM_DECL:
	case VAR_DECL: {
		tree var;
		location_t loc;

		gcc_assert(VAR_P(rhs) || !sideeffects);

		loc = gimple_location(stmt);
		var = rhs;
		TREE_ADDRESSABLE(var) = 1;
		rhs = build1_loc(loc, ADDR_EXPR, TYPE_POINTER_TO(TREE_TYPE(var)), var);
		rhs = build2_loc(loc, MEM_REF, TREE_TYPE(TREE_TYPE(rhs)), rhs, build_int_cst(TREE_TYPE(rhs), 0));
		redo_volatile(rhs);
		gimple_assign_set_rhs1(stmt, rhs);
		gimple_set_read_once(stmt, sideeffects);
		update_stmt(stmt);

		remove_ssa_names(var);
		return true;
	}
	}

	gimple_set_has_volatile_ops(stmt, !sideeffects);
	gimple_set_read_once(stmt, sideeffects);
	update_stmt(stmt);
	return true;
}

static unsigned int once_before_execute(void)
{
	basic_block bb;

	FOR_EACH_BB_FN(bb, cfun) {
		gimple_stmt_iterator gsi;

		for (gsi = gsi_start_bb(bb); !gsi_end_p(gsi); gsi_next(&gsi)) {
			gimple_stmt_ptr stmt;

			stmt = gsi_stmt(gsi);
			gimple_set_write_once(stmt, false);
			gimple_set_read_once(stmt, false);
			if (!write_once(stmt, true))
				read_once(stmt, true);
		}
	}

	return 0;
}

#define PASS_NAME once_before
#define NO_GATE
#define PROPERTIES_REQUIRED PROP_cfg
#define TODO_FLAGS_FINISH TODO_verify_il | TODO_update_ssa | TODO_update_address_taken
#include "gcc-generate-gimple-pass.h"

unsigned int once_after_execute(void)
{
	basic_block bb;

	FOR_EACH_BB_FN(bb, cfun) {
		gimple_stmt_iterator gsi;

		for (gsi = gsi_start_bb(bb); !gsi_end_p(gsi); gsi_next(&gsi)) {
			gimple_stmt_ptr stmt;

			stmt = gsi_stmt(gsi);
			gcc_assert(!gimple_read_once_p(stmt) || !gimple_write_once_p(stmt));
			gcc_assert(!gimple_write_once_p(stmt) || write_once(stmt, false));
			gcc_assert(!gimple_read_once_p(stmt) || read_once(stmt, false));
		}
	}

	return 0;
}

#define PASS_NAME once_after
#define NO_GATE
#define PROPERTIES_REQUIRED PROP_cfg
#define TODO_FLAGS_FINISH TODO_verify_il | TODO_remove_unused_locals | TODO_update_ssa | TODO_update_address_taken
#include "gcc-generate-gimple-pass.h"
