/*
 * Copyright 2020-2024 by PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 *
 * allow specifying some ELF section flags not supported by gcc's section attr
 */

#include "utilities.h"

static bool parse_section_flag(char c, int *flags)
{
	switch (c) {
	default:
		return false;

	case 0:
		return true;

	case '1':
	case '2':
	case '4':
	case '8':
		*flags |= c - '0';
		return true;

	case 'b':
		*flags &= ~SECTION_NOTYPE;
		*flags |= SECTION_BSS;
		return true;

	case 'M':
		*flags |= SECTION_MERGE;
		return true;

	case 'n':
		*flags &= ~SECTION_NOTYPE;
		return true;

	case 'r':
		*flags = 0;
		return true;

	case 'S':
		*flags |= SECTION_STRINGS;
		return true;
	}
}

static tree handle_section_type_attribute(tree *node, tree name, tree args, int flags, bool *no_add_attrs)
{
	tree decl;
	int i, e;

	*no_add_attrs = true;

	decl = *node;
	if (!VAR_OR_FUNCTION_DECL_P(decl)) {
		error_at(DECL_SOURCE_LOCATION(decl), "attribute %qE applies to variables and functions only", name);
		return NULL_TREE;
	}

	args = TREE_VALUE(args);
	if (TREE_CODE(args) != STRING_CST) {
		error_at(DECL_SOURCE_LOCATION(decl), "attribute %qE requires a string argument", name);
		return NULL_TREE;
	}

	for (i = 0, e = TREE_STRING_LENGTH(args); i < e; i++) {
		char c = TREE_STRING_POINTER(args)[i];
		if (parse_section_flag(c, &flags))
			continue;

		error_at(DECL_SOURCE_LOCATION(decl), "unknown value %c provided to attribute %qE", c, name);
		return NULL_TREE;
	}

	*no_add_attrs = false;
	return NULL_TREE;
}

struct attribute_spec section_type_attr = ATTRIBUTE_SPEC_INIT("section_type", 1, 1, true, false, false, handle_section_type_attribute, false, NULL);

static unsigned int (*old_section_type_flags)(tree decl, const char *name, int reloc);

static unsigned int utilities_section_type_flags(tree decl, const char *name, int reloc)
{
	tree section_type;
	unsigned int flags;
	int i, e;

	flags = old_section_type_flags(decl, name, reloc);
	if (decl == NULL_TREE)
		return flags;

	section_type = lookup_attribute("section_type", DECL_ATTRIBUTES(decl));
	if (section_type == NULL_TREE)
		return flags;

	section_type = TREE_VALUE(TREE_VALUE(section_type));
	for (i = 0, e = TREE_STRING_LENGTH(section_type); i < e; i++) {
		char c = TREE_STRING_POINTER(section_type)[i];
		if (parse_section_flag(c, (int *)&flags))
			continue;

		debug_tree(decl);
		gcc_unreachable();
	}

	return flags;
}

void section_type_start_unit(void __unused *gcc_data, void __unused *user_data)
{
	old_section_type_flags = targetm.section_type_flags;
	targetm.section_type_flags = utilities_section_type_flags;
}
